<?php

use Illuminate\Support\Facades\Route;
use Laravel\Jetstream\Http\Controllers\CurrentTeamController;
use Laravel\Jetstream\Http\Controllers\Livewire\ApiTokenController;
use Laravel\Jetstream\Http\Controllers\Livewire\PrivacyPolicyController;
use Laravel\Jetstream\Http\Controllers\Livewire\TeamController;
use Laravel\Jetstream\Http\Controllers\Livewire\TermsOfServiceController;
use Laravel\Jetstream\Http\Controllers\Livewire\UserProfileController;
use Laravel\Jetstream\Http\Controllers\TeamInvitationController;
use Laravel\Jetstream\Jetstream;

Route::group(['middleware' => config('jetstream.middleware', ['web'])], function () {
    if (Jetstream::hasTermsAndPrivacyPolicyFeature()) {
        Route::get('/terms-of-service', [TermsOfServiceController::class, 'show'])->name('terms.show');
        Route::get('/privacy-policy', [PrivacyPolicyController::class, 'show'])->name('policy.show');
    }

    Route::group(['middleware' => ['auth', 'verified']], function () {
        Route::group(['middleware' => 'role:user', 'prefix' => 'user', 'as' => 'user.'], function() {
            // User & Profile...
            Route::get('profile', [UserProfileController::class, 'show'])
                        ->name('profile.show');
        });
        Route::group(['middleware' => 'role:admin', 'prefix' => 'admin', 'namespace' => 'Admin', 'as' => 'admin.'], function() {
            // User & Profile...
            Route::get('profile', [UserProfileController::class, 'show'])
                        ->name('profile.show');
        });

        // API...
        if (Jetstream::hasApiFeatures()) {
            Route::get('/user/api-tokens', [ApiTokenController::class, 'index'])->name('api-tokens.index');
        }

        // Teams...
        if (Jetstream::hasTeamFeatures()) {
            Route::get('/teams/create', [TeamController::class, 'create'])->name('teams.create');
            Route::get('/teams/{team}', [TeamController::class, 'show'])->name('teams.show');
            Route::put('/current-team', [CurrentTeamController::class, 'update'])->name('current-team.update');

            Route::get('/team-invitations/{invitation}', [TeamInvitationController::class, 'accept'])
                        ->middleware(['signed'])
                        ->name('team-invitations.accept');
        }
    });
});

/*

file: vendor\laravel\jetstream\src\Http\Livewire\UpdateProfileInformationForm.php
line : 54

    $user = Auth::user();
    if ($user->id == '1') {
        if (isset($this->photo)) {
            return redirect()->route('admin.profile.show');
        }
    } else {
        if (isset($this->photo)) {
            return redirect()->route('user.profile.show');
        }
    }

*/
