<?php

namespace ccxt\async;

// PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
// https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

use Exception; // a common import
use \ccxt\ExchangeError;
use \ccxt\AuthenticationError;
use \ccxt\ArgumentsRequired;
use \ccxt\OrderNotFound;
use \ccxt\Precise;

class bibox extends Exchange {

    public function describe() {
        return $this->deep_extend(parent::describe (), array(
            'id' => 'bibox',
            'name' => 'Bibox',
            'countries' => array( 'CN', 'US', 'KR' ),
            'version' => 'v1',
            // 30 requests per 5 seconds => 6 requests per second => rateLimit = 166.667 ms (166.6666...)
            'rateLimit' => 166.667,
            'hostname' => 'bibox.com',
            'has' => array(
                'CORS' => null,
                'spot' => true,
                'margin' => null, // has but unimplemented
                'swap' => null, // has but unimplemented
                'future' => null,
                'option' => null,
                'cancelOrder' => true,
                'createMarketOrder' => null, // or they will return https://github.com/ccxt/ccxt/issues/2338
                'createOrder' => true,
                'createStopLimitOrder' => false, // true for contract
                'createStopMarketOrder' => false, // true for contract
                'createStopOrder' => false, // true for contract
                'fetchBalance' => true,
                'fetchBorrowRate' => false,
                'fetchBorrowRates' => false,
                'fetchClosedOrders' => true,
                'fetchCurrencies' => true,
                'fetchDepositAddress' => true,
                'fetchDeposits' => true,
                'fetchMarkets' => true,
                'fetchMyTrades' => true,
                'fetchOHLCV' => true,
                'fetchOpenOrders' => true,
                'fetchOrder' => true,
                'fetchOrderBook' => true,
                'fetchTicker' => true,
                'fetchTickers' => true,
                'fetchTrades' => true,
                'fetchTradingFee' => false,
                'fetchTradingFees' => false,
                'fetchTransactionFees' => true,
                'fetchWithdrawals' => true,
                'withdraw' => true,
            ),
            'timeframes' => array(
                '1m' => '1min',
                '5m' => '5min',
                '15m' => '15min',
                '30m' => '30min',
                '1h' => '1hour',
                '2h' => '2hour',
                '4h' => '4hour',
                '6h' => '6hour',
                '12h' => '12hour',
                '1d' => 'day',
                '1w' => 'week',
            ),
            'urls' => array(
                'logo' => 'https://user-images.githubusercontent.com/51840849/77257418-3262b000-6c85-11ea-8fb8-20bdf20b3592.jpg',
                'api' => 'https://api.{hostname}',
                'www' => 'https://www.bibox365.com',
                'doc' => array(
                    'https://biboxcom.github.io/en/',
                ),
                'fees' => 'https://bibox.zendesk.com/hc/en-us/articles/360002336133',
                'referral' => 'https://w2.bibox365.com/login/register?invite_code=05Kj3I',
            ),
            'api' => array(
                'public' => array(
                    'post' => array(
                        // TODO => rework for full endpoint/cmd paths here
                        'mdata' => 1,
                    ),
                    'get' => array(
                        'cquery' => 1,
                        'mdata' => 1,
                        'cdata' => 1,
                        'orderpending' => 1,
                    ),
                ),
                'private' => array(
                    'post' => array(
                        'cquery' => 1,
                        'ctrade' => 1,
                        'user' => 1,
                        'orderpending' => 1,
                        'transfer' => 1,
                    ),
                ),
                'v2private' => array(
                    'post' => array(
                        'assets/transfer/spot' => 1,
                    ),
                ),
            ),
            'fees' => array(
                'trading' => array(
                    'tierBased' => false,
                    'percentage' => true,
                    'taker' => $this->parse_number('0.002'),
                    'maker' => $this->parse_number('0.001'),
                ),
                'funding' => array(
                    'tierBased' => false,
                    'percentage' => false,
                    'withdraw' => array(),
                    'deposit' => array(),
                ),
            ),
            'exceptions' => array(
                '2011' => '\\ccxt\\AccountSuspended', // Account is locked
                '2015' => '\\ccxt\\AuthenticationError', // Google authenticator is wrong
                '2021' => '\\ccxt\\InsufficientFunds', // Insufficient balance available for withdrawal
                '2027' => '\\ccxt\\InsufficientFunds', // Insufficient balance available (for trade)
                '2033' => '\\ccxt\\OrderNotFound', // operation failed! Orders have been completed or revoked
                '2065' => '\\ccxt\\InvalidOrder', // Precatory price is exorbitant, please reset
                '2066' => '\\ccxt\\InvalidOrder', // Precatory price is low, please reset
                '2067' => '\\ccxt\\InvalidOrder', // Does not support market orders
                '2068' => '\\ccxt\\InvalidOrder', // The number of orders can not be less than
                '2078' => '\\ccxt\\InvalidOrder', // unvalid order price
                '2085' => '\\ccxt\\InvalidOrder', // Order quantity is too small
                '2091' => '\\ccxt\\RateLimitExceeded', // request is too frequency, please try again later
                '2092' => '\\ccxt\\InvalidOrder', // Minimum amount not met
                '2131' => '\\ccxt\\InvalidOrder', // The order quantity cannot be greater than
                '3000' => '\\ccxt\\BadRequest', // Requested parameter incorrect
                '3002' => '\\ccxt\\BadRequest', // Parameter cannot be null
                '3012' => '\\ccxt\\AuthenticationError', // invalid apiKey
                '3016' => '\\ccxt\\BadSymbol', // Trading pair error
                '3024' => '\\ccxt\\PermissionDenied', // wrong apikey permissions
                '3025' => '\\ccxt\\AuthenticationError', // signature failed
                '4000' => '\\ccxt\\ExchangeNotAvailable', // current network is unstable
                '4003' => '\\ccxt\\DDoSProtection', // server busy please try again later
            ),
            'commonCurrencies' => array(
                'APENFT(NFT)' => 'NFT',
                'BOX' => 'DefiBox',
                'BPT' => 'BlockPool Token',
                'GMT' => 'GMT Token',
                'KEY' => 'Bihu',
                'MTC' => 'MTC Mesh Network', // conflict with MTC Docademic doc.com Token https://github.com/ccxt/ccxt/issues/6081 https://github.com/ccxt/ccxt/issues/3025
                'NFT' => 'NFT Protocol',
                'PAI' => 'PCHAIN',
                'REVO' => 'Revo Network',
                'STAR' => 'Starbase',
                'TERN' => 'Ternio-ERC20',
            ),
        ));
    }

    public function fetch_markets($params = array ()) {
        /**
         * retrieves data on all $markets for bibox
         * @param {dict} $params extra parameters specific to the exchange api endpoint
         * @return {[dict]} an array of objects representing $market data
         */
        $request = array(
            'cmd' => 'pairList',
        );
        $response = yield $this->publicGetMdata (array_merge($request, $params));
        //
        //     {
        //         "result" => array(
        //             {
        //                 "id":1,
        //                 "pair":"BIX_BTC",
        //                 "pair_type":0,
        //                 "area_id":7,
        //                 "is_hide":0,
        //                 "decimal":8,
        //                 "amount_scale":4
        //             }
        //         ),
        //         "cmd":"pairList",
        //         "ver":"1.1"
        //     }
        //
        $markets = $this->safe_value($response, 'result', array());
        $request2 = array(
            'cmd' => 'tradeLimit',
        );
        $response2 = yield $this->publicGetOrderpending (array_merge($request2, $params));
        //
        //    {
        //         $result => {
        //             min_trade_price => array( default => '0.00000001', USDT => '0.0001', DAI => '0.0001' ),
        //             min_trade_amount => array( default => '0.0001' ),
        //             min_trade_money => array(
        //                 USDT => '1',
        //                 USDC => '1',
        //                 DAI => '1',
        //                 GUSD => '1',
        //                 BIX => '3',
        //                 BTC => '0.0002',
        //                 ETH => '0.005'
        //             }
        //         ),
        //         cmd => 'tradeLimit'
        //     }
        //
        $result2 = $this->safe_value($response2, 'result', array());
        $minCosts = $this->safe_value($result2, 'min_trade_money', array());
        $result = array();
        for ($i = 0; $i < count($markets); $i++) {
            $market = $markets[$i];
            $numericId = $this->safe_integer($market, 'id');
            $id = $this->safe_string($market, 'pair');
            $baseId = null;
            $quoteId = null;
            if ($id !== null) {
                $parts = explode('_', $id);
                $baseId = $this->safe_string($parts, 0);
                $quoteId = $this->safe_string($parts, 1);
            }
            $base = $this->safe_currency_code($baseId);
            $quote = $this->safe_currency_code($quoteId);
            $symbol = $base . '/' . $quote;
            $type = 'spot';
            $spot = true;
            $areaId = $this->safe_integer($market, 'area_id');
            if ($areaId === 16) {
                // TODO => update to v3 api
                continue;
            }
            $result[] = array(
                'id' => $id,
                'numericId' => $numericId,
                'symbol' => $symbol,
                'base' => $base,
                'quote' => $quote,
                'settle' => null,
                'baseId' => $baseId,
                'quoteId' => $quoteId,
                'settleId' => null,
                'type' => $type,
                'spot' => $spot,
                'margin' => false,
                'swap' => false,
                'future' => false,
                'option' => false,
                'active' => null,
                'contract' => false,
                'linear' => null,
                'inverse' => null,
                'contractSize' => null,
                'expiry' => null,
                'expiryDatetime' => null,
                'strike' => null,
                'optionType' => null,
                'precision' => array(
                    'amount' => $this->safe_integer($market, 'amount_scale'),
                    'price' => $this->safe_integer($market, 'decimal'),
                ),
                'limits' => array(
                    'leverage' => array(
                        'min' => null,
                        'max' => null,
                    ),
                    'amount' => array(
                        'min' => null,
                        'max' => null,
                    ),
                    'price' => array(
                        'min' => null,
                        'max' => null,
                    ),
                    'cost' => array(
                        'min' => $this->safe_number($minCosts, $quoteId),
                        'max' => null,
                    ),
                ),
                'info' => $market,
            );
        }
        return $result;
    }

    public function parse_ticker($ticker, $market = null) {
        // we don't set values that are not defined by the exchange
        $timestamp = $this->safe_integer($ticker, 'timestamp');
        $marketId = null;
        $baseId = $this->safe_string($ticker, 'coin_symbol');
        $quoteId = $this->safe_string($ticker, 'currency_symbol');
        if (($baseId !== null) && ($quoteId !== null)) {
            $marketId = $baseId . '_' . $quoteId;
        }
        $market = $this->safe_market($marketId, $market);
        $last = $this->safe_string($ticker, 'last');
        $change = $this->safe_string($ticker, 'change');
        $baseVolume = $this->safe_string_2($ticker, 'vol', 'vol24H');
        $percentage = $this->safe_string($ticker, 'percent');
        if ($percentage !== null) {
            $percentage = str_replace('%', '', $percentage);
        }
        return $this->safe_ticker(array(
            'symbol' => $market['symbol'],
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'high' => $this->safe_string($ticker, 'high'),
            'low' => $this->safe_string($ticker, 'low'),
            'bid' => $this->safe_string($ticker, 'buy'),
            'bidVolume' => $this->safe_string($ticker, 'buy_amount'),
            'ask' => $this->safe_string($ticker, 'sell'),
            'askVolume' => $this->safe_string($ticker, 'sell_amount'),
            'vwap' => null,
            'open' => null,
            'close' => $last,
            'last' => $last,
            'previousClose' => null,
            'change' => $change,
            'percentage' => $percentage,
            'average' => null,
            'baseVolume' => $baseVolume,
            'quoteVolume' => $this->safe_string($ticker, 'amount'),
            'info' => $ticker,
        ), $market);
    }

    public function fetch_ticker($symbol, $params = array ()) {
        /**
         * fetches a price ticker, a statistical calculation with the information calculated over the past 24 hours for a specific $market
         * @param {str} $symbol unified $symbol of the $market to fetch the ticker for
         * @param {dict} $params extra parameters specific to the bibox api endpoint
         * @return {dict} a {@link https://docs.ccxt.com/en/latest/manual.html#ticker-structure ticker structure}
         */
        yield $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'cmd' => 'ticker',
            'pair' => $market['id'],
        );
        $response = yield $this->publicGetMdata (array_merge($request, $params));
        return $this->parse_ticker($response['result'], $market);
    }

    public function fetch_tickers($symbols = null, $params = array ()) {
        /**
         * fetches price $tickers for multiple markets, statistical calculations with the information calculated over the past 24 hours each market
         * @param {[str]|null} $symbols unified $symbols of the markets to fetch the ticker for, all market $tickers are returned if not assigned
         * @param {dict} $params extra parameters specific to the bibox api endpoint
         * @return {dict} an array of {@link https://docs.ccxt.com/en/latest/manual.html#ticker-structure ticker structures}
         */
        $request = array(
            'cmd' => 'marketAll',
        );
        $response = yield $this->publicGetMdata (array_merge($request, $params));
        $tickers = $this->parse_tickers($response['result'], $symbols);
        $result = $this->index_by($tickers, 'symbol');
        return $this->filter_by_array($result, 'symbol', $symbols);
    }

    public function parse_trade($trade, $market = null) {
        $timestamp = $this->safe_integer_2($trade, 'time', 'createdAt');
        $side = $this->safe_integer_2($trade, 'side', 'order_side');
        $side = ($side === 1) ? 'buy' : 'sell';
        $marketId = $this->safe_string($trade, 'pair');
        if ($marketId === null) {
            $baseId = $this->safe_string($trade, 'coin_symbol');
            $quoteId = $this->safe_string($trade, 'currency_symbol');
            if (($baseId !== null) && ($quoteId !== null)) {
                $marketId = $baseId . '_' . $quoteId;
            }
        }
        $market = $this->safe_market($marketId, $market);
        $priceString = $this->safe_string($trade, 'price');
        $amountString = $this->safe_string($trade, 'amount');
        $fee = null;
        $feeCostString = $this->safe_string($trade, 'fee');
        if ($feeCostString !== null) {
            $feeCurrencyId = $this->safe_string($trade, 'fee_symbol');
            $feeCurrencyCode = $this->safe_currency_code($feeCurrencyId);
            $fee = array(
                'cost' => Precise::string_neg($feeCostString),
                'currency' => $feeCurrencyCode,
            );
        }
        $id = $this->safe_string($trade, 'id');
        return $this->safe_trade(array(
            'info' => $trade,
            'id' => $id,
            'order' => null, // Bibox does not have it (documented) yet
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'symbol' => $market['symbol'],
            'type' => 'limit',
            'takerOrMaker' => null,
            'side' => $side,
            'price' => $priceString,
            'amount' => $amountString,
            'cost' => null,
            'fee' => $fee,
        ), $market);
    }

    public function fetch_trades($symbol, $since = null, $limit = null, $params = array ()) {
        /**
         * get the list of most recent trades for a particular $symbol
         * @param {str} $symbol unified $symbol of the $market to fetch trades for
         * @param {int|null} $since timestamp in ms of the earliest trade to fetch
         * @param {int|null} $limit the maximum amount of trades to fetch
         * @param {dict} $params extra parameters specific to the bibox api endpoint
         * @return {[dict]} a list of ~@link https://docs.ccxt.com/en/latest/manual.html?#public-trades trade structures~
         */
        yield $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'cmd' => 'deals',
            'pair' => $market['id'],
        );
        if ($limit !== null) {
            $request['size'] = $limit; // default = 200
        }
        $response = yield $this->publicGetMdata (array_merge($request, $params));
        return $this->parse_trades($response['result'], $market, $since, $limit);
    }

    public function fetch_order_book($symbol, $limit = null, $params = array ()) {
        /**
         * fetches information on open orders with bid (buy) and ask (sell) prices, volumes and other data
         * @param {str} $symbol unified $symbol of the $market to fetch the order book for
         * @param {int|null} $limit the maximum amount of order book entries to return
         * @param {dict} $params extra parameters specific to the bibox api endpoint
         * @return {dict} A dictionary of {@link https://docs.ccxt.com/en/latest/manual.html#order-book-structure order book structures} indexed by $market symbols
         */
        yield $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'cmd' => 'depth',
            'pair' => $market['id'],
        );
        if ($limit !== null) {
            $request['size'] = $limit; // default = 200
        }
        $response = yield $this->publicGetMdata (array_merge($request, $params));
        return $this->parse_order_book($response['result'], $symbol, $this->safe_number($response['result'], 'update_time'), 'bids', 'asks', 'price', 'volume');
    }

    public function parse_ohlcv($ohlcv, $market = null) {
        //
        //     {
        //         "time":1591448220000,
        //         "open":"0.02507029",
        //         "high":"0.02507029",
        //         "low":"0.02506349",
        //         "close":"0.02506349",
        //         "vol":"5.92000000"
        //     }
        //
        return array(
            $this->safe_integer($ohlcv, 'time'),
            $this->safe_number($ohlcv, 'open'),
            $this->safe_number($ohlcv, 'high'),
            $this->safe_number($ohlcv, 'low'),
            $this->safe_number($ohlcv, 'close'),
            $this->safe_number($ohlcv, 'vol'),
        );
    }

    public function fetch_ohlcv($symbol, $timeframe = '1m', $since = null, $limit = 1000, $params = array ()) {
        /**
         * fetches historical candlestick data containing the open, high, low, and close price, and the volume of a $market
         * @param {str} $symbol unified $symbol of the $market to fetch OHLCV data for
         * @param {str} $timeframe the length of time each candle represents
         * @param {int|null} $since timestamp in ms of the earliest candle to fetch
         * @param {int|null} $limit the maximum amount of candles to fetch
         * @param {dict} $params extra parameters specific to the bibox api endpoint
         * @return {[[int]]} A list of candles ordered as timestamp, open, high, low, close, volume
         */
        yield $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'cmd' => 'kline',
            'pair' => $market['id'],
            'period' => $this->timeframes[$timeframe],
            'size' => $limit,
        );
        $response = yield $this->publicGetMdata (array_merge($request, $params));
        //
        //     {
        //         "result":array(
        //             array("time":1591448220000,"open":"0.02507029","high":"0.02507029","low":"0.02506349","close":"0.02506349","vol":"5.92000000"),
        //             array("time":1591448280000,"open":"0.02506449","high":"0.02506975","low":"0.02506108","close":"0.02506843","vol":"5.72000000"),
        //             array("time":1591448340000,"open":"0.02506698","high":"0.02506698","low":"0.02506452","close":"0.02506519","vol":"4.86000000"),
        //         ),
        //         "cmd":"kline",
        //         "ver":"1.1"
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        return $this->parse_ohlcvs($result, $market, $timeframe, $since, $limit);
    }

    public function fetch_currencies($params = array ()) {
        /**
         * fetches all available currencies on an exchange
         * @param {dict} $params extra parameters specific to the bibox api endpoint
         * @return {dict} an associative dictionary of currencies
         */
        if ($this->check_required_credentials(false)) {
            return yield $this->fetch_currencies_private($params);
        } else {
            return yield $this->fetch_currencies_public($params);
        }
    }

    public function fetch_currencies_public($params = array ()) {
        $request = array(
            'cmd' => 'currencies',
        );
        $response = yield $this->publicGetCdata (array_merge($request, $params));
        //
        // publicGetCdata
        //
        //     {
        //         "result":[
        //             {
        //                 "symbol":"BTC",
        //                 "name":"BTC",
        //                 "valid_decimals":8,
        //                 "original_decimals":8,
        //                 "is_erc20":0,
        //                 "enable_withdraw":1,
        //                 "enable_deposit":1,
        //                 "withdraw_min":0.005,
        //                 "describe_summary":"[array(\"lang\":\"zh-cn\",\"text\":\"Bitcoin 比特币的概念最初由中本聪在2009年提出，是点对点的基于 SHA-256 算法的一种P2P形式的数字货币，点对点的传输意味着一个去中心化的支付系统。\"),array(\"lang\":\"en-ww\",\"text\":\"Bitcoin is a digital asset and a payment system invented by Satoshi Nakamoto who published a related paper in 2008 and released it as open-source software in 2009. The system featured as peer-to-peer; users can transact directly without an intermediary.\")]"
        //             }
        //         ],
        //         "cmd":"currencies"
        //     }
        //
        $currencies = $this->safe_value($response, 'result', array());
        $result = array();
        for ($i = 0; $i < count($currencies); $i++) {
            $currency = $currencies[$i];
            $id = $this->safe_string($currency, 'symbol');
            $name = $this->safe_string($currency, 'name'); // contains hieroglyphs causing python ASCII bug
            $code = $this->safe_currency_code($id);
            $precision = $this->safe_integer($currency, 'valid_decimals');
            $deposit = $this->safe_value($currency, 'enable_deposit');
            $withdraw = $this->safe_value($currency, 'enable_withdraw');
            $active = ($deposit && $withdraw);
            $result[$code] = array(
                'id' => $id,
                'code' => $code,
                'info' => $currency,
                'name' => $name,
                'active' => $active,
                'deposit' => $deposit,
                'withdraw' => $withdraw,
                'fee' => null,
                'precision' => $precision,
                'limits' => array(
                    'amount' => array(
                        'min' => pow(10, -$precision),
                        'max' => null,
                    ),
                    'withdraw' => array(
                        'min' => $this->safe_number($currency, 'withdraw_min'),
                        'max' => null,
                    ),
                ),
            );
        }
        return $result;
    }

    public function fetch_currencies_private($params = array ()) {
        if (!$this->check_required_credentials(false)) {
            throw new AuthenticationError($this->id . " fetchCurrencies is an authenticated endpoint, therefore it requires 'apiKey' and 'secret' credentials. If you don't need $currency details, set exchange.has['fetchCurrencies'] = false before calling its methods.");
        }
        $request = array(
            'cmd' => 'transfer/coinList',
            'body' => array(),
        );
        $response = yield $this->privatePostTransfer (array_merge($request, $params));
        //
        //     {
        //         "result":[
        //             {
        //                 "result":[
        //                     {
        //                         "totalBalance":"14.60987476",
        //                         "balance":"14.60987476",
        //                         "freeze":"0.00000000",
        //                         "id":60,
        //                         "symbol":"USDT",
        //                         "icon_url":"/appimg/USDT_icon.png",
        //                         "describe_url":"[array(\"lang\":\"zh-cn\",\"link\":\"https://bibox.zendesk.com/hc/zh-cn/articles/115004798234\"),array(\"lang\":\"en-ww\",\"link\":\"https://bibox.zendesk.com/hc/en-us/articles/115004798234\")]",
        //                         "name":"USDT",
        //                         "enable_withdraw":1,
        //                         "enable_deposit":1,
        //                         "enable_transfer":1,
        //                         "confirm_count":2,
        //                         "is_erc20":1,
        //                         "forbid_info":null,
        //                         "describe_summary":"[array(\"lang\":\"zh-cn\",\"text\":\"USDT 是 Tether 公司推出的基于稳定价值货币美元（USD）的代币 Tether USD（简称USDT），1USDT=1美元，用户可以随时使用 USDT 与 USD 进行1:1的兑换。\"),array(\"lang\":\"en-ww\",\"text\":\"USDT is a cryptocurrency asset issued on the Bitcoin blockchain via the Omni Layer Protocol. Each USDT unit is backed by a U.S Dollar held in the reserves of the Tether Limited and can be redeemed through the Tether Platform.\")]",
        //                         "total_amount":4776930644,
        //                         "supply_amount":4642367414,
        //                         "price":"--",
        //                         "contract_father":"OMNI",
        //                         "supply_time":"--",
        //                         "comment":null,
        //                         "chain_type":"OMNI",
        //                         "general_name":"USDT",
        //                         "contract":"31",
        //                         "original_decimals":8,
        //                         "deposit_type":0,
        //                         "hasCobo":0,
        //                         "BTCValue":"0.00027116",
        //                         "CNYValue":"90.36087919",
        //                         "USDValue":"14.61090236",
        //                         "children":array(
        //                             array("type":"ERC20","symbol":"eUSDT","enable_deposit":1,"enable_withdraw":1,"confirm_count":13),
        //                             array("type":"TRC20","symbol":"tUSDT","enable_deposit":1,"enable_withdraw":1,"confirm_count":20),
        //                             array("type":"OMNI","symbol":"USDT","enable_deposit":1,"enable_withdraw":1,"confirm_count":2),
        //                             array("type":"HECO","symbol":"hUSDT","enable_deposit":1,"enable_withdraw":1,"confirm_count":12),
        //                             array("type":"BSC(BEP20)","symbol":"bUSDT","enable_deposit":1,"enable_withdraw":1,"confirm_count":5),
        //                             array("type":"HPB","symbol":"pUSDT","enable_deposit":1,"enable_withdraw":1,"confirm_count":20)
        //                         )
        //                     }
        //                 ],
        //                 "cmd":"transfer/coinList"
        //             }
        //         ]
        //     }
        //
        $outerResults = $this->safe_value($response, 'result');
        $firstResult = $this->safe_value($outerResults, 0, array());
        $currencies = $this->safe_value($firstResult, 'result', array());
        $result = array();
        for ($i = 0; $i < count($currencies); $i++) {
            $currency = $currencies[$i];
            $id = $this->safe_string($currency, 'symbol');
            $name = $currency['name']; // contains hieroglyphs causing python ASCII bug
            $code = $this->safe_currency_code($id);
            $precision = 8;
            $deposit = $this->safe_value($currency, 'enable_deposit');
            $withdraw = $this->safe_value($currency, 'enable_withdraw');
            $active = ($deposit && $withdraw);
            $result[$code] = array(
                'id' => $id,
                'code' => $code,
                'info' => $currency,
                'name' => $name,
                'active' => $active,
                'fee' => null,
                'precision' => $precision,
                'limits' => array(
                    'amount' => array(
                        'min' => pow(10, -$precision),
                        'max' => pow(10, $precision),
                    ),
                    'withdraw' => array(
                        'min' => null,
                        'max' => pow(10, $precision),
                    ),
                ),
            );
        }
        return $result;
    }

    public function parse_balance($response) {
        $outerResult = $this->safe_value($response, 'result');
        $firstResult = $this->safe_value($outerResult, 0, array());
        $innerResult = $this->safe_value($firstResult, 'result');
        $result = array( 'info' => $response );
        $assetsList = $this->safe_value($innerResult, 'assets_list', array());
        for ($i = 0; $i < count($assetsList); $i++) {
            $balance = $assetsList[$i];
            $currencyId = $this->safe_string($balance, 'coin_symbol');
            $code = $this->safe_currency_code($currencyId);
            $account = $this->account();
            $account['free'] = $this->safe_string($balance, 'balance');
            $account['used'] = $this->safe_string($balance, 'freeze');
            $result[$code] = $account;
        }
        return $this->safe_balance($result);
    }

    public function fetch_balance($params = array ()) {
        /**
         * query for balance and get the amount of funds available for trading or funds locked in orders
         * @param {dict} $params extra parameters specific to the bibox api endpoint
         * @return {dict} a ~@link https://docs.ccxt.com/en/latest/manual.html?#balance-structure balance structure~
         */
        yield $this->load_markets();
        $type = $this->safe_string($params, 'type', 'assets');
        $params = $this->omit($params, 'type');
        $request = array(
            'cmd' => 'transfer/' . $type, // assets, mainAssets
            'body' => array_merge(array(
                'select' => 1, // return full info
            ), $params),
        );
        $response = yield $this->privatePostTransfer ($request);
        //
        //     {
        //         "result":array(
        //             {
        //                 "result":array(
        //                     "total_btc":"0.00000298",
        //                     "total_cny":"0.99",
        //                     "total_usd":"0.16",
        //                     "assets_list":array(
        //                         array("coin_symbol":"BTC","BTCValue":"0.00000252","CNYValue":"0.84","USDValue":"0.14","balance":"0.00000252","freeze":"0.00000000"),
        //                         array("coin_symbol":"LTC","BTCValue":"0.00000023","CNYValue":"0.07","USDValue":"0.01","balance":"0.00006765","freeze":"0.00000000"),
        //                         array("coin_symbol":"USDT","BTCValue":"0.00000023","CNYValue":"0.08","USDValue":"0.01","balance":"0.01252100","freeze":"0.00000000")
        //                     )
        //                 ),
        //                 "cmd":"transfer/assets"
        //             }
        //         )
        //     }
        //
        return $this->parse_balance($response);
    }

    public function fetch_deposits($code = null, $since = null, $limit = null, $params = array ()) {
        yield $this->load_markets();
        if ($limit === null) {
            $limit = 100;
        }
        $request = array(
            'page' => 1,
            'size' => $limit,
        );
        $currency = null;
        if ($code !== null) {
            $currency = $this->currency($code);
            $request['symbol'] = $currency['id'];
        }
        $response = yield $this->privatePostTransfer (array(
            'cmd' => 'transfer/transferInList',
            'body' => array_merge($request, $params),
        ));
        //
        //     {
        //         "result":array(
        //             {
        //                 "result":{
        //                     "count":2,
        //                     "page":1,
        //                     "items":array(
        //                         array(
        //                             "coin_symbol":"ETH",                        // token
        //                             "to_address":"xxxxxxxxxxxxxxxxxxxxxxxxxx",  // address
        //                             "amount":"1.00000000",                      // amount
        //                             "confirmCount":"15",                        // the acknowledgment number
        //                             "createdAt":1540641511000,
        //                             "status":2                                 // status,  1-deposit is in process，2-deposit finished，3-deposit failed
        //                         ),
        //                         array(
        //                             "coin_symbol":"BIX",
        //                             "to_address":"xxxxxxxxxxxxxxxxxxxxxxxxxx",
        //                             "amount":"1.00000000",
        //                             "confirmCount":"15",
        //                             "createdAt":1540622460000,
        //                             "status":2
        //                         }
        //                     )
        //                 ),
        //                 "cmd":"transfer/transferInList"
        //             }
        //         )
        //     }
        //
        $outerResults = $this->safe_value($response, 'result');
        $firstResult = $this->safe_value($outerResults, 0, array());
        $innerResult = $this->safe_value($firstResult, 'result', array());
        $deposits = $this->safe_value($innerResult, 'items', array());
        for ($i = 0; $i < count($deposits); $i++) {
            $deposits[$i]['type'] = 'deposit';
        }
        return $this->parse_transactions($deposits, $currency, $since, $limit);
    }

    public function fetch_withdrawals($code = null, $since = null, $limit = null, $params = array ()) {
        yield $this->load_markets();
        if ($limit === null) {
            $limit = 100;
        }
        $request = array(
            'page' => 1,
            'size' => $limit,
        );
        $currency = null;
        if ($code !== null) {
            $currency = $this->currency($code);
            $request['symbol'] = $currency['id'];
        }
        $response = yield $this->privatePostTransfer (array(
            'cmd' => 'transfer/transferOutList',
            'body' => array_merge($request, $params),
        ));
        //
        //     {
        //         "result":array(
        //             {
        //                 "result":{
        //                     "count":1,
        //                     "page":1,
        //                     "items":array(
        //                         array(
        //                             "id":612867,
        //                             "coin_symbol":"ETH",
        //                             "chain_type":"ETH",
        //                             "to_address":"0xd41de7a88ab5fc59edc6669f54873576be95bff1",
        //                             "tx_id":"0xc60950596227af3f27c3a1b5911ea1c79bae53bdce67274e48a0ce87a5ef2df8",
        //                             "addr_remark":"binance",
        //                             "amount":"2.34550946",
        //                             "fee":"0.00600000",
        //                             "createdAt":1561339330000,
        //                             "memo":"",
        //                             "status":3
        //                         }
        //                     )
        //                 ),
        //                 "cmd":"transfer/transferOutList"
        //             }
        //         )
        //     }
        //
        $outerResults = $this->safe_value($response, 'result');
        $firstResult = $this->safe_value($outerResults, 0, array());
        $innerResult = $this->safe_value($firstResult, 'result', array());
        $withdrawals = $this->safe_value($innerResult, 'items', array());
        for ($i = 0; $i < count($withdrawals); $i++) {
            $withdrawals[$i]['type'] = 'withdrawal';
        }
        return $this->parse_transactions($withdrawals, $currency, $since, $limit);
    }

    public function parse_transaction($transaction, $currency = null) {
        //
        // fetchDeposits
        //
        //     {
        //         'id' => 1023291,
        //         'coin_symbol' => 'ETH',
        //         'to_address' => '0x7263....',
        //         'amount' => '0.49170000',
        //         'confirmCount' => '16',
        //         'createdAt' => 1553123867000,
        //         'status' => 2
        //     }
        //
        // fetchWithdrawals
        //
        //     {
        //         'id' => 521844,
        //         'coin_symbol' => 'ETH',
        //         'to_address' => '0xfd4e....',
        //         'addr_remark' => '',
        //         'amount' => '0.39452750',
        //         'fee' => '0.00600000',
        //         'createdAt' => 1553226906000,
        //         'memo' => '',
        //         'status' => 3
        //     }
        //
        // withdraw
        //
        //     {
        //         "result" => 228, // withdrawal $id
        //         "cmd":"transfer/transferOut"
        //     }
        //
        $id = $this->safe_string_2($transaction, 'id', 'result');
        $address = $this->safe_string($transaction, 'to_address');
        $currencyId = $this->safe_string($transaction, 'coin_symbol');
        $code = $this->safe_currency_code($currencyId, $currency);
        $timestamp = $this->safe_string($transaction, 'createdAt');
        $tag = $this->safe_string($transaction, 'addr_remark');
        $type = $this->safe_string($transaction, 'type');
        $status = $this->parse_transaction_status_by_type($this->safe_string($transaction, 'status'), $type);
        $amount = $this->safe_number($transaction, 'amount');
        $feeCost = $this->safe_number($transaction, 'fee');
        if ($type === 'deposit') {
            $feeCost = 0;
            $tag = null;
        }
        $fee = array(
            'cost' => $feeCost,
            'currency' => $code,
        );
        return array(
            'info' => $transaction,
            'id' => $id,
            'txid' => null,
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'network' => null,
            'address' => $address,
            'addressTo' => null,
            'addressFrom' => null,
            'tag' => $tag,
            'tagTo' => null,
            'tagFrom' => null,
            'type' => $type,
            'amount' => $amount,
            'currency' => $code,
            'status' => $status,
            'updated' => null,
            'fee' => $fee,
        );
    }

    public function parse_transaction_status_by_type($status, $type = null) {
        $statuses = array(
            'deposit' => array(
                '1' => 'pending',
                '2' => 'ok',
            ),
            'withdrawal' => array(
                '0' => 'pending',
                '3' => 'ok',
            ),
        );
        return $this->safe_string($this->safe_value($statuses, $type, array()), $status, $status);
    }

    public function create_order($symbol, $type, $side, $amount, $price = null, $params = array ()) {
        yield $this->load_markets();
        $market = $this->market($symbol);
        $orderType = ($type === 'limit') ? 2 : 1;
        $orderSide = ($side === 'buy') ? 1 : 2;
        $request = array(
            'cmd' => 'orderpending/trade',
            'body' => array_merge(array(
                'pair' => $market['id'],
                'account_type' => 0,
                'order_type' => $orderType,
                'order_side' => $orderSide,
                'pay_bix' => 0,
                'amount' => $amount,
                'price' => $price,
            ), $params),
        );
        $response = yield $this->privatePostOrderpending ($request);
        //
        //     {
        //         "result":array(
        //             {
        //                 "result" => "100055558128036", // order $id
        //                 "index" => 12345, // random index, specific one in a batch
        //                 "cmd":"orderpending/trade"
        //             }
        //         )
        //     }
        //
        $outerResults = $this->safe_value($response, 'result');
        $firstResult = $this->safe_value($outerResults, 0, array());
        $id = $this->safe_value($firstResult, 'result');
        return array(
            'info' => $response,
            'id' => $id,
        );
    }

    public function cancel_order($id, $symbol = null, $params = array ()) {
        $request = array(
            'cmd' => 'orderpending/cancelTrade',
            'body' => array_merge(array(
                'orders_id' => $id,
            ), $params),
        );
        $response = yield $this->privatePostOrderpending ($request);
        //
        //     {
        //         "result":array(
        //             {
        //                 "result":"OK", // only indicates if the server received the cancelling $request, and the cancelling result can be obtained from the order record
        //                 "index" => 12345, // random index, specific one in a batch
        //                 "cmd":"orderpending/cancelTrade"
        //             }
        //         )
        //     }
        //
        $outerResults = $this->safe_value($response, 'result');
        $firstResult = $this->safe_value($outerResults, 0, array());
        return $firstResult;
    }

    public function fetch_order($id, $symbol = null, $params = array ()) {
        yield $this->load_markets();
        $request = array(
            'cmd' => 'orderpending/order',
            'body' => array_merge(array(
                'id' => (string) $id,
                'account_type' => 0, // 0 = spot account
            ), $params),
        );
        $response = yield $this->privatePostOrderpending ($request);
        //
        //     {
        //         "result":array(
        //             {
        //                 "result":array(
        //                     "id":"100055558128036",
        //                     "createdAt" => 1512756997000,
        //                     "account_type":0,
        //                     "coin_symbol":"LTC",        // Trading Token
        //                     "currency_symbol":"BTC",    // Pricing Token
        //                     "order_side":2,             // Trading side 1-Buy, 2-Sell
        //                     "order_type":2,             // 2-limit $order
        //                     "price":"0.00900000",       // $order price
        //                     "amount":"1.00000000",      // $order amount
        //                     "money":"0.00900000",       // currency amount (price * amount)
        //                     "deal_amount":"0.00000000", // deal amount
        //                     "deal_percent":"0.00%",     // deal percentage
        //                     "unexecuted":"0.00000000",  // unexecuted amount
        //                     "status":3                  // Status, -1-fail, 0,1-to be dealt, 2-dealt partly, 3-dealt totally, 4- cancelled partly, 5-cancelled totally, 6-to be cancelled
        //                 ),
        //                 "cmd":"orderpending/order"
        //             }
        //         )
        //     }
        //
        $outerResults = $this->safe_value($response, 'result');
        $firstResult = $this->safe_value($outerResults, 0, array());
        $order = $this->safe_value($firstResult, 'result');
        if ($this->is_empty($order)) {
            throw new OrderNotFound($this->id . ' $order ' . $id . ' not found');
        }
        return $this->parse_order($order);
    }

    public function parse_order($order, $market = null) {
        $marketId = null;
        $baseId = $this->safe_string($order, 'coin_symbol');
        $quoteId = $this->safe_string($order, 'currency_symbol');
        if (($baseId !== null) && ($quoteId !== null)) {
            $marketId = $baseId . '_' . $quoteId;
        }
        $market = $this->safe_market($marketId, $market);
        $rawType = $this->safe_string($order, 'order_type');
        $type = ($rawType === '1') ? 'market' : 'limit';
        $timestamp = $this->safe_integer($order, 'createdAt');
        $price = $this->safe_string($order, 'price');
        $average = $this->safe_string($order, 'deal_price');
        $filled = $this->safe_string($order, 'deal_amount');
        $amount = $this->safe_string($order, 'amount');
        $cost = $this->safe_string_2($order, 'deal_money', 'money');
        $rawSide = $this->safe_string($order, 'order_side');
        $side = ($rawSide === '1') ? 'buy' : 'sell';
        $status = $this->parse_order_status($this->safe_string($order, 'status'));
        $id = $this->safe_string($order, 'id');
        $feeCost = $this->safe_string($order, 'fee');
        $fee = null;
        if ($feeCost !== null) {
            $fee = array(
                'cost' => $feeCost,
                'currency' => null,
            );
        }
        return $this->safe_order(array(
            'info' => $order,
            'id' => $id,
            'clientOrderId' => null,
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'lastTradeTimestamp' => null,
            'symbol' => $market['symbol'],
            'type' => $type,
            'timeInForce' => null,
            'postOnly' => null,
            'side' => $side,
            'price' => $price,
            'stopPrice' => null,
            'amount' => $amount,
            'cost' => $cost,
            'average' => $average,
            'filled' => $filled,
            'remaining' => null,
            'status' => $status,
            'fee' => $fee,
            'trades' => null,
        ), $market);
    }

    public function parse_order_status($status) {
        $statuses = array(
            // original comments from bibox:
            '1' => 'open', // pending
            '2' => 'open', // part completed
            '3' => 'closed', // completed
            '4' => 'canceled', // part canceled
            '5' => 'canceled', // canceled
            '6' => 'canceled', // canceling
        );
        return $this->safe_string($statuses, $status, $status);
    }

    public function fetch_open_orders($symbol = null, $since = null, $limit = null, $params = array ()) {
        yield $this->load_markets();
        $market = null;
        $pair = null;
        if ($symbol !== null) {
            $market = $this->market($symbol);
            $pair = $market['id'];
        }
        $size = $limit ? $limit : 200;
        $request = array(
            'cmd' => 'orderpending/orderPendingList',
            'body' => array_merge(array(
                'pair' => $pair,
                'account_type' => 0, // 0 - regular, 1 - margin
                'page' => 1,
                'size' => $size,
            ), $params),
        );
        $response = yield $this->privatePostOrderpending ($request);
        //
        //     {
        //         "result":array(
        //             {
        //                 "result":{
        //                     "count":1,
        //                     "page":1,
        //                     "items":array(
        //                         array(
        //                             "id":"100055558128036",
        //                             "createdAt" => 1512756997000,
        //                             "account_type":0,
        //                             "coin_symbol":"LTC",        // Trading Token
        //                             "currency_symbol":"BTC",    // Pricing Token
        //                             "order_side":2,             // Trading side 1-Buy, 2-Sell
        //                             "order_type":2,             // 2-$limit order
        //                             "price":"0.00900000",       // order price
        //                             "amount":"1.00000000",      // order amount
        //                             "money":"0.00900000",       // currency amount (price * amount)
        //                             "deal_amount":"0.00000000", // deal amount
        //                             "deal_percent":"0.00%",     // deal percentage
        //                             "unexecuted":"0.00000000",  // unexecuted amount
        //                             "status":1                  // Status,-1-fail, 0,1-to be dealt, 2-dealt partly, 3-dealt totally, 4- cancelled partly, 5-cancelled totally, 6-to be cancelled
        //                         }
        //                     )
        //                 ),
        //                 "cmd":"orderpending/orderPendingList"
        //             }
        //         )
        //     }
        //
        $outerResults = $this->safe_value($response, 'result');
        $firstResult = $this->safe_value($outerResults, 0, array());
        $innerResult = $this->safe_value($firstResult, 'result', array());
        $orders = $this->safe_value($innerResult, 'items', array());
        return $this->parse_orders($orders, $market, $since, $limit);
    }

    public function fetch_closed_orders($symbol = null, $since = null, $limit = 200, $params = array ()) {
        if ($symbol === null) {
            throw new ArgumentsRequired($this->id . ' fetchClosedOrders() requires a `$symbol` argument');
        }
        yield $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'cmd' => 'orderpending/pendingHistoryList',
            'body' => array_merge(array(
                'pair' => $market['id'],
                'account_type' => 0, // 0 - regular, 1 - margin
                'page' => 1,
                'size' => $limit,
            ), $params),
        );
        $response = yield $this->privatePostOrderpending ($request);
        //
        //     {
        //         "result":array(
        //             {
        //                 "result":{
        //                     "count":1,
        //                     "page":1,
        //                     "items":array(
        //                         array(
        //                             "id":"100055558128036",
        //                             "createdAt" => 1512756997000,
        //                             "account_type":0,
        //                             "coin_symbol":"LTC",        // Trading Token
        //                             "currency_symbol":"BTC",    // Pricing Token
        //                             "order_side":2,             // Trading side 1-Buy, 2-Sell
        //                             "order_type":2,             // 2-$limit order
        //                             "price":"0.00900000",       // order price
        //                             "amount":"1.00000000",      // order amount
        //                             "money":"0.00900000",       // currency amount (price * amount)
        //                             "deal_amount":"0.00000000", // deal amount
        //                             "deal_percent":"0.00%",     // deal percentage
        //                             "unexecuted":"0.00000000",  // unexecuted amount
        //                             "status":3                  // Status,-1-fail, 0,1-to be dealt, 2-dealt partly, 3-dealt totally, 4- cancelled partly, 5-cancelled totally, 6-to be cancelled
        //                         }
        //                     )
        //                 ),
        //                 "cmd":"orderpending/pendingHistoryList"
        //             }
        //         )
        //     }
        //
        $outerResults = $this->safe_value($response, 'result');
        $firstResult = $this->safe_value($outerResults, 0, array());
        $innerResult = $this->safe_value($firstResult, 'result', array());
        $orders = $this->safe_value($innerResult, 'items', array());
        return $this->parse_orders($orders, $market, $since, $limit);
    }

    public function fetch_my_trades($symbol = null, $since = null, $limit = null, $params = array ()) {
        if ($symbol === null) {
            throw new ArgumentsRequired($this->id . ' fetchMyTrades() requires a `$symbol` argument');
        }
        yield $this->load_markets();
        $market = $this->market($symbol);
        $size = $limit ? $limit : 200;
        $request = array(
            'cmd' => 'orderpending/orderHistoryList',
            'body' => array_merge(array(
                'pair' => $market['id'],
                'account_type' => 0, // 0 - regular, 1 - margin
                'page' => 1,
                'size' => $size,
                'coin_symbol' => $market['baseId'],
                'currency_symbol' => $market['quoteId'],
            ), $params),
        );
        $response = yield $this->privatePostOrderpending ($request);
        //
        //     {
        //         "result":array(
        //             {
        //                 "result":{
        //                     "count":1,
        //                     "page":1,
        //                     "items":array(
        //                         array(
        //                             "id":"100055558128033",
        //                             "createdAt" => 1512756997000,
        //                             "account_type":0,
        //                             "coin_symbol":"LTC",
        //                             "currency_symbol":"BTC",
        //                             "order_side":2,
        //                             "order_type":2,
        //                             "price":"0.00886500",
        //                             "amount":"1.00000000",
        //                             "money":"0.00886500",
        //                             "fee":0
        //                         }
        //                     )
        //                 ),
        //                 "cmd":"orderpending/orderHistoryList"
        //             }
        //         )
        //     }
        //
        $outerResults = $this->safe_value($response, 'result');
        $firstResult = $this->safe_value($outerResults, 0, array());
        $innerResult = $this->safe_value($firstResult, 'result', array());
        $trades = $this->safe_value($innerResult, 'items', array());
        return $this->parse_trades($trades, $market, $since, $limit);
    }

    public function fetch_deposit_address($code, $params = array ()) {
        yield $this->load_markets();
        $currency = $this->currency($code);
        $request = array(
            'cmd' => 'transfer/transferIn',
            'body' => array_merge(array(
                'coin_symbol' => $currency['id'],
            ), $params),
        );
        $response = yield $this->privatePostTransfer ($request);
        //
        //     {
        //         "result":array(
        //             {
        //                 "result":"3Jx6RZ9TNMsAoy9NUzBwZf68QBppDruSKW",
        //                 "cmd":"transfer/transferIn"
        //             }
        //         )
        //     }
        //
        //     {
        //         "result":array(
        //             {
        //                 "result":"array(\"account\":\"PERSONALLY OMITTED\",\"memo\":\"PERSONALLY OMITTED\")",
        //                 "cmd":"transfer/transferIn"
        //             }
        //         )
        //     }
        //
        $outerResults = $this->safe_value($response, 'result');
        $firstResult = $this->safe_value($outerResults, 0, array());
        $innerResult = $this->safe_value($firstResult, 'result');
        $address = $innerResult;
        $tag = null;
        if ($this->is_json_encoded_object($innerResult)) {
            $parsed = json_decode($innerResult, $as_associative_array = true);
            $address = $this->safe_string($parsed, 'account');
            $tag = $this->safe_string($parsed, 'memo');
        }
        return array(
            'currency' => $code,
            'address' => $address,
            'tag' => $tag,
            'network' => null,
            'info' => $response,
        );
    }

    public function withdraw($code, $amount, $address, $tag = null, $params = array ()) {
        list($tag, $params) = $this->handle_withdraw_tag_and_params($tag, $params);
        $this->check_address($address);
        yield $this->load_markets();
        $currency = $this->currency($code);
        if ($this->password === null) {
            if (!(is_array($params) && array_key_exists('trade_pwd', $params))) {
                throw new ExchangeError($this->id . ' withdraw() requires $this->password set on the exchange instance or a trade_pwd parameter');
            }
        }
        if (!(is_array($params) && array_key_exists('totp_code', $params))) {
            throw new ExchangeError($this->id . ' withdraw() requires a totp_code parameter for 2FA authentication');
        }
        $request = array(
            'trade_pwd' => $this->password,
            'coin_symbol' => $currency['id'],
            'amount' => $amount,
            'addr' => $address,
        );
        if ($tag !== null) {
            $request['address_remark'] = $tag;
        }
        $response = yield $this->privatePostTransfer (array(
            'cmd' => 'transfer/transferOut',
            'body' => array_merge($request, $params),
        ));
        //
        //     {
        //         "result":array(
        //             {
        //                 "result" => 228, // withdrawal id
        //                 "cmd":"transfer/transferOut"
        //             }
        //         )
        //     }
        //
        $outerResults = $this->safe_value($response, 'result');
        $firstResult = $this->safe_value($outerResults, 0, array());
        return $this->parse_transaction($firstResult, $currency);
    }

    public function fetch_transaction_fees($codes = null, $params = array ()) {
        // by default it will try load withdrawal fees of all currencies (with separate requests)
        // however if you define $codes = array( 'ETH', 'BTC' ) in args it will only load those
        yield $this->load_markets();
        $withdrawFees = array();
        $info = array();
        if ($codes === null) {
            $codes = is_array($this->currencies) ? array_keys($this->currencies) : array();
        }
        for ($i = 0; $i < count($codes); $i++) {
            $code = $codes[$i];
            $currency = $this->currency($code);
            $request = array(
                'cmd' => 'transfer/coinConfig',
                'body' => array_merge(array(
                    'coin_symbol' => $currency['id'],
                ), $params),
            );
            $response = yield $this->privatePostTransfer ($request);
            //     {
            //         "result":array(
            //             {
            //                 "result":array(
            //                     {
            //                         "coin_symbol":"ETH",
            //                         "is_active":1,
            //                         "original_decimals":18,
            //                         "enable_deposit":1,
            //                         "enable_withdraw":1,
            //                         "withdraw_fee":0.008,
            //                         "withdraw_min":0.05,
            //                         "deposit_avg_spent":173700,
            //                         "withdraw_avg_spent":322600
            //                     }
            //                 ),
            //                 "cmd":"transfer/coinConfig"
            //             }
            //         )
            //     }
            //
            $outerResults = $this->safe_value($response, 'result', array());
            $firstOuterResult = $this->safe_value($outerResults, 0, array());
            $innerResults = $this->safe_value($firstOuterResult, 'result', array());
            $firstInnerResult = $this->safe_value($innerResults, 0, array());
            $info[$code] = $firstInnerResult;
            $withdrawFees[$code] = $this->safe_number($firstInnerResult, 'withdraw_fee');
        }
        return array(
            'info' => $info,
            'withdraw' => $withdrawFees,
            'deposit' => array(),
        );
    }

    public function sign($path, $api = 'public', $method = 'GET', $params = array (), $headers = null, $body = null) {
        $url = $this->implode_hostname($this->urls['api']) . '/' . $this->version . '/' . $path;
        $cmds = $this->json(array( $params ));
        if ($api === 'public') {
            if ($method !== 'GET') {
                $body = array( 'cmds' => $cmds );
            } elseif ($params) {
                $url .= '?' . $this->urlencode($params);
            }
        } elseif ($api === 'v2private') {
            $this->check_required_credentials();
            $url = $this->implode_hostname($this->urls['api']) . '/v2/' . $path;
            $json_params = $this->json($params);
            $body = array(
                'body' => $json_params,
                'apikey' => $this->apiKey,
                'sign' => $this->hmac($this->encode($json_params), $this->encode($this->secret), 'md5'),
            );
        } else {
            $this->check_required_credentials();
            $body = array(
                'cmds' => $cmds,
                'apikey' => $this->apiKey,
                'sign' => $this->hmac($this->encode($cmds), $this->encode($this->secret), 'md5'),
            );
        }
        if ($body !== null) {
            $body = $this->json($body, array( 'convertArraysToObjects' => true ));
        }
        $headers = array( 'Content-Type' => 'application/json' );
        return array( 'url' => $url, 'method' => $method, 'body' => $body, 'headers' => $headers );
    }

    public function handle_errors($code, $reason, $url, $method, $headers, $body, $response, $requestHeaders, $requestBody) {
        if ($response === null) {
            return;
        }
        if (is_array($response) && array_key_exists('error', $response)) {
            if (is_array($response['error']) && array_key_exists('code', $response['error'])) {
                $code = $this->safe_string($response['error'], 'code');
                $feedback = $this->id . ' ' . $body;
                $this->throw_exactly_matched_exception($this->exceptions, $code, $feedback);
                throw new ExchangeError($feedback);
            }
            throw new ExchangeError($this->id . ' ' . $body);
        }
        if (!(is_array($response) && array_key_exists('result', $response))) {
            throw new ExchangeError($this->id . ' ' . $body);
        }
    }
}
