<?php

namespace ccxt\async;

// PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
// https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

use Exception; // a common import
use \ccxt\ExchangeError;
use \ccxt\ArgumentsRequired;
use \ccxt\Precise;

class bitbns extends Exchange {

    public function describe() {
        return $this->deep_extend(parent::describe (), array(
            'id' => 'bitbns',
            'name' => 'Bitbns',
            'countries' => array( 'IN' ), // India
            'rateLimit' => 1000,
            'certified' => false,
            'pro' => false,
            'version' => 'v2',
            // new metainfo interface
            'has' => array(
                'CORS' => null,
                'spot' => true,
                'margin' => null, // has but unimplemented
                'swap' => false,
                'future' => false,
                'option' => null, // coming soon
                'cancelOrder' => true,
                'createOrder' => true,
                'fetchBalance' => true,
                'fetchDepositAddress' => true,
                'fetchDeposits' => true,
                'fetchFundingHistory' => false,
                'fetchFundingRate' => false,
                'fetchFundingRateHistory' => false,
                'fetchFundingRates' => false,
                'fetchIndexOHLCV' => false,
                'fetchMarkets' => true,
                'fetchMarkOHLCV' => false,
                'fetchMyTrades' => true,
                'fetchOHLCV' => null,
                'fetchOpenOrders' => true,
                'fetchOrder' => true,
                'fetchOrderBook' => true,
                'fetchPremiumIndexOHLCV' => false,
                'fetchStatus' => true,
                'fetchTicker' => 'emulated',
                'fetchTickers' => true,
                'fetchTrades' => true,
                'fetchTradingFee' => false,
                'fetchTradingFees' => false,
                'fetchTransfer' => false,
                'fetchTransfers' => false,
                'fetchWithdrawal' => false,
                'fetchWithdrawals' => true,
                'transfer' => false,
                'withdraw' => false,
            ),
            'timeframes' => array(
            ),
            'urls' => array(
                'logo' => 'https://user-images.githubusercontent.com/1294454/117201933-e7a6e780-adf5-11eb-9d80-98fc2a21c3d6.jpg',
                'api' => array(
                    'www' => 'https://bitbns.com',
                    'v1' => 'https://api.bitbns.com/api/trade/v1',
                    'v2' => 'https://api.bitbns.com/api/trade/v2',
                ),
                'www' => 'https://bitbns.com',
                'referral' => 'https://ref.bitbns.com/1090961',
                'doc' => array(
                    'https://bitbns.com/trade/#/api-trading/',
                ),
                'fees' => 'https://bitbns.com/fees',
            ),
            'api' => array(
                'www' => array(
                    'get' => array(
                        'order/fetchMarkets',
                        'order/fetchTickers',
                        'order/fetchOrderbook',
                        'order/getTickerWithVolume',
                        'exchangeData/ohlc', // ?coin=${coin_name}&page=${page}
                        'exchangeData/orderBook',
                        'exchangeData/tradedetails',
                    ),
                ),
                'v1' => array(
                    'get' => array(
                        'platform/status',
                        'tickers',
                        'orderbook/sell/{symbol}',
                        'orderbook/buy/{symbol}',
                    ),
                    'post' => array(
                        'currentCoinBalance/EVERYTHING',
                        'getApiUsageStatus/USAGE',
                        'getOrderSocketToken/USAGE',
                        'currentCoinBalance/{symbol}',
                        'orderStatus/{symbol}',
                        'depositHistory/{symbol}',
                        'withdrawHistory/{symbol}',
                        'withdrawHistoryAll/{symbol}',
                        'depositHistoryAll/{symbol}',
                        'listOpenOrders/{symbol}',
                        'listOpenStopOrders/{symbol}',
                        'getCoinAddress/{symbol}',
                        'placeSellOrder/{symbol}',
                        'placeBuyOrder/{symbol}',
                        'buyStopLoss/{symbol}',
                        'sellStopLoss/{symbol}',
                        'placeSellOrder/{symbol}',
                        'cancelOrder/{symbol}',
                        'cancelStopLossOrder/{symbol}',
                        'listExecutedOrders/{symbol}',
                        'placeMarketOrder/{symbol}',
                        'placeMarketOrderQnty/{symbol}',
                    ),
                ),
                'v2' => array(
                    'post' => array(
                        'orders',
                        'cancel',
                        'getordersnew',
                        'marginOrders',
                    ),
                ),
            ),
            'fees' => array(
                'trading' => array(
                    'feeSide' => 'quote',
                    'tierBased' => false,
                    'percentage' => true,
                    'taker' => $this->parse_number('0.0025'),
                    'maker' => $this->parse_number('0.0025'),
                ),
            ),
            'exceptions' => array(
                'exact' => array(
                    '400' => '\\ccxt\\BadRequest', // array("msg":"Invalid Request","status":-1,"code":400)
                    '409' => '\\ccxt\\BadSymbol', // array("data":"","status":0,"error":"coin name not supplied or not yet supported","code":409)
                    '416' => '\\ccxt\\InsufficientFunds', // array("data":"Oops ! Not sufficient currency to sell","status":0,"error":null,"code":416)
                    '417' => '\\ccxt\\OrderNotFound', // array("data":array(),"status":0,"error":"Nothing to show","code":417)
                ),
                'broad' => array(),
            ),
        ));
    }

    public function fetch_status($params = array ()) {
        /**
         * the latest known information on the availability of the exchange API
         * @param {dict} $params extra parameters specific to the bitbns api endpoint
         * @return {dict} a {@link https://docs.ccxt.com/en/latest/manual.html#exchange-status-structure status structure}
         */
        $response = yield $this->v1GetPlatformStatus ($params);
        //
        //     {
        //         "data":array(
        //             "BTC":array("status":1),
        //             "ETH":array("status":1),
        //             "XRP":array("status":1),
        //         ),
        //         "status":1,
        //         "error":null,
        //         "code":200
        //     }
        //
        $statusRaw = $this->safe_string($response, 'status');
        return array(
            'status' => $this->safe_string(array( '1' => 'ok' ), $statusRaw, $statusRaw),
            'updated' => null,
            'eta' => null,
            'url' => null,
            'info' => $response,
        );
    }

    public function fetch_markets($params = array ()) {
        /**
         * retrieves data on all markets for bitbns
         * @param {dict} $params extra parameters specific to the exchange api endpoint
         * @return {[dict]} an array of objects representing $market data
         */
        $response = yield $this->wwwGetOrderFetchMarkets ($params);
        //
        //     array(
        //         array(
        //             "id":"BTC",
        //             "symbol":"BTC/INR",
        //             "base":"BTC",
        //             "quote":"INR",
        //             "baseId":"BTC",
        //             "quoteId":"",
        //             "active":true,
        //             "limits":array(
        //                 "amount":array("min":"0.00017376","max":20),
        //                 "price":array("min":2762353.2359999996,"max":6445490.883999999),
        //                 "cost":array("min":800,"max":128909817.67999998)
        //             ),
        //             "precision":array(
        //                 "amount":8,
        //                 "price":2
        //             ),
        //             "info":array()
        //         ),
        //     )
        //
        $result = array();
        for ($i = 0; $i < count($response); $i++) {
            $market = $response[$i];
            $id = $this->safe_string($market, 'id');
            $baseId = $this->safe_string($market, 'base');
            $quoteId = $this->safe_string($market, 'quote');
            $base = $this->safe_currency_code($baseId);
            $quote = $this->safe_currency_code($quoteId);
            $marketPrecision = $this->safe_value($market, 'precision', array());
            $marketLimits = $this->safe_value($market, 'limits', array());
            $amountLimits = $this->safe_value($marketLimits, 'amount', array());
            $priceLimits = $this->safe_value($marketLimits, 'price', array());
            $costLimits = $this->safe_value($marketLimits, 'cost', array());
            $usdt = ($quoteId === 'USDT');
            // INR markets don't need a _INR prefix
            $uppercaseId = $usdt ? ($baseId . '_' . $quoteId) : $baseId;
            $result[] = array(
                'id' => $id,
                'uppercaseId' => $uppercaseId,
                'symbol' => $base . '/' . $quote,
                'base' => $base,
                'quote' => $quote,
                'settle' => null,
                'baseId' => $baseId,
                'quoteId' => $quoteId,
                'settleId' => null,
                'type' => 'spot',
                'spot' => true,
                'margin' => false,
                'swap' => false,
                'future' => false,
                'option' => false,
                'active' => null,
                'contract' => false,
                'linear' => null,
                'inverse' => null,
                'contractSize' => null,
                'expiry' => null,
                'expiryDatetime' => null,
                'strike' => null,
                'optionType' => null,
                'precision' => array(
                    'amount' => $this->safe_integer($marketPrecision, 'amount'),
                    'price' => $this->safe_integer($marketPrecision, 'price'),
                ),
                'limits' => array(
                    'leverage' => array(
                        'min' => null,
                        'max' => null,
                    ),
                    'amount' => array(
                        'min' => $this->safe_number($amountLimits, 'min'),
                        'max' => $this->safe_number($amountLimits, 'max'),
                    ),
                    'price' => array(
                        'min' => $this->safe_number($priceLimits, 'min'),
                        'max' => $this->safe_number($priceLimits, 'max'),
                    ),
                    'cost' => array(
                        'min' => $this->safe_number($costLimits, 'min'),
                        'max' => $this->safe_number($costLimits, 'max'),
                    ),
                ),
                'info' => $market,
            );
        }
        return $result;
    }

    public function fetch_order_book($symbol, $limit = null, $params = array ()) {
        /**
         * fetches information on open orders with bid (buy) and ask (sell) prices, volumes and other data
         * @param {str} $symbol unified $symbol of the $market to fetch the order book for
         * @param {int|null} $limit the maximum amount of order book entries to return
         * @param {dict} $params extra parameters specific to the bitbns api endpoint
         * @return {dict} A dictionary of {@link https://docs.ccxt.com/en/latest/manual.html#order-book-structure order book structures} indexed by $market symbols
         */
        yield $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'symbol' => $market['id'],
        );
        if ($limit !== null) {
            $request['limit'] = $limit; // default 100, max 5000, see https://github.com/binance-exchange/binance-official-api-docs/blob/master/rest-api.md#order-book
        }
        $response = yield $this->wwwGetOrderFetchOrderbook (array_merge($request, $params));
        //
        //     {
        //         "bids":[
        //             [49352.04,0.843948],
        //             [49352.03,0.742048],
        //             [49349.78,0.686239],
        //         ],
        //         "asks":[
        //             [49443.59,0.065137],
        //             [49444.63,0.098211],
        //             [49449.01,0.066309],
        //         ],
        //         "timestamp":1619172786577,
        //         "datetime":"2021-04-23T10:13:06.577Z",
        //         "nonce":""
        //     }
        //
        $timestamp = $this->safe_integer($response, 'timestamp');
        return $this->parse_order_book($response, $symbol, $timestamp);
    }

    public function parse_ticker($ticker, $market = null) {
        //
        //     {
        //         "symbol":"BTC/INR",
        //         "info":array(
        //             "highest_buy_bid":4368494.31,
        //             "lowest_sell_bid":4374835.09,
        //             "last_traded_price":4374835.09,
        //             "yes_price":4531016.27,
        //             "volume":array("max":"4569119.23","min":"4254552.13","volume":62.17722344)
        //         ),
        //         "timestamp":1619100020845,
        //         "datetime":1619100020845,
        //         "high":"4569119.23",
        //         "low":"4254552.13",
        //         "bid":4368494.31,
        //         "bidVolume":"",
        //         "ask":4374835.09,
        //         "askVolume":"",
        //         "vwap":"",
        //         "open":4531016.27,
        //         "close":4374835.09,
        //         "last":4374835.09,
        //         "baseVolume":62.17722344,
        //         "quoteVolume":"",
        //         "previousClose":"",
        //         "change":-156181.1799999997,
        //         "percentage":-3.446934874943623,
        //         "average":4452925.68
        //     }
        //
        $timestamp = $this->safe_integer($ticker, 'timestamp');
        $marketId = $this->safe_string($ticker, 'symbol');
        $symbol = $this->safe_symbol($marketId, $market);
        $last = $this->safe_string($ticker, 'last');
        return $this->safe_ticker(array(
            'symbol' => $symbol,
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'high' => $this->safe_string($ticker, 'high'),
            'low' => $this->safe_string($ticker, 'low'),
            'bid' => $this->safe_string($ticker, 'bid'),
            'bidVolume' => $this->safe_string($ticker, 'bidVolume'),
            'ask' => $this->safe_string($ticker, 'ask'),
            'askVolume' => $this->safe_string($ticker, 'askVolume'),
            'vwap' => $this->safe_string($ticker, 'vwap'),
            'open' => $this->safe_string($ticker, 'open'),
            'close' => $last,
            'last' => $last,
            'previousClose' => $this->safe_string($ticker, 'previousClose'), // previous day close
            'change' => $this->safe_string($ticker, 'change'),
            'percentage' => $this->safe_string($ticker, 'percentage'),
            'average' => $this->safe_string($ticker, 'average'),
            'baseVolume' => $this->safe_string($ticker, 'baseVolume'),
            'quoteVolume' => $this->safe_string($ticker, 'quoteVolume'),
            'info' => $ticker,
        ), $market);
    }

    public function fetch_tickers($symbols = null, $params = array ()) {
        /**
         * fetches price tickers for multiple markets, statistical calculations with the information calculated over the past 24 hours each market
         * @param {[str]|null} $symbols unified $symbols of the markets to fetch the ticker for, all market tickers are returned if not assigned
         * @param {dict} $params extra parameters specific to the bitbns api endpoint
         * @return {dict} an array of {@link https://docs.ccxt.com/en/latest/manual.html#ticker-structure ticker structures}
         */
        yield $this->load_markets();
        $response = yield $this->wwwGetOrderFetchTickers ($params);
        //
        //     {
        //         "BTC/INR":{
        //             "symbol":"BTC/INR",
        //             "info":array(
        //                 "highest_buy_bid":4368494.31,
        //                 "lowest_sell_bid":4374835.09,
        //                 "last_traded_price":4374835.09,
        //                 "yes_price":4531016.27,
        //                 "volume":array("max":"4569119.23","min":"4254552.13","volume":62.17722344)
        //             ),
        //             "timestamp":1619100020845,
        //             "datetime":1619100020845,
        //             "high":"4569119.23",
        //             "low":"4254552.13",
        //             "bid":4368494.31,
        //             "bidVolume":"",
        //             "ask":4374835.09,
        //             "askVolume":"",
        //             "vwap":"",
        //             "open":4531016.27,
        //             "close":4374835.09,
        //             "last":4374835.09,
        //             "baseVolume":62.17722344,
        //             "quoteVolume":"",
        //             "previousClose":"",
        //             "change":-156181.1799999997,
        //             "percentage":-3.446934874943623,
        //             "average":4452925.68
        //         }
        //     }
        //
        return $this->parse_tickers($response, $symbols);
    }

    public function parse_balance($response) {
        $timestamp = null;
        $result = array(
            'info' => $response,
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
        );
        $data = $this->safe_value($response, 'data', array());
        $keys = is_array($data) ? array_keys($data) : array();
        for ($i = 0; $i < count($keys); $i++) {
            $key = $keys[$i];
            $parts = explode('availableorder', $key);
            $numParts = is_array($parts) ? count($parts) : 0;
            if ($numParts > 1) {
                $currencyId = $this->safe_string($parts, 1);
                // note that "Money" stands for INR - the only fiat in bitbns
                if ($currencyId === 'Money') {
                    $currencyId = 'INR';
                }
                $code = $this->safe_currency_code($currencyId);
                $account = $this->account();
                $account['free'] = $this->safe_string($data, $key);
                $account['used'] = $this->safe_string($data, 'inorder' . $currencyId);
                $result[$code] = $account;
            }
        }
        return $this->safe_balance($result);
    }

    public function fetch_balance($params = array ()) {
        /**
         * query for balance and get the amount of funds available for trading or funds locked in orders
         * @param {dict} $params extra parameters specific to the bitbns api endpoint
         * @return {dict} a ~@link https://docs.ccxt.com/en/latest/manual.html?#balance-structure balance structure~
         */
        yield $this->load_markets();
        $response = yield $this->v1PostCurrentCoinBalanceEVERYTHING ($params);
        //
        //     {
        //         "data":array(
        //             "availableorderMoney":12.34, // INR
        //             "availableorderBTC":0,
        //             "availableorderXRP":0,
        //             "inorderMoney":0, // INR
        //             "inorderBTC":0,
        //             "inorderXRP":0,
        //             "inorderNEO":0,
        //         ),
        //         "status":1,
        //         "error":null,
        //         "code":200
        //     }
        //
        // note that "Money" stands for INR - the only fiat in bitbns
        return $this->parse_balance($response);
    }

    public function parse_order_status($status) {
        $statuses = array(
            '0' => 'open',
            // 'PARTIALLY_FILLED' => 'open',
            // 'FILLED' => 'closed',
            // 'CANCELED' => 'canceled',
            // 'PENDING_CANCEL' => 'canceling', // currently unused
            // 'REJECTED' => 'rejected',
            // 'EXPIRED' => 'expired',
        );
        return $this->safe_string($statuses, $status, $status);
    }

    public function parse_order($order, $market = null) {
        //
        // createOrder
        //
        //     {
        //         "data":"Successfully placed bid to purchase currency",
        //         "status":1,
        //         "error":null,
        //         "id":5424475,
        //         "code":200
        //     }
        //
        // fetchOrder
        //
        //     {
        //         "entry_id":5424475,
        //         "btc":0.01,
        //         "rate":2000,
        //         "time":"2021-04-25T17:05:42.000Z",
        //         "type":0,
        //         "status":0,
        //         "total":0.01,
        //         "avg_cost":null,
        //         "side":"BUY",
        //         "amount":0.01,
        //         "remaining":0.01,
        //         "filled":0,
        //         "cost":null,
        //         "fee":0.05
        //     }
        //
        // fetchOpenOrders
        //
        //     {
        //         "entry_id":5424475,
        //         "btc":0.01,
        //         "rate":2000,
        //         "time":"2021-04-25T17:05:42.000Z",
        //         "type":0,
        //         "status":0
        //     }
        //
        $id = $this->safe_string_2($order, 'id', 'entry_id');
        $marketId = $this->safe_string($order, 'symbol');
        $symbol = $this->safe_symbol($marketId, $market);
        $timestamp = $this->parse8601($this->safe_string($order, 'time'));
        $price = $this->safe_string($order, 'rate');
        $amount = $this->safe_string_2($order, 'amount', 'btc');
        $filled = $this->safe_string($order, 'filled');
        $remaining = $this->safe_string($order, 'remaining');
        $average = $this->safe_string($order, 'avg_cost');
        $cost = $this->safe_string($order, 'cost');
        $type = $this->safe_string_lower($order, 'type');
        if ($type === '0') {
            $type = 'limit';
        }
        $status = $this->parse_order_status($this->safe_string($order, 'status'));
        $side = $this->safe_string_lower($order, 'side');
        $feeCost = $this->safe_number($order, 'fee');
        $fee = null;
        if ($feeCost !== null) {
            $feeCurrencyCode = null;
            $fee = array(
                'cost' => $feeCost,
                'currency' => $feeCurrencyCode,
            );
        }
        return $this->safe_order(array(
            'info' => $order,
            'id' => $id,
            'clientOrderId' => null,
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'lastTradeTimestamp' => null,
            'symbol' => $symbol,
            'type' => $type,
            'timeInForce' => null,
            'postOnly' => null,
            'side' => $side,
            'price' => $price,
            'stopPrice' => null,
            'amount' => $amount,
            'cost' => $cost,
            'average' => $average,
            'filled' => $filled,
            'remaining' => $remaining,
            'status' => $status,
            'fee' => $fee,
            'trades' => null,
        ), $market);
    }

    public function create_order($symbol, $type, $side, $amount, $price = null, $params = array ()) {
        if ($type !== 'limit' && $type !== 'market') {
            throw new ExchangeError($this->id . ' allows limit and $market orders only');
        }
        yield $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'side' => strtoupper($side),
            'symbol' => $market['uppercaseId'],
            'quantity' => $this->amount_to_precision($symbol, $amount),
            // 'target_rate' => $this->price_to_precision($symbol, targetRate),
            // 't_rate' => $this->price_to_precision($symbol, stopPrice),
            // 'trail_rate' => $this->price_to_precision($symbol, trailRate),
            // To Place Simple Buy or Sell Order use rate
            // To Place Stoploss Buy or Sell Order use rate & t_rate
            // To Place Bracket Buy or Sell Order use rate , t_rate, target_rate & trail_rate
        );
        $method = 'v2PostOrders';
        if ($type === 'limit') {
            $request['rate'] = $this->price_to_precision($symbol, $price);
        } elseif ($type === 'market') {
            $method = 'v1PostPlaceMarketOrderQntySymbol';
            $request['market'] = $market['quoteId'];
        } else {
            throw new ExchangeError($this->id . ' allows limit and $market orders only');
        }
        $response = yield $this->$method (array_merge($request, $params));
        //
        //     {
        //         "data":"Successfully placed bid to purchase currency",
        //         "status":1,
        //         "error":null,
        //         "id":5424475,
        //         "code":200
        //     }
        //
        return $this->parse_order($response, $market);
    }

    public function cancel_order($id, $symbol = null, $params = array ()) {
        if ($symbol === null) {
            throw new ArgumentsRequired($this->id . ' cancelOrder() requires a $symbol argument');
        }
        yield $this->load_markets();
        $market = $this->market($symbol);
        $quoteSide = ($market['quoteId'] === 'USDT') ? 'usdtcancelOrder' : 'cancelOrder';
        $request = array(
            'entry_id' => $id,
            'symbol' => $market['uppercaseId'],
            'side' => $quoteSide,
        );
        $response = yield $this->v2PostCancel (array_merge($request, $params));
        return $this->parse_order($response, $market);
    }

    public function fetch_order($id, $symbol = null, $params = array ()) {
        if ($symbol === null) {
            throw new ArgumentsRequired($this->id . ' fetchOrder() requires a $symbol argument');
        }
        yield $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'symbol' => $market['id'],
            'entry_id' => $id,
        );
        $response = yield $this->v1PostOrderStatusSymbol (array_merge($request, $params));
        //
        //     {
        //         "data":array(
        //             {
        //                 "entry_id":5424475,
        //                 "btc":0.01,
        //                 "rate":2000,
        //                 "time":"2021-04-25T17:05:42.000Z",
        //                 "type":0,
        //                 "status":0,
        //                 "total":0.01,
        //                 "avg_cost":null,
        //                 "side":"BUY",
        //                 "amount":0.01,
        //                 "remaining":0.01,
        //                 "filled":0,
        //                 "cost":null,
        //                 "fee":0.05
        //             }
        //         ),
        //         "status":1,
        //         "error":null,
        //         "code":200
        //     }
        //
        $data = $this->safe_value($response, 'data', array());
        $first = $this->safe_value($data, 0);
        return $this->parse_order($first, $market);
    }

    public function fetch_open_orders($symbol = null, $since = null, $limit = null, $params = array ()) {
        if ($symbol === null) {
            throw new ArgumentsRequired($this->id . ' fetchOpenOrders() requires a $symbol argument');
        }
        yield $this->load_markets();
        $market = $this->market($symbol);
        $quoteSide = ($market['quoteId'] === 'USDT') ? 'usdtListOpenOrders' : 'listOpenOrders';
        $request = array(
            'symbol' => $market['uppercaseId'],
            'side' => $quoteSide,
            'page' => 0,
        );
        $response = yield $this->v2PostGetordersnew (array_merge($request, $params));
        //
        //     {
        //         "data":array(
        //             {
        //                 "entry_id":5424475,
        //                 "btc":0.01,
        //                 "rate":2000,
        //                 "time":"2021-04-25T17:05:42.000Z",
        //                 "type":0,
        //                 "status":0
        //             }
        //         ),
        //         "status":1,
        //         "error":null,
        //         "code":200
        //     }
        //
        $data = $this->safe_value($response, 'data', array());
        return $this->parse_orders($data, $market, $since, $limit);
    }

    public function parse_trade($trade, $market = null) {
        //
        // fetchMyTrades
        //
        //     {
        //         "type" => "BTC Sell order executed",
        //         "typeI" => 6,
        //         "crypto" => 5000,
        //         "amount" => 35.4,
        //         "rate" => 709800,
        //         "date" => "2020-05-22T15:05:34.000Z",
        //         "unit" => "INR",
        //         "factor" => 100000000,
        //         "fee" => 0.09,
        //         "delh_btc" => -5000,
        //         "delh_inr" => 0,
        //         "del_btc" => 0,
        //         "del_inr" => 35.4,
        //         "id" => "2938823"
        //     }
        //
        // fetchTrades
        //
        //     {
        //         "tradeId":"1909151",
        //         "price":"61904.6300",
        //         "quote_volume":1618.05,
        //         "base_volume":0.02607254,
        //         "timestamp":1634548602000,
        //         "type":"buy"
        //     }
        //
        $market = $this->safe_market(null, $market);
        $orderId = $this->safe_string_2($trade, 'id', 'tradeId');
        $timestamp = $this->parse8601($this->safe_string($trade, 'date'));
        $timestamp = $this->safe_integer($trade, 'timestamp', $timestamp);
        $priceString = $this->safe_string_2($trade, 'rate', 'price');
        $amountString = $this->safe_string($trade, 'amount');
        $side = $this->safe_string_lower($trade, 'type');
        if ($side !== null) {
            if (mb_strpos($side, 'buy') !== false) {
                $side = 'buy';
            } elseif (mb_strpos($side, 'sell') !== false) {
                $side = 'sell';
            }
        }
        $factor = $this->safe_string($trade, 'factor');
        $costString = null;
        if ($factor !== null) {
            $amountString = Precise::string_div($amountString, $factor);
        } else {
            $amountString = $this->safe_string($trade, 'base_volume');
            $costString = $this->safe_string($trade, 'quote_volume');
        }
        $symbol = $market['symbol'];
        $fee = null;
        $feeCostString = $this->safe_string($trade, 'fee');
        if ($feeCostString !== null) {
            $feeCurrencyCode = $market['quote'];
            $fee = array(
                'cost' => $feeCostString,
                'currency' => $feeCurrencyCode,
            );
        }
        return $this->safe_trade(array(
            'info' => $trade,
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'symbol' => $symbol,
            'id' => $orderId,
            'order' => $orderId,
            'type' => null,
            'side' => $side,
            'takerOrMaker' => null,
            'price' => $priceString,
            'amount' => $amountString,
            'cost' => $costString,
            'fee' => $fee,
        ), $market);
    }

    public function fetch_my_trades($symbol = null, $since = null, $limit = null, $params = array ()) {
        if ($symbol === null) {
            throw new ArgumentsRequired($this->id . ' fetchMyTrades() requires a $symbol argument');
        }
        yield $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'symbol' => $market['id'],
            'page' => 0,
        );
        if ($since !== null) {
            $request['since'] = $this->iso8601($since);
        }
        $response = yield $this->v1PostListExecutedOrdersSymbol (array_merge($request, $params));
        //
        //     {
        //         "data" => array(
        //             array(
        //                 "type" => "BTC Sell order executed",
        //                 "typeI" => 6,
        //                 "crypto" => 5000,
        //                 "amount" => 35.4,
        //                 "rate" => 709800,
        //                 "date" => "2020-05-22T15:05:34.000Z",
        //                 "unit" => "INR",
        //                 "factor" => 100000000,
        //                 "fee" => 0.09,
        //                 "delh_btc" => -5000,
        //                 "delh_inr" => 0,
        //                 "del_btc" => 0,
        //                 "del_inr" => 35.4,
        //                 "id" => "2938823"
        //             ),
        //             {
        //                 "type" => "BTC Sell order executed",
        //                 "typeI" => 6,
        //                 "crypto" => 195000,
        //                 "amount" => 1380.58,
        //                 "rate" => 709765.5,
        //                 "date" => "2020-05-22T15:05:34.000Z",
        //                 "unit" => "INR",
        //                 "factor" => 100000000,
        //                 "fee" => 3.47,
        //                 "delh_btc" => -195000,
        //                 "delh_inr" => 0,
        //                 "del_btc" => 0,
        //                 "del_inr" => 1380.58,
        //                 "id" => "2938823"
        //             }
        //         ),
        //         "status" => 1,
        //         "error" => null,
        //         "code" => 200
        //     }
        //
        $data = $this->safe_value($response, 'data', array());
        return $this->parse_trades($data, $market, $since, $limit);
    }

    public function fetch_trades($symbol, $since = null, $limit = null, $params = array ()) {
        /**
         * get the list of most recent trades for a particular $symbol
         * @param {str} $symbol unified $symbol of the $market to fetch trades for
         * @param {int|null} $since timestamp in ms of the earliest trade to fetch
         * @param {int|null} $limit the maximum amount of trades to fetch
         * @param {dict} $params extra parameters specific to the bitbns api endpoint
         * @return {[dict]} a list of ~@link https://docs.ccxt.com/en/latest/manual.html?#public-trades trade structures~
         */
        if ($symbol === null) {
            throw new ArgumentsRequired($this->id . ' fetchTrades() requires a $symbol argument');
        }
        yield $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'coin' => $market['baseId'],
            'market' => $market['quoteId'],
        );
        $response = yield $this->wwwGetExchangeDataTradedetails (array_merge($request, $params));
        //
        //     [
        //         array("tradeId":"1909151","price":"61904.6300","quote_volume":1618.05,"base_volume":0.02607254,"timestamp":1634548602000,"type":"buy"),
        //         array("tradeId":"1909153","price":"61893.9000","quote_volume":16384.42,"base_volume":0.26405767,"timestamp":1634548999000,"type":"sell"),
        //         array("tradeId":"1909155","price":"61853.1100","quote_volume":2304.37,"base_volume":0.03716263,"timestamp":1634549670000,"type":"sell")
        //     }
        //
        return $this->parse_trades($response, $market, $since, $limit);
    }

    public function fetch_deposits($code = null, $since = null, $limit = null, $params = array ()) {
        if ($code === null) {
            throw new ArgumentsRequired($this->id . ' fetchDeposits() requires a $currency $code argument');
        }
        yield $this->load_markets();
        $currency = $this->currency($code);
        $request = array(
            'symbol' => $currency['id'],
            'page' => 0,
        );
        $response = yield $this->v1PostDepositHistorySymbol (array_merge($request, $params));
        //
        //     {
        //         "data":array(
        //             {
        //                 "type":"USDT deposited",
        //                 "typeI":1,
        //                 "amount":100,
        //                 "date":"2021-04-24T14:56:04.000Z",
        //                 "unit":"USDT",
        //                 "factor":100,
        //                 "fee":0,
        //                 "delh_btc":0,
        //                 "delh_inr":0,
        //                 "rate":0,
        //                 "del_btc":10000,
        //                 "del_inr":0
        //             }
        //         ),
        //         "status":1,
        //         "error":null,
        //         "code":200
        //     }
        //
        $data = $this->safe_value($response, 'data', array());
        return $this->parse_transactions($data, $currency, $since, $limit);
    }

    public function fetch_withdrawals($code = null, $since = null, $limit = null, $params = array ()) {
        if ($code === null) {
            throw new ArgumentsRequired($this->id . ' fetchWithdrawals() requires a $currency $code argument');
        }
        yield $this->load_markets();
        $currency = $this->currency($code);
        $request = array(
            'symbol' => $currency['id'],
            'page' => 0,
        );
        $response = yield $this->v1PostWithdrawHistorySymbol (array_merge($request, $params));
        //
        //     ...
        //
        $data = $this->safe_value($response, 'data', array());
        return $this->parse_transactions($data, $currency, $since, $limit);
    }

    public function parse_transaction_status_by_type($status, $type = null) {
        $statusesByType = array(
            'deposit' => array(
                '0' => 'pending',
                '1' => 'ok',
            ),
            'withdrawal' => array(
                '0' => 'pending', // Email Sent
                '1' => 'canceled', // Cancelled (different from 1 = ok in deposits)
                '2' => 'pending', // Awaiting Approval
                '3' => 'failed', // Rejected
                '4' => 'pending', // Processing
                '5' => 'failed', // Failure
                '6' => 'ok', // Completed
            ),
        );
        $statuses = $this->safe_value($statusesByType, $type, array());
        return $this->safe_string($statuses, $status, $status);
    }

    public function parse_transaction($transaction, $currency = null) {
        //
        // fetchDeposits
        //
        //     {
        //         "type":"USDT deposited",
        //         "typeI":1,
        //         "amount":100,
        //         "date":"2021-04-24T14:56:04.000Z",
        //         "unit":"USDT",
        //         "factor":100,
        //         "fee":0,
        //         "delh_btc":0,
        //         "delh_inr":0,
        //         "rate":0,
        //         "del_btc":10000,
        //         "del_inr":0
        //     }
        //
        // fetchWithdrawals
        //
        //     ...
        //
        $currencyId = $this->safe_string($transaction, 'unit');
        $code = $this->safe_currency_code($currencyId, $currency);
        $timestamp = $this->parse8601($this->safe_string($transaction, 'date'));
        $type = $this->safe_string($transaction, 'type');
        $status = null;
        if ($type !== null) {
            if (mb_strpos($type, 'deposit') !== false) {
                $type = 'deposit';
                $status = 'ok';
            } elseif (mb_strpos($type, 'withdraw') !== false) {
                $type = 'withdrawal';
            }
        }
        // $status = $this->parse_transaction_status_by_type($this->safe_string($transaction, 'status'), $type);
        $amount = $this->safe_number($transaction, 'amount');
        $feeCost = $this->safe_number($transaction, 'fee');
        $fee = null;
        if ($feeCost !== null) {
            $fee = array( 'currency' => $code, 'cost' => $feeCost );
        }
        return array(
            'info' => $transaction,
            'id' => null,
            'txid' => null,
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'network' => null,
            'address' => null,
            'addressTo' => null,
            'addressFrom' => null,
            'tag' => null,
            'tagTo' => null,
            'tagFrom' => null,
            'type' => $type,
            'amount' => $amount,
            'currency' => $code,
            'status' => $status,
            'updated' => null,
            'internal' => null,
            'fee' => $fee,
        );
    }

    public function fetch_deposit_address($code, $params = array ()) {
        yield $this->load_markets();
        $currency = $this->currency($code);
        $request = array(
            'symbol' => $currency['id'],
        );
        $response = yield $this->v1PostGetCoinAddressSymbol (array_merge($request, $params));
        //
        //     {
        //         "data":array(
        //             "token":"0x680dee9edfff0c397736e10b017cf6a0aee4ba31",
        //             "expiry":"2022-04-24 22:30:11"
        //         ),
        //         "status":1,
        //         "error":null
        //     }
        //
        $data = $this->safe_value($response, 'data', array());
        $address = $this->safe_string($data, 'token');
        $tag = $this->safe_string($data, 'tag');
        $this->check_address($address);
        return array(
            'currency' => $code,
            'address' => $address,
            'tag' => $tag,
            'network' => null,
            'info' => $response,
        );
    }

    public function nonce() {
        return $this->milliseconds();
    }

    public function sign($path, $api = 'www', $method = 'GET', $params = array (), $headers = null, $body = null) {
        if (!(is_array($this->urls['api']) && array_key_exists($api, $this->urls['api']))) {
            throw new ExchangeError($this->id . ' does not have a testnet/sandbox URL for ' . $api . ' endpoints');
        }
        if ($api !== 'www') {
            $this->check_required_credentials();
            $headers = array(
                'X-BITBNS-APIKEY' => $this->apiKey,
            );
        }
        $baseUrl = $this->implode_hostname($this->urls['api'][$api]);
        $url = $baseUrl . '/' . $this->implode_params($path, $params);
        $query = $this->omit($params, $this->extract_params($path));
        $nonce = (string) $this->nonce();
        if ($method === 'GET') {
            if ($query) {
                $url .= '?' . $this->urlencode($query);
            }
        } elseif ($method === 'POST') {
            if ($query) {
                $body = $this->json($query);
            } else {
                $body = '{}';
            }
            $auth = array(
                'timeStamp_nonce' => $nonce,
                'body' => $body,
            );
            $payload = base64_encode($this->json($auth));
            $signature = $this->hmac($payload, $this->encode($this->secret), 'sha512');
            $headers['X-BITBNS-PAYLOAD'] = $this->decode($payload);
            $headers['X-BITBNS-SIGNATURE'] = $signature;
            $headers['Content-Type'] = 'application/x-www-form-urlencoded';
        }
        return array( 'url' => $url, 'method' => $method, 'body' => $body, 'headers' => $headers );
    }

    public function handle_errors($httpCode, $reason, $url, $method, $headers, $body, $response, $requestHeaders, $requestBody) {
        if ($response === null) {
            return; // fallback to default $error handler
        }
        //
        //     array("msg":"Invalid Request","status":-1,"code":400)
        //     array("data":array(),"status":0,"error":"Nothing to show","code":417)
        //
        $code = $this->safe_string($response, 'code');
        $message = $this->safe_string($response, 'msg');
        $error = ($code !== null) && ($code !== '200');
        if ($error || ($message !== null)) {
            $feedback = $this->id . ' ' . $body;
            $this->throw_exactly_matched_exception($this->exceptions['exact'], $code, $feedback);
            $this->throw_exactly_matched_exception($this->exceptions['exact'], $message, $feedback);
            $this->throw_broadly_matched_exception($this->exceptions['broad'], $message, $feedback);
            throw new ExchangeError($feedback); // unknown $message
        }
    }
}
