<?php

namespace ccxt;

// PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
// https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

use Exception; // a common import
use \ccxt\ExchangeError;
use \ccxt\ArgumentsRequired;
use \ccxt\BadRequest;
use \ccxt\InvalidOrder;

class exmo extends Exchange {

    public function describe() {
        return $this->deep_extend(parent::describe (), array(
            'id' => 'exmo',
            'name' => 'EXMO',
            'countries' => array( 'LT' ), // Lithuania
            'rateLimit' => 350, // once every 350 ms ≈ 180 requests per minute ≈ 3 requests per second
            'version' => 'v1.1',
            'has' => array(
                'CORS' => null,
                'spot' => true,
                'margin' => true,
                'swap' => false,
                'future' => false,
                'option' => false,
                'addMargin' => true,
                'cancelOrder' => true,
                'cancelOrders' => false,
                'createDepositAddress' => false,
                'createOrder' => true,
                'createStopLimitOrder' => true,
                'createStopMarketOrder' => true,
                'createStopOrder' => true,
                'fetchAccounts' => false,
                'fetchBalance' => true,
                'fetchCanceledOrders' => true,
                'fetchCurrencies' => true,
                'fetchDeposit' => true,
                'fetchDepositAddress' => true,
                'fetchDeposits' => true,
                'fetchFundingHistory' => false,
                'fetchFundingRate' => false,
                'fetchFundingRateHistory' => false,
                'fetchFundingRates' => false,
                'fetchIndexOHLCV' => false,
                'fetchMarkets' => true,
                'fetchMarkOHLCV' => false,
                'fetchMyTrades' => true,
                'fetchOHLCV' => true,
                'fetchOpenInterestHistory' => false,
                'fetchOpenOrders' => true,
                'fetchOrder' => 'emulated',
                'fetchOrderBook' => true,
                'fetchOrderBooks' => true,
                'fetchOrderTrades' => true,
                'fetchPremiumIndexOHLCV' => false,
                'fetchTicker' => true,
                'fetchTickers' => true,
                'fetchTrades' => true,
                'fetchTradingFee' => false,
                'fetchTradingFees' => true,
                'fetchTransactionFees' => true,
                'fetchTransactions' => true,
                'fetchTransfer' => false,
                'fetchTransfers' => false,
                'fetchWithdrawal' => true,
                'fetchWithdrawals' => true,
                'reduceMargin' => true,
                'setMargin' => false,
                'transfer' => false,
                'withdraw' => true,
            ),
            'timeframes' => array(
                '1m' => '1',
                '5m' => '5',
                '15m' => '15',
                '30m' => '30',
                '45m' => '45',
                '1h' => '60',
                '2h' => '120',
                '3h' => '180',
                '4h' => '240',
                '1d' => 'D',
                '1w' => 'W',
                '1M' => 'M',
            ),
            'urls' => array(
                'logo' => 'https://user-images.githubusercontent.com/1294454/27766491-1b0ea956-5eda-11e7-9225-40d67b481b8d.jpg',
                'api' => array(
                    'public' => 'https://api.exmo.com',
                    'private' => 'https://api.exmo.com',
                    'web' => 'https://exmo.me',
                ),
                'www' => 'https://exmo.me',
                'referral' => 'https://exmo.me/?ref=131685',
                'doc' => array(
                    'https://exmo.me/en/api_doc?ref=131685',
                ),
                'fees' => 'https://exmo.com/en/docs/fees',
            ),
            'api' => array(
                'web' => array(
                    'get' => array(
                        'ctrl/feesAndLimits',
                        'en/docs/fees',
                    ),
                ),
                'public' => array(
                    'get' => array(
                        'currency',
                        'currency/list/extended',
                        'order_book',
                        'pair_settings',
                        'ticker',
                        'trades',
                        'candles_history',
                        'required_amount',
                        'payments/providers/crypto/list',
                    ),
                ),
                'private' => array(
                    'post' => array(
                        'user_info',
                        'order_create',
                        'order_cancel',
                        'stop_market_order_create',
                        'stop_market_order_cancel',
                        'user_open_orders',
                        'user_trades',
                        'user_cancelled_orders',
                        'order_trades',
                        'deposit_address',
                        'withdraw_crypt',
                        'withdraw_get_txid',
                        'excode_create',
                        'excode_load',
                        'code_check',
                        'wallet_history',
                        'wallet_operations',
                        'margin/user/order/create',
                        'margin/user/order/update',
                        'margin/user/order/cancel',
                        'margin/user/position/close',
                        'margin/user/position/margin_add',
                        'margin/user/position/margin_remove',
                        'margin/currency/list',
                        'margin/pair/list',
                        'margin/settings',
                        'margin/funding/list',
                        'margin/user/info',
                        'margin/user/order/list',
                        'margin/user/order/history',
                        'margin/user/order/trades',
                        'margin/user/order/max_quantity',
                        'margin/user/position/list',
                        'margin/user/position/margin_remove_info',
                        'margin/user/position/margin_add_info',
                        'margin/user/wallet/list',
                        'margin/user/wallet/history',
                        'margin/user/trade/list',
                        'margin/trades',
                        'margin/liquidation/feed',
                    ),
                ),
            ),
            'fees' => array(
                'trading' => array(
                    'feeSide' => 'get',
                    'tierBased' => true,
                    'percentage' => true,
                    'maker' => $this->parse_number('0.004'),
                    'taker' => $this->parse_number('0.004'),
                ),
                'funding' => array(
                    'tierBased' => false,
                    'percentage' => false, // fixed funding fees for crypto, see fetchTransactionFees below
                ),
            ),
            'options' => array(
                'networks' => array(
                    'ETH' => 'ERC20',
                    'TRX' => 'TRC20',
                ),
                'fetchTradingFees' => array(
                    'method' => 'fetchPrivateTradingFees', // or 'fetchPublicTradingFees'
                ),
                'margin' => array(
                    'fillResponseFromRequest' => true,
                ),
            ),
            'commonCurrencies' => array(
                'GMT' => 'GMT Token',
            ),
            'exceptions' => array(
                'exact' => array(
                    '40005' => '\\ccxt\\AuthenticationError', // Authorization error, incorrect signature
                    '40009' => '\\ccxt\\InvalidNonce', //
                    '40015' => '\\ccxt\\ExchangeError', // API function do not exist
                    '40016' => '\\ccxt\\OnMaintenance', // array("result":false,"error":"Error 40016 => Maintenance work in progress")
                    '40017' => '\\ccxt\\AuthenticationError', // Wrong API Key
                    '40032' => '\\ccxt\\PermissionDenied', // array("result":false,"error":"Error 40032 => Access is denied for this API key")
                    '40033' => '\\ccxt\\PermissionDenied', // array("result":false,"error":"Error 40033 => Access is denied, this resources are temporarily blocked to user")
                    '40034' => '\\ccxt\\RateLimitExceeded', // array("result":false,"error":"Error 40034 => Access is denied, rate limit is exceeded")
                    '50052' => '\\ccxt\\InsufficientFunds',
                    '50054' => '\\ccxt\\InsufficientFunds',
                    '50304' => '\\ccxt\\OrderNotFound', // "Order was not found '123456789'" (fetching order trades for an order that does not have trades yet)
                    '50173' => '\\ccxt\\OrderNotFound', // "Order with id X was not found." (cancelling non-existent, closed and cancelled order)
                    '50277' => '\\ccxt\\InvalidOrder',
                    '50319' => '\\ccxt\\InvalidOrder', // Price by order is less than permissible minimum for this pair
                    '50321' => '\\ccxt\\InvalidOrder', // Price by order is more than permissible maximum for this pair
                    '50381' => '\\ccxt\\InvalidOrder', // array("result":false,"error":"Error 50381 => More than 2 decimal places are not permitted for pair BTC_USD")
                ),
                'broad' => array(
                    'range period is too long' => '\\ccxt\\BadRequest',
                    'invalid syntax' => '\\ccxt\\BadRequest',
                    'API rate limit exceeded' => '\\ccxt\\RateLimitExceeded', // array("result":false,"error":"API rate limit exceeded for x.x.x.x. Retry after 60 sec.","history":array(),"begin":1579392000,"end":1579478400)
                ),
            ),
        ));
    }

    public function modify_margin_helper($symbol, $amount, $type, $params = array ()) {
        $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'position_id' => $market['id'],
            'quantity' => $amount,
        );
        $method = null;
        if ($type === 'add') {
            $method = 'privatePostMarginUserPositionMarginAdd';
        } elseif ($type === 'reduce') {
            $method = 'privatePostMarginUserPositionMarginReduce';
        }
        $response = $this->$method (array_merge($request, $params));
        //
        //      array()
        //
        $margin = $this->parse_margin_modification($response, $market);
        $options = $this->safe_value($this->options, 'margin', array());
        $fillResponseFromRequest = $this->safe_value($options, 'fillResponseFromRequest', true);
        if ($fillResponseFromRequest) {
            $margin['type'] = $type;
            $margin['amount'] = $amount;
        }
        return $margin;
    }

    public function parse_margin_modification($data, $market = null) {
        //
        //      array()
        //
        return array(
            'info' => $data,
            'type' => null,
            'amount' => null,
            'code' => $this->safe_value($market, 'quote'),
            'symbol' => $this->safe_symbol(null, $market),
            'total' => null,
            'status' => 'ok',
        );
    }

    public function reduce_margin($symbol, $amount, $params = array ()) {
        return $this->modify_margin_helper($symbol, $amount, 'reduce', $params);
    }

    public function add_margin($symbol, $amount, $params = array ()) {
        return $this->modify_margin_helper($symbol, $amount, 'add', $params);
    }

    public function fetch_trading_fees($params = array ()) {
        $method = $this->safe_string($params, 'method');
        $params = $this->omit($params, 'method');
        if ($method === null) {
            $options = $this->safe_value($this->options, 'fetchTradingFees', array());
            $method = $this->safe_string($options, 'method', 'fetchPrivateTradingFees');
        }
        return $this->$method ($params);
    }

    public function fetch_private_trading_fees($params = array ()) {
        $this->load_markets();
        $response = $this->privatePostMarginPairList ($params);
        //
        //     {
        //         $pairs => [{
        //             name => 'EXM_USD',
        //             buy_price => '0.02728391',
        //             sell_price => '0.0276',
        //             last_trade_price => '0.0276',
        //             ticker_updated => '1646956050056696046',
        //             is_fair_price => true,
        //             max_price_precision => '8',
        //             min_order_quantity => '1',
        //             max_order_quantity => '50000',
        //             min_order_price => '0.00000001',
        //             max_order_price => '1000',
        //             max_position_quantity => '50000',
        //             trade_taker_fee => '0.05',
        //             trade_maker_fee => '0',
        //             liquidation_fee => '0.5',
        //             max_leverage => '3',
        //             default_leverage => '3',
        //             liquidation_level => '5',
        //             margin_call_level => '7.5',
        //             position => '1',
        //             updated => '1638976144797807397'
        //         }
        //         ...
        //         ]
        //     }
        //
        $pairs = $this->safe_value($response, 'pairs', array());
        $result = array();
        for ($i = 0; $i < count($pairs); $i++) {
            $pair = $pairs[$i];
            $marketId = $this->safe_string($pair, 'name');
            $symbol = $this->safe_symbol($marketId, null, '_');
            $makerString = $this->safe_string($pair, 'trade_maker_fee');
            $takerString = $this->safe_string($pair, 'trade_taker_fee');
            $maker = $this->parse_number(Precise::string_div($makerString, '100'));
            $taker = $this->parse_number(Precise::string_div($takerString, '100'));
            $result[$symbol] = array(
                'info' => $pair,
                'symbol' => $symbol,
                'maker' => $maker,
                'taker' => $taker,
                'percentage' => true,
                'tierBased' => true,
            );
        }
        return $result;
    }

    public function fetch_public_trading_fees($params = array ()) {
        $this->load_markets();
        $response = $this->publicGetPairSettings ($params);
        //
        //     {
        //         BTC_USD => array(
        //             min_quantity => '0.00002',
        //             max_quantity => '1000',
        //             min_price => '1',
        //             max_price => '150000',
        //             max_amount => '500000',
        //             min_amount => '1',
        //             price_precision => '2',
        //             commission_taker_percent => '0.3',
        //             commission_maker_percent => '0.3'
        //         ),
        //     }
        //
        $result = array();
        for ($i = 0; $i < count($this->symbols); $i++) {
            $symbol = $this->symbols[$i];
            $market = $this->market($symbol);
            $fee = $this->safe_value($response, $market['id'], array());
            $makerString = $this->safe_string($fee, 'commission_maker_percent');
            $takerString = $this->safe_string($fee, 'commission_taker_percent');
            $maker = $this->parse_number(Precise::string_div($makerString, '100'));
            $taker = $this->parse_number(Precise::string_div($takerString, '100'));
            $result[$symbol] = array(
                'info' => $fee,
                'symbol' => $symbol,
                'maker' => $maker,
                'taker' => $taker,
                'percentage' => true,
                'tierBased' => true,
            );
        }
        return $result;
    }

    public function parse_fixed_float_value($input) {
        if (($input === null) || ($input === '-')) {
            return null;
        }
        if ($input === '') {
            return 0;
        }
        $isPercentage = (mb_strpos($input, '%') !== false);
        $parts = explode(' ', $input);
        $value = str_replace('%', '', $parts[0]);
        $result = floatval($value);
        if (($result > 0) && $isPercentage) {
            throw new ExchangeError($this->id . ' parseFixedFloatValue() detected an unsupported non-zero percentage-based fee ' . $input);
        }
        return $result;
    }

    public function fetch_transaction_fees($codes = null, $params = array ()) {
        $this->load_markets();
        $currencyList = $this->publicGetCurrencyListExtended ($params);
        //
        //     array(
        //         array("name":"VLX","description":"Velas"),
        //         array("name":"RUB","description":"Russian Ruble"),
        //         array("name":"BTC","description":"Bitcoin"),
        //         array("name":"USD","description":"US Dollar")
        //     )
        //
        $cryptoList = $this->publicGetPaymentsProvidersCryptoList ($params);
        //
        //     {
        //         "BTC":array(
        //             array( "type":"deposit", "name":"BTC", "currency_name":"BTC", "min":"0.001", "max":"0", "enabled":true,"comment":"Minimum deposit amount is 0.001 BTC. We do not support BSC and BEP20 network, please consider this when sending funds", "commission_desc":"0%", "currency_confirmations":1 ),
        //             array( "type":"withdraw", "name":"BTC", "currency_name":"BTC", "min":"0.001", "max":"350", "enabled":true,"comment":"Do not withdraw directly to the Crowdfunding or ICO address as your account will not be credited with tokens from such sales.", "commission_desc":"0.0005 BTC", "currency_confirmations":6 )
        //         ),
        //         "ETH":array(
        //             array( "type":"withdraw", "name":"ETH", "currency_name":"ETH", "min":"0.01", "max":"500", "enabled":true,"comment":"Do not withdraw directly to the Crowdfunding or ICO address as your account will not be credited with tokens from such sales.", "commission_desc":"0.004 ETH", "currency_confirmations":4 ),
        //             array( "type":"deposit", "name":"ETH", "currency_name":"ETH", "min":"0.01", "max":"0", "enabled":true,"comment":"Minimum deposit amount is 0.01 ETH. We do not support BSC and BEP20 network, please consider this when sending funds", "commission_desc":"0%", "currency_confirmations":1 )
        //         ),
        //         "USDT":array(
        //             array( "type":"deposit", "name":"USDT (OMNI)", "currency_name":"USDT", "min":"10", "max":"0", "enabled":false,"comment":"Minimum deposit amount is 10 USDT", "commission_desc":"0%", "currency_confirmations":2 ),
        //             array( "type":"withdraw", "name":"USDT (OMNI)", "currency_name":"USDT", "min":"10", "max":"100000", "enabled":false,"comment":"Do not withdraw directly to the Crowdfunding or ICO address as your account will not be credited with tokens from such sales.", "commission_desc":"5 USDT", "currency_confirmations":6 ),
        //             array( "type":"deposit", "name":"USDT (ERC20)", "currency_name":"USDT", "min":"10", "max":"0", "enabled":true,"comment":"Minimum deposit amount is 10 USDT", "commission_desc":"0%", "currency_confirmations":2 ),
        //             array(
        //                 "type":"withdraw",
        //                 "name":"USDT (ERC20)",
        //                 "currency_name":"USDT",
        //                 "min":"55",
        //                 "max":"200000",
        //                 "enabled":true,
        //                 "comment":"Caution! Do not withdraw directly to a crowdfund or ICO address, as your account will not be credited with tokens from such sales. Recommendation => Due to the high load of ERC20 network, using TRC20 address for withdrawal is recommended.",
        //                 "commission_desc":"10 USDT",
        //                 "currency_confirmations":6
        //             ),
        //             array( "type":"deposit", "name":"USDT (TRC20)", "currency_name":"USDT", "min":"10", "max":"100000", "enabled":true,"comment":"Minimum deposit amount is 10 USDT. Only TRON main network supported", "commission_desc":"0%", "currency_confirmations":2 ),
        //             array( "type":"withdraw", "name":"USDT (TRC20)", "currency_name":"USDT", "min":"10", "max":"150000", "enabled":true,"comment":"Caution! Do not withdraw directly to a crowdfund or ICO address, as your account will not be credited with tokens from such sales. Only TRON main network supported.", "commission_desc":"1 USDT", "currency_confirmations":6 )
        //         ),
        //         "XLM":array(
        //             array( "type":"deposit", "name":"XLM", "currency_name":"XLM", "min":"1", "max":"1000000", "enabled":true,"comment":"Attention! A deposit without memo(invoice) will not be credited. Minimum deposit amount is 1 XLM. We do not support BSC and BEP20 network, please consider this when sending funds", "commission_desc":"0%", "currency_confirmations":1 ),
        //             array( "type":"withdraw", "name":"XLM", "currency_name":"XLM", "min":"21", "max":"1000000", "enabled":true,"comment":"Caution! Do not withdraw directly to a crowdfund or ICO address, as your account will not be credited with tokens from such sales.", "commission_desc":"0.01 XLM", "currency_confirmations":1 )
        //         ),
        //     }
        //
        $result = array(
            'info' => $cryptoList,
            'withdraw' => array(),
            'deposit' => array(),
        );
        for ($i = 0; $i < count($currencyList); $i++) {
            $currency = $currencyList[$i];
            $currencyId = $this->safe_string($currency, 'name');
            $code = $this->safe_currency_code($currencyId);
            $providers = $this->safe_value($cryptoList, $currencyId, array());
            for ($j = 0; $j < count($providers); $j++) {
                $provider = $providers[$j];
                $type = $this->safe_string($provider, 'type');
                $commissionDesc = $this->safe_string($provider, 'commission_desc');
                $newFee = $this->parse_fixed_float_value($commissionDesc);
                $previousFee = $this->safe_number($result[$type], $code);
                if (($previousFee === null) || (($newFee !== null) && ($newFee < $previousFee))) {
                    $result[$type][$code] = $newFee;
                }
            }
        }
        // cache them for later use
        $this->options['fundingFees'] = $result;
        return $result;
    }

    public function fetch_currencies($params = array ()) {
        /**
         * fetches all available currencies on an exchange
         * @param {dict} $params extra parameters specific to the exmo api endpoint
         * @return {dict} an associative dictionary of currencies
         */
        //
        $currencyList = $this->publicGetCurrencyListExtended ($params);
        //
        //     array(
        //         array("name":"VLX","description":"Velas"),
        //         array("name":"RUB","description":"Russian Ruble"),
        //         array("name":"BTC","description":"Bitcoin"),
        //         array("name":"USD","description":"US Dollar")
        //     )
        //
        $cryptoList = $this->publicGetPaymentsProvidersCryptoList ($params);
        //
        //     {
        //         "BTC":array(
        //             array( "type":"deposit", "name":"BTC", "currency_name":"BTC", "min":"0.001", "max":"0", "enabled":true,"comment":"Minimum deposit amount is 0.001 BTC. We do not support BSC and BEP20 network, please consider this when sending funds", "commission_desc":"0%", "currency_confirmations":1 ),
        //             array( "type":"withdraw", "name":"BTC", "currency_name":"BTC", "min":"0.001", "max":"350", "enabled":true,"comment":"Do not withdraw directly to the Crowdfunding or ICO address as your account will not be credited with tokens from such sales.", "commission_desc":"0.0005 BTC", "currency_confirmations":6 )
        //         ),
        //         "ETH":array(
        //             array( "type":"withdraw", "name":"ETH", "currency_name":"ETH", "min":"0.01", "max":"500", "enabled":true,"comment":"Do not withdraw directly to the Crowdfunding or ICO address as your account will not be credited with tokens from such sales.", "commission_desc":"0.004 ETH", "currency_confirmations":4 ),
        //             array( "type":"deposit", "name":"ETH", "currency_name":"ETH", "min":"0.01", "max":"0", "enabled":true,"comment":"Minimum deposit amount is 0.01 ETH. We do not support BSC and BEP20 network, please consider this when sending funds", "commission_desc":"0%", "currency_confirmations":1 )
        //         ),
        //         "USDT":array(
        //             array( "type":"deposit", "name":"USDT (OMNI)", "currency_name":"USDT", "min":"10", "max":"0", "enabled":false,"comment":"Minimum deposit amount is 10 USDT", "commission_desc":"0%", "currency_confirmations":2 ),
        //             array( "type":"withdraw", "name":"USDT (OMNI)", "currency_name":"USDT", "min":"10", "max":"100000", "enabled":false,"comment":"Do not withdraw directly to the Crowdfunding or ICO address as your account will not be credited with tokens from such sales.", "commission_desc":"5 USDT", "currency_confirmations":6 ),
        //             array( "type":"deposit", "name":"USDT (ERC20)", "currency_name":"USDT", "min":"10", "max":"0", "enabled":true,"comment":"Minimum deposit amount is 10 USDT", "commission_desc":"0%", "currency_confirmations":2 ),
        //             array(
        //                 "type":"withdraw",
        //                 "name":"USDT (ERC20)",
        //                 "currency_name":"USDT",
        //                 "min":"55",
        //                 "max":"200000",
        //                 "enabled":true,
        //                 "comment":"Caution! Do not withdraw directly to a crowdfund or ICO address, as your account will not be credited with tokens from such sales. Recommendation => Due to the high load of ERC20 network, using TRC20 address for withdrawal is recommended.",
        //                 "commission_desc":"10 USDT",
        //                 "currency_confirmations":6
        //             ),
        //             array( "type":"deposit", "name":"USDT (TRC20)", "currency_name":"USDT", "min":"10", "max":"100000", "enabled":true,"comment":"Minimum deposit amount is 10 USDT. Only TRON main network supported", "commission_desc":"0%", "currency_confirmations":2 ),
        //             array( "type":"withdraw", "name":"USDT (TRC20)", "currency_name":"USDT", "min":"10", "max":"150000", "enabled":true,"comment":"Caution! Do not withdraw directly to a crowdfund or ICO address, as your account will not be credited with tokens from such sales. Only TRON main network supported.", "commission_desc":"1 USDT", "currency_confirmations":6 )
        //         ),
        //         "XLM":array(
        //             array( "type":"deposit", "name":"XLM", "currency_name":"XLM", "min":"1", "max":"1000000", "enabled":true,"comment":"Attention! A deposit without memo(invoice) will not be credited. Minimum deposit amount is 1 XLM. We do not support BSC and BEP20 network, please consider this when sending funds", "commission_desc":"0%", "currency_confirmations":1 ),
        //             array( "type":"withdraw", "name":"XLM", "currency_name":"XLM", "min":"21", "max":"1000000", "enabled":true,"comment":"Caution! Do not withdraw directly to a crowdfund or ICO address, as your account will not be credited with tokens from such sales.", "commission_desc":"0.01 XLM", "currency_confirmations":1 )
        //         ),
        //     }
        //
        $result = array();
        for ($i = 0; $i < count($currencyList); $i++) {
            $currency = $currencyList[$i];
            $currencyId = $this->safe_string($currency, 'name');
            $name = $this->safe_string($currency, 'description');
            $providers = $this->safe_value($cryptoList, $currencyId);
            $active = false;
            $type = 'crypto';
            $limits = array(
                'deposit' => array(
                    'min' => null,
                    'max' => null,
                ),
                'withdraw' => array(
                    'min' => null,
                    'max' => null,
                ),
            );
            $fee = null;
            $depositEnabled = null;
            $withdrawEnabled = null;
            if ($providers === null) {
                $active = true;
                $type = 'fiat';
            } else {
                for ($j = 0; $j < count($providers); $j++) {
                    $provider = $providers[$j];
                    $type = $this->safe_string($provider, 'type');
                    $minValue = $this->safe_number($provider, 'min');
                    $maxValue = $this->safe_number($provider, 'max');
                    if ($maxValue === 0.0) {
                        $maxValue = null;
                    }
                    $activeProvider = $this->safe_value($provider, 'enabled');
                    if ($type === 'deposit') {
                        if ($activeProvider && !$depositEnabled) {
                            $depositEnabled = true;
                        } elseif (!$activeProvider) {
                            $depositEnabled = false;
                        }
                    } elseif ($type === 'withdraw') {
                        if ($activeProvider && !$withdrawEnabled) {
                            $withdrawEnabled = true;
                        } elseif (!$activeProvider) {
                            $withdrawEnabled = false;
                        }
                    }
                    if ($activeProvider) {
                        $active = true;
                        if (($limits[$type]['min'] === null) || ($minValue < $limits[$type]['min'])) {
                            $limits[$type]['min'] = $minValue;
                            $limits[$type]['max'] = $maxValue;
                            if ($type === 'withdraw') {
                                $commissionDesc = $this->safe_string($provider, 'commission_desc');
                                $fee = $this->parse_fixed_float_value($commissionDesc);
                            }
                        }
                    }
                }
            }
            $code = $this->safe_currency_code($currencyId);
            $result[$code] = array(
                'id' => $currencyId,
                'code' => $code,
                'name' => $name,
                'type' => $type,
                'active' => $active,
                'deposit' => $depositEnabled,
                'withdraw' => $withdrawEnabled,
                'fee' => $fee,
                'precision' => 8,
                'limits' => $limits,
                'info' => $providers,
            );
        }
        return $result;
    }

    public function fetch_markets($params = array ()) {
        /**
         * retrieves data on all markets for exmo
         * @param {dict} $params extra parameters specific to the exchange api endpoint
         * @return {[dict]} an array of objects representing $market data
         */
        $response = $this->publicGetPairSettings ($params);
        //
        //     {
        //         "BTC_USD":array(
        //             "min_quantity":"0.0001",
        //             "max_quantity":"1000",
        //             "min_price":"1",
        //             "max_price":"30000",
        //             "max_amount":"500000",
        //             "min_amount":"1",
        //             "price_precision":8,
        //             "commission_taker_percent":"0.4",
        //             "commission_maker_percent":"0.4"
        //         ),
        //     }
        //
        $keys = is_array($response) ? array_keys($response) : array();
        $result = array();
        for ($i = 0; $i < count($keys); $i++) {
            $id = $keys[$i];
            $market = $response[$id];
            $symbol = str_replace('_', '/', $id);
            list($baseId, $quoteId) = explode('/', $symbol);
            $base = $this->safe_currency_code($baseId);
            $quote = $this->safe_currency_code($quoteId);
            $takerString = $this->safe_string($market, 'commission_taker_percent');
            $makerString = $this->safe_string($market, 'commission_maker_percent');
            $result[] = array(
                'id' => $id,
                'symbol' => $symbol,
                'base' => $base,
                'quote' => $quote,
                'settle' => null,
                'baseId' => $baseId,
                'quoteId' => $quoteId,
                'settleId' => null,
                'type' => 'spot',
                'spot' => true,
                'margin' => true,
                'swap' => false,
                'future' => false,
                'option' => false,
                'active' => true,
                'contract' => false,
                'linear' => null,
                'inverse' => null,
                'taker' => $this->parse_number(Precise::string_div($takerString, '100')),
                'maker' => $this->parse_number(Precise::string_div($makerString, '100')),
                'contractSize' => null,
                'expiry' => null,
                'expiryDatetime' => null,
                'strike' => null,
                'optionType' => null,
                'precision' => array(
                    'amount' => intval('8'),
                    'price' => $this->safe_integer($market, 'price_precision'),
                ),
                'limits' => array(
                    'leverage' => array(
                        'min' => null,
                        'max' => null,
                    ),
                    'amount' => array(
                        'min' => $this->safe_number($market, 'min_quantity'),
                        'max' => $this->safe_number($market, 'max_quantity'),
                    ),
                    'price' => array(
                        'min' => $this->safe_number($market, 'min_price'),
                        'max' => $this->safe_number($market, 'max_price'),
                    ),
                    'cost' => array(
                        'min' => $this->safe_number($market, 'min_amount'),
                        'max' => $this->safe_number($market, 'max_amount'),
                    ),
                ),
                'info' => $market,
            );
        }
        return $result;
    }

    public function fetch_ohlcv($symbol, $timeframe = '1m', $since = null, $limit = null, $params = array ()) {
        /**
         * fetches historical candlestick data containing the open, high, low, and close price, and the volume of a $market
         * @param {str} $symbol unified $symbol of the $market $to fetch OHLCV data for
         * @param {str} $timeframe the length of time each candle represents
         * @param {int|null} $since timestamp in ms of the earliest candle $to fetch
         * @param {int|null} $limit the maximum amount of $candles $to fetch
         * @param {dict} $params extra parameters specific $to the exmo api endpoint
         * @return {[[int]]} A list of $candles ordered as timestamp, open, high, low, close, volume
         */
        $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'symbol' => $market['id'],
            'resolution' => $this->timeframes[$timeframe],
        );
        $options = $this->safe_value($this->options, 'fetchOHLCV');
        $maxLimit = $this->safe_integer($options, 'maxLimit', 3000);
        $duration = $this->parse_timeframe($timeframe);
        $now = $this->milliseconds();
        if ($since === null) {
            if ($limit === null) {
                throw new ArgumentsRequired($this->id . ' fetchOHLCV() requires a $since argument or a $limit argument');
            } else {
                if ($limit > $maxLimit) {
                    throw new BadRequest($this->id . ' fetchOHLCV() will serve ' . (string) $maxLimit . ' $candles at most');
                }
                $request['from'] = intval($now / 1000) - $limit * $duration - 1;
                $request['to'] = intval($now / 1000);
            }
        } else {
            $request['from'] = intval($since / 1000) - 1;
            if ($limit === null) {
                $request['to'] = intval($now / 1000);
            } else {
                if ($limit > $maxLimit) {
                    throw new BadRequest($this->id . ' fetchOHLCV() will serve ' . (string) $maxLimit . ' $candles at most');
                }
                $to = $this->sum($since, $limit * $duration * 1000);
                $request['to'] = intval($to / 1000);
            }
        }
        $response = $this->publicGetCandlesHistory (array_merge($request, $params));
        //
        //     {
        //         "candles":array(
        //             array("t":1584057600000,"o":0.02235144,"c":0.02400233,"h":0.025171,"l":0.02221,"v":5988.34031761),
        //             array("t":1584144000000,"o":0.0240373,"c":0.02367413,"h":0.024399,"l":0.0235,"v":2027.82522329),
        //             array("t":1584230400000,"o":0.02363458,"c":0.02319242,"h":0.0237948,"l":0.02223196,"v":1707.96944997),
        //         )
        //     }
        //
        $candles = $this->safe_value($response, 'candles', array());
        return $this->parse_ohlcvs($candles, $market, $timeframe, $since, $limit);
    }

    public function parse_ohlcv($ohlcv, $market = null) {
        //
        //     {
        //         "t":1584057600000,
        //         "o":0.02235144,
        //         "c":0.02400233,
        //         "h":0.025171,
        //         "l":0.02221,
        //         "v":5988.34031761
        //     }
        //
        return array(
            $this->safe_integer($ohlcv, 't'),
            $this->safe_number($ohlcv, 'o'),
            $this->safe_number($ohlcv, 'h'),
            $this->safe_number($ohlcv, 'l'),
            $this->safe_number($ohlcv, 'c'),
            $this->safe_number($ohlcv, 'v'),
        );
    }

    public function parse_balance($response) {
        $result = array( 'info' => $response );
        $free = $this->safe_value($response, 'balances', array());
        $used = $this->safe_value($response, 'reserved', array());
        $currencyIds = is_array($free) ? array_keys($free) : array();
        for ($i = 0; $i < count($currencyIds); $i++) {
            $currencyId = $currencyIds[$i];
            $code = $this->safe_currency_code($currencyId);
            $account = $this->account();
            if (is_array($free) && array_key_exists($currencyId, $free)) {
                $account['free'] = $this->safe_string($free, $currencyId);
            }
            if (is_array($used) && array_key_exists($currencyId, $used)) {
                $account['used'] = $this->safe_string($used, $currencyId);
            }
            $result[$code] = $account;
        }
        return $this->safe_balance($result);
    }

    public function fetch_balance($params = array ()) {
        /**
         * query for balance and get the amount of funds available for trading or funds locked in orders
         * @param {dict} $params extra parameters specific to the exmo api endpoint
         * @return {dict} a ~@link https://docs.ccxt.com/en/latest/manual.html?#balance-structure balance structure~
         */
        $this->load_markets();
        $response = $this->privatePostUserInfo ($params);
        //
        //     {
        //         "uid":131685,
        //         "server_date":1628999600,
        //         "balances":array(
        //             "EXM":"0",
        //             "USD":"0",
        //             "EUR":"0",
        //             "GBP":"0",
        //         ),
        //     }
        //
        return $this->parse_balance($response);
    }

    public function fetch_order_book($symbol, $limit = null, $params = array ()) {
        /**
         * fetches information on open orders with bid (buy) and ask (sell) prices, volumes and other data
         * @param {str} $symbol unified $symbol of the $market to fetch the order book for
         * @param {int|null} $limit the maximum amount of order book entries to return
         * @param {dict} $params extra parameters specific to the exmo api endpoint
         * @return {dict} A dictionary of {@link https://docs.ccxt.com/en/latest/manual.html#order-book-structure order book structures} indexed by $market symbols
         */
        $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'pair' => $market['id'],
        );
        if ($limit !== null) {
            $request['limit'] = $limit;
        }
        $response = $this->publicGetOrderBook (array_merge($request, $params));
        $result = $this->safe_value($response, $market['id']);
        return $this->parse_order_book($result, $symbol, null, 'bid', 'ask');
    }

    public function fetch_order_books($symbols = null, $limit = null, $params = array ()) {
        $this->load_markets();
        $ids = null;
        if ($symbols === null) {
            $ids = implode(',', $this->ids);
            // max URL length is 2083 $symbols, including http schema, hostname, tld, etc...
            if (strlen($ids) > 2048) {
                $numIds = is_array($this->ids) ? count($this->ids) : 0;
                throw new ExchangeError($this->id . ' fetchOrderBooks() has ' . (string) $numIds . ' $symbols exceeding max URL length, you are required to specify a list of $symbols in the first argument to fetchOrderBooks');
            }
        } else {
            $ids = $this->market_ids($symbols);
            $ids = implode(',', $ids);
        }
        $request = array(
            'pair' => $ids,
        );
        if ($limit !== null) {
            $request['limit'] = $limit;
        }
        $response = $this->publicGetOrderBook (array_merge($request, $params));
        $result = array();
        $marketIds = is_array($response) ? array_keys($response) : array();
        for ($i = 0; $i < count($marketIds); $i++) {
            $marketId = $marketIds[$i];
            $symbol = $marketId;
            if (is_array($this->markets_by_id) && array_key_exists($marketId, $this->markets_by_id)) {
                $market = $this->markets_by_id[$marketId];
                $symbol = $market['symbol'];
            }
            $result[$symbol] = $this->parse_order_book($response[$marketId], $symbol, null, 'bid', 'ask');
        }
        return $result;
    }

    public function parse_ticker($ticker, $market = null) {
        //
        //     {
        //         "buy_price":"0.00002996",
        //         "sell_price":"0.00003002",
        //         "last_trade":"0.00002992",
        //         "high":"0.00003028",
        //         "low":"0.00002935",
        //         "avg":"0.00002963",
        //         "vol":"1196546.3163222",
        //         "vol_curr":"35.80066578",
        //         "updated":1642291733
        //     }
        //
        $timestamp = $this->safe_timestamp($ticker, 'updated');
        $market = $this->safe_market(null, $market);
        $last = $this->safe_string($ticker, 'last_trade');
        return $this->safe_ticker(array(
            'symbol' => $market['symbol'],
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'high' => $this->safe_string($ticker, 'high'),
            'low' => $this->safe_string($ticker, 'low'),
            'bid' => $this->safe_string($ticker, 'buy_price'),
            'bidVolume' => null,
            'ask' => $this->safe_string($ticker, 'sell_price'),
            'askVolume' => null,
            'vwap' => null,
            'open' => null,
            'close' => $last,
            'last' => $last,
            'previousClose' => null,
            'change' => null,
            'percentage' => null,
            'average' => $this->safe_string($ticker, 'avg'),
            'baseVolume' => $this->safe_string($ticker, 'vol'),
            'quoteVolume' => $this->safe_string($ticker, 'vol_curr'),
            'info' => $ticker,
        ), $market);
    }

    public function fetch_tickers($symbols = null, $params = array ()) {
        /**
         * fetches price tickers for multiple markets, statistical calculations with the information calculated over the past 24 hours each $market
         * @param {[str]|null} $symbols unified $symbols of the markets to fetch the $ticker for, all $market tickers are returned if not assigned
         * @param {dict} $params extra parameters specific to the exmo api endpoint
         * @return {dict} an array of {@link https://docs.ccxt.com/en/latest/manual.html#$ticker-structure $ticker structures}
         */
        $this->load_markets();
        $response = $this->publicGetTicker ($params);
        //
        //     {
        //         "ADA_BTC":{
        //             "buy_price":"0.00002996",
        //             "sell_price":"0.00003002",
        //             "last_trade":"0.00002992",
        //             "high":"0.00003028",
        //             "low":"0.00002935",
        //             "avg":"0.00002963",
        //             "vol":"1196546.3163222",
        //             "vol_curr":"35.80066578",
        //             "updated":1642291733
        //         }
        //     }
        //
        $result = array();
        $marketIds = is_array($response) ? array_keys($response) : array();
        for ($i = 0; $i < count($marketIds); $i++) {
            $marketId = $marketIds[$i];
            $market = $this->safe_market($marketId, null, '_');
            $symbol = $market['symbol'];
            $ticker = $this->safe_value($response, $marketId);
            $result[$symbol] = $this->parse_ticker($ticker, $market);
        }
        return $this->filter_by_array($result, 'symbol', $symbols);
    }

    public function fetch_ticker($symbol, $params = array ()) {
        /**
         * fetches a price ticker, a statistical calculation with the information calculated over the past 24 hours for a specific $market
         * @param {str} $symbol unified $symbol of the $market to fetch the ticker for
         * @param {dict} $params extra parameters specific to the exmo api endpoint
         * @return {dict} a {@link https://docs.ccxt.com/en/latest/manual.html#ticker-structure ticker structure}
         */
        $this->load_markets();
        $response = $this->publicGetTicker ($params);
        $market = $this->market($symbol);
        return $this->parse_ticker($response[$market['id']], $market);
    }

    public function parse_trade($trade, $market = null) {
        //
        // fetchTrades (public)
        //
        //     array(
        //         "trade_id":165087520,
        //         "date":1587470005,
        //         "type":"buy",
        //         "quantity":"1.004",
        //         "price":"0.02491461",
        //         "amount":"0.02501426"
        //     ),
        //
        // fetchMyTrades, fetchOrderTrades
        //
        //     {
        //         "trade_id" => 3,
        //         "date" => 1435488248,
        //         "type" => "buy",
        //         "pair" => "BTC_USD",
        //         "order_id" => 12345,
        //         "quantity" => 1,
        //         "price" => 100,
        //         "amount" => 100,
        //         "exec_type" => "taker",
        //         "commission_amount" => "0.02",
        //         "commission_currency" => "BTC",
        //         "commission_percent" => "0.2"
        //     }
        //
        $timestamp = $this->safe_timestamp($trade, 'date');
        $symbol = null;
        $id = $this->safe_string($trade, 'trade_id');
        $orderId = $this->safe_string($trade, 'order_id');
        $priceString = $this->safe_string($trade, 'price');
        $amountString = $this->safe_string($trade, 'quantity');
        $costString = $this->safe_string($trade, 'amount');
        $side = $this->safe_string($trade, 'type');
        $type = null;
        $marketId = $this->safe_string($trade, 'pair');
        if ($marketId !== null) {
            if (is_array($this->markets_by_id) && array_key_exists($marketId, $this->markets_by_id)) {
                $market = $this->markets_by_id[$marketId];
            } else {
                list($baseId, $quoteId) = explode('_', $marketId);
                $base = $this->safe_currency_code($baseId);
                $quote = $this->safe_currency_code($quoteId);
                $symbol = $base . '/' . $quote;
            }
        }
        if (($symbol === null) && ($market !== null)) {
            $symbol = $market['symbol'];
        }
        $takerOrMaker = $this->safe_string($trade, 'exec_type');
        $fee = null;
        $feeCostString = $this->safe_string($trade, 'commission_amount');
        if ($feeCostString !== null) {
            $feeCurrencyId = $this->safe_string($trade, 'commission_currency');
            $feeCurrencyCode = $this->safe_currency_code($feeCurrencyId);
            $feeRateString = $this->safe_string($trade, 'commission_percent');
            if ($feeRateString !== null) {
                $feeRateString = Precise::string_div($feeRateString, '1000', 18);
            }
            $fee = array(
                'cost' => $feeCostString,
                'currency' => $feeCurrencyCode,
                'rate' => $feeRateString,
            );
        }
        return $this->safe_trade(array(
            'id' => $id,
            'info' => $trade,
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'symbol' => $symbol,
            'order' => $orderId,
            'type' => $type,
            'side' => $side,
            'takerOrMaker' => $takerOrMaker,
            'price' => $priceString,
            'amount' => $amountString,
            'cost' => $costString,
            'fee' => $fee,
        ), $market);
    }

    public function fetch_trades($symbol, $since = null, $limit = null, $params = array ()) {
        /**
         * get the list of most recent trades for a particular $symbol
         * @param {str} $symbol unified $symbol of the $market to fetch trades for
         * @param {int|null} $since timestamp in ms of the earliest trade to fetch
         * @param {int|null} $limit the maximum amount of trades to fetch
         * @param {dict} $params extra parameters specific to the exmo api endpoint
         * @return {[dict]} a list of ~@link https://docs.ccxt.com/en/latest/manual.html?#public-trades trade structures~
         */
        $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'pair' => $market['id'],
        );
        $response = $this->publicGetTrades (array_merge($request, $params));
        //
        //     {
        //         "ETH_BTC":array(
        //             array(
        //                 "trade_id":165087520,
        //                 "date":1587470005,
        //                 "type":"buy",
        //                 "quantity":"1.004",
        //                 "price":"0.02491461",
        //                 "amount":"0.02501426"
        //             ),
        //             {
        //                 "trade_id":165087369,
        //                 "date":1587469938,
        //                 "type":"buy",
        //                 "quantity":"0.94",
        //                 "price":"0.02492348",
        //                 "amount":"0.02342807"
        //             }
        //         )
        //     }
        //
        $data = $this->safe_value($response, $market['id'], array());
        return $this->parse_trades($data, $market, $since, $limit);
    }

    public function fetch_my_trades($symbol = null, $since = null, $limit = null, $params = array ()) {
        // a $symbol is required but it can be a single string, or a non-empty array
        if ($symbol === null) {
            throw new ArgumentsRequired($this->id . ' fetchMyTrades() requires a $symbol argument (a single $symbol or an array)');
        }
        $this->load_markets();
        $pair = null;
        $market = null;
        if (gettype($symbol) === 'array' && count(array_filter(array_keys($symbol), 'is_string')) == 0) {
            $numSymbols = is_array($symbol) ? count($symbol) : 0;
            if ($numSymbols < 1) {
                throw new ArgumentsRequired($this->id . ' fetchMyTrades() requires a non-empty $symbol array');
            }
            $marketIds = $this->market_ids($symbol);
            $pair = implode(',', $marketIds);
        } else {
            $market = $this->market($symbol);
            $pair = $market['id'];
        }
        $request = array(
            'pair' => $pair,
        );
        if ($limit !== null) {
            $request['limit'] = $limit;
        }
        $response = $this->privatePostUserTrades (array_merge($request, $params));
        $result = array();
        $marketIds = is_array($response) ? array_keys($response) : array();
        for ($i = 0; $i < count($marketIds); $i++) {
            $marketId = $marketIds[$i];
            $symbol = null;
            if (is_array($this->markets_by_id) && array_key_exists($marketId, $this->markets_by_id)) {
                $market = $this->markets_by_id[$marketId];
                $symbol = $market['symbol'];
            } else {
                list($baseId, $quoteId) = explode('_', $marketId);
                $base = $this->safe_currency_code($baseId);
                $quote = $this->safe_currency_code($quoteId);
                $symbol = $base . '/' . $quote;
            }
            $items = $response[$marketId];
            $trades = $this->parse_trades($items, $market, $since, $limit, array(
                'symbol' => $symbol,
            ));
            $result = $this->array_concat($result, $trades);
        }
        return $this->filter_by_since_limit($result, $since, $limit);
    }

    public function create_order($symbol, $type, $side, $amount, $price = null, $params = array ()) {
        $this->load_markets();
        $market = $this->market($symbol);
        $prefix = ($type === 'market') ? ($type . '_') : '';
        $orderType = $prefix . $side;
        $orderPrice = $price;
        if (($type === 'market') && ($price === null)) {
            $orderPrice = 0;
        }
        $request = array(
            'pair' => $market['id'],
            // 'leverage' => 2,
            'quantity' => $this->amount_to_precision($symbol, $amount),
            // spot - buy, sell, market_buy, market_sell, market_buy_total, market_sell_total
            // margin - limit_buy, limit_sell, market_buy, market_sell, stop_buy, stop_sell, stop_limit_buy, stop_limit_sell, trailing_stop_buy, trailing_stop_sell
            'type' => $orderType,
            'price' => $this->price_to_precision($symbol, $orderPrice),
            // 'stop_price' => $this->price_to_precision($symbol, $stopPrice),
            // 'distance' => 0, // distance for trailing stop orders
            // 'expire' => 0, // expiration $timestamp in UTC timezone for the order, unless expire is 0
            // 'client_id' => 123, // optional, must be a positive integer
            // 'comment' => '', // up to 50 latin symbols, whitespaces, underscores
        );
        $method = 'privatePostOrderCreate';
        $clientOrderId = $this->safe_value_2($params, 'client_id', 'clientOrderId');
        if ($clientOrderId !== null) {
            $clientOrderId = $this->safe_integer_2($params, 'client_id', 'clientOrderId');
            if ($clientOrderId === null) {
                throw new BadRequest($this->id . ' createOrder() client order $id must be an integer / numeric literal');
            } else {
                $request['client_id'] = $clientOrderId;
            }
            $params = $this->omit($params, array( 'client_id', 'clientOrderId' ));
        }
        if (($type === 'stop') || ($type === 'stop_limit') || ($type === 'trailing_stop')) {
            $stopPrice = $this->safe_number_2($params, 'stop_price', 'stopPrice');
            if ($stopPrice === null) {
                throw new InvalidOrder($this->id . ' createOrder() requires a $stopPrice extra param for a ' . $type . ' order');
            } else {
                $params = $this->omit($params, array( 'stopPrice', 'stop_price' ));
                $request['stop_price'] = $this->price_to_precision($symbol, $stopPrice);
                $method = 'privatePostMarginUserOrderCreate';
            }
        }
        $response = $this->$method (array_merge($request, $params));
        $id = $this->safe_string($response, 'order_id');
        $timestamp = $this->milliseconds();
        $status = 'open';
        return array(
            'id' => $id,
            'info' => $response,
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'lastTradeTimestamp' => null,
            'status' => $status,
            'symbol' => $symbol,
            'type' => $type,
            'side' => $side,
            'price' => $price,
            'cost' => null,
            'amount' => $amount,
            'remaining' => $amount,
            'filled' => 0.0,
            'fee' => null,
            'trades' => null,
            'clientOrderId' => $clientOrderId,
            'average' => null,
        );
    }

    public function cancel_order($id, $symbol = null, $params = array ()) {
        $this->load_markets();
        $request = array( 'order_id' => $id );
        return $this->privatePostOrderCancel (array_merge($request, $params));
    }

    public function fetch_order($id, $symbol = null, $params = array ()) {
        $this->load_markets();
        $request = array(
            'order_id' => (string) $id,
        );
        $response = $this->privatePostOrderTrades (array_merge($request, $params));
        //
        //     {
        //         "type" => "buy",
        //         "in_currency" => "BTC",
        //         "in_amount" => "1",
        //         "out_currency" => "USD",
        //         "out_amount" => "100",
        //         "trades" => array(
        //             {
        //                 "trade_id" => 3,
        //                 "date" => 1435488248,
        //                 "type" => "buy",
        //                 "pair" => "BTC_USD",
        //                 "order_id" => 12345,
        //                 "quantity" => 1,
        //                 "price" => 100,
        //                 "amount" => 100
        //             }
        //         )
        //     }
        //
        $order = $this->parse_order($response);
        return array_merge($order, array(
            'id' => (string) $id,
        ));
    }

    public function fetch_order_trades($id, $symbol = null, $since = null, $limit = null, $params = array ()) {
        $market = null;
        if ($symbol !== null) {
            $market = $this->market($symbol);
        }
        $request = array(
            'order_id' => (string) $id,
        );
        $response = $this->privatePostOrderTrades (array_merge($request, $params));
        //
        //     {
        //         "type" => "buy",
        //         "in_currency" => "BTC",
        //         "in_amount" => "1",
        //         "out_currency" => "USD",
        //         "out_amount" => "100",
        //         "trades" => array(
        //             {
        //                 "trade_id" => 3,
        //                 "date" => 1435488248,
        //                 "type" => "buy",
        //                 "pair" => "BTC_USD",
        //                 "order_id" => 12345,
        //                 "quantity" => 1,
        //                 "price" => 100,
        //                 "amount" => 100,
        //                 "exec_type" => "taker",
        //                 "commission_amount" => "0.02",
        //                 "commission_currency" => "BTC",
        //                 "commission_percent" => "0.2"
        //             }
        //         )
        //     }
        //
        $trades = $this->safe_value($response, 'trades');
        return $this->parse_trades($trades, $market, $since, $limit);
    }

    public function fetch_open_orders($symbol = null, $since = null, $limit = null, $params = array ()) {
        $this->load_markets();
        if ($symbol !== null) {
            $market = $this->market($symbol);
            $symbol = $market['symbol'];
        }
        $response = $this->privatePostUserOpenOrders ($params);
        $marketIds = is_array($response) ? array_keys($response) : array();
        $orders = array();
        for ($i = 0; $i < count($marketIds); $i++) {
            $marketId = $marketIds[$i];
            $market = null;
            if (is_array($this->markets_by_id) && array_key_exists($marketId, $this->markets_by_id)) {
                $market = $this->markets_by_id[$marketId];
            }
            $parsedOrders = $this->parse_orders($response[$marketId], $market);
            $orders = $this->array_concat($orders, $parsedOrders);
        }
        return $this->filter_by_symbol_since_limit($orders, $symbol, $since, $limit);
    }

    public function parse_order($order, $market = null) {
        //
        // fetchOrders, fetchOpenOrders, fetchClosedOrders, fetchCanceledOrders
        //
        //     {
        //         "order_id" => "14",
        //         "created" => "1435517311",
        //         "type" => "buy",
        //         "pair" => "BTC_USD",
        //         "price" => "100",
        //         "quantity" => "1",
        //         "amount" => "100"
        //     }
        //
        // fetchOrder
        //
        //     {
        //         "type" => "buy",
        //         "in_currency" => "BTC",
        //         "in_amount" => "1",
        //         "out_currency" => "USD",
        //         "out_amount" => "100",
        //         "trades" => array(
        //             {
        //                 "trade_id" => 3,
        //                 "date" => 1435488248,
        //                 "type" => "buy",
        //                 "pair" => "BTC_USD",
        //                 "order_id" => 12345,
        //                 "quantity" => 1,
        //                 "price" => 100,
        //                 "amount" => 100
        //             }
        //         )
        //     }
        //
        $id = $this->safe_string($order, 'order_id');
        $timestamp = $this->safe_timestamp($order, 'created');
        $symbol = null;
        $side = $this->safe_string($order, 'type');
        $marketId = null;
        if (is_array($order) && array_key_exists('pair', $order)) {
            $marketId = $order['pair'];
        } elseif ((is_array($order) && array_key_exists('in_currency', $order)) && (is_array($order) && array_key_exists('out_currency', $order))) {
            if ($side === 'buy') {
                $marketId = $order['in_currency'] . '_' . $order['out_currency'];
            } else {
                $marketId = $order['out_currency'] . '_' . $order['in_currency'];
            }
        }
        if (($marketId !== null) && (is_array($this->markets_by_id) && array_key_exists($marketId, $this->markets_by_id))) {
            $market = $this->markets_by_id[$marketId];
        }
        $amount = $this->safe_number($order, 'quantity');
        if ($amount === null) {
            $amountField = ($side === 'buy') ? 'in_amount' : 'out_amount';
            $amount = $this->safe_number($order, $amountField);
        }
        $price = $this->safe_number($order, 'price');
        $cost = $this->safe_number($order, 'amount');
        $filled = 0.0;
        $trades = array();
        $transactions = $this->safe_value($order, 'trades', array());
        $feeCost = null;
        $lastTradeTimestamp = null;
        $average = null;
        $numTransactions = is_array($transactions) ? count($transactions) : 0;
        if ($numTransactions > 0) {
            $feeCost = 0;
            for ($i = 0; $i < $numTransactions; $i++) {
                $trade = $this->parse_trade($transactions[$i], $market);
                if ($id === null) {
                    $id = $trade['order'];
                }
                if ($timestamp === null) {
                    $timestamp = $trade['timestamp'];
                }
                if ($timestamp > $trade['timestamp']) {
                    $timestamp = $trade['timestamp'];
                }
                $filled = $this->sum($filled, $trade['amount']);
                $feeCost = $this->sum($feeCost, $trade['fee']['cost']);
                $trades[] = $trade;
            }
            $lastTradeTimestamp = $trades[$numTransactions - 1]['timestamp'];
        }
        $status = $this->safe_string($order, 'status'); // in case we need to redefine it for canceled orders
        $remaining = null;
        if ($amount !== null) {
            $remaining = $amount - $filled;
            if ($filled >= $amount) {
                $status = 'closed';
            } else {
                $status = 'open';
            }
        }
        if ($market === null) {
            $market = $this->get_market_from_trades($trades);
        }
        $feeCurrency = null;
        if ($market !== null) {
            $symbol = $market['symbol'];
            $feeCurrency = $market['quote'];
        }
        if ($cost === null) {
            if ($price !== null) {
                $cost = $price * $filled;
            }
        } else {
            if ($filled > 0) {
                if ($average === null) {
                    $average = $cost / $filled;
                }
                if ($price === null) {
                    $price = $cost / $filled;
                }
            }
        }
        $fee = array(
            'cost' => $feeCost,
            'currency' => $feeCurrency,
        );
        $clientOrderId = $this->safe_integer($order, 'client_id');
        return array(
            'id' => $id,
            'clientOrderId' => $clientOrderId,
            'datetime' => $this->iso8601($timestamp),
            'timestamp' => $timestamp,
            'lastTradeTimestamp' => $lastTradeTimestamp,
            'status' => $status,
            'symbol' => $symbol,
            'type' => 'limit',
            'timeInForce' => null,
            'postOnly' => null,
            'side' => $side,
            'price' => $price,
            'stopPrice' => null,
            'cost' => $cost,
            'amount' => $amount,
            'filled' => $filled,
            'remaining' => $remaining,
            'average' => $average,
            'trades' => $trades,
            'fee' => $fee,
            'info' => $order,
        );
    }

    public function fetch_canceled_orders($symbol = null, $since = null, $limit = null, $params = array ()) {
        $this->load_markets();
        $request = array();
        if ($since !== null) {
            $request['offset'] = $limit;
        }
        if ($limit !== null) {
            $request['limit'] = $limit;
        }
        $market = null;
        if ($symbol !== null) {
            $market = $this->market($symbol);
        }
        $response = $this->privatePostUserCancelledOrders (array_merge($request, $params));
        //
        //     [array(
        //         "order_id" => "27056153840",
        //         "client_id" => "0",
        //         "created" => "1653428646",
        //         "type" => "buy",
        //         "pair" => "BTC_USDT",
        //         "quantity" => "0.1",
        //         "price" => "10",
        //         "amount" => "1"
        //     )]
        //
        return $this->parse_orders($response, $market, $since, $limit, $params);
    }

    public function fetch_deposit_address($code, $params = array ()) {
        $this->load_markets();
        $response = $this->privatePostDepositAddress ($params);
        $depositAddress = $this->safe_string($response, $code);
        $address = null;
        $tag = null;
        if ($depositAddress) {
            $addressAndTag = explode(',', $depositAddress);
            $address = $addressAndTag[0];
            $numParts = is_array($addressAndTag) ? count($addressAndTag) : 0;
            if ($numParts > 1) {
                $tag = $addressAndTag[1];
            }
        }
        $this->check_address($address);
        return array(
            'currency' => $code,
            'address' => $address,
            'tag' => $tag,
            'network' => null,
            'info' => $response,
        );
    }

    public function get_market_from_trades($trades) {
        $tradesBySymbol = $this->index_by($trades, 'pair');
        $symbols = is_array($tradesBySymbol) ? array_keys($tradesBySymbol) : array();
        $numSymbols = is_array($symbols) ? count($symbols) : 0;
        if ($numSymbols === 1) {
            return $this->markets[$symbols[0]];
        }
        return null;
    }

    public function withdraw($code, $amount, $address, $tag = null, $params = array ()) {
        list($tag, $params) = $this->handle_withdraw_tag_and_params($tag, $params);
        $this->load_markets();
        $currency = $this->currency($code);
        $request = array(
            'amount' => $amount,
            'currency' => $currency['id'],
            'address' => $address,
        );
        if ($tag !== null) {
            $request['invoice'] = $tag;
        }
        $networks = $this->safe_value($this->options, 'networks', array());
        $network = $this->safe_string_upper($params, 'network'); // this line allows the user to specify either ERC20 or ETH
        $network = $this->safe_string($networks, $network, $network); // handle ERC20>ETH alias
        if ($network !== null) {
            $request['transport'] = $network;
            $params = $this->omit($params, 'network');
        }
        $response = $this->privatePostWithdrawCrypt (array_merge($request, $params));
        return $this->parse_transaction($response, $currency);
    }

    public function parse_transaction_status($status) {
        $statuses = array(
            'transferred' => 'ok',
            'paid' => 'ok',
            'pending' => 'pending',
            'processing' => 'pending',
            'verifying' => 'pending',
        );
        return $this->safe_string($statuses, $status, $status);
    }

    public function parse_transaction($transaction, $currency = null) {
        //
        // fetchTransactions
        //
        //          {
        //            "dt" => 1461841192,
        //            "type" => "deposit",
        //            "curr" => "RUB",
        //            "status" => "processing",
        //            "provider" => "Qiwi (LA) [12345]",
        //            "amount" => "1",
        //            "account" => "",
        //            "txid" => "ec46f784ad976fd7f7539089d1a129fe46...",
        //          }
        //
        // fetchWithdrawals
        //
        //          array(
        //             "operation_id" => 47412538520634344,
        //             "created" => 1573760013,
        //             "updated" => 1573760013,
        //             "type" => "withdraw",
        //             "currency" => "DOGE",
        //             "status" => "Paid",
        //             "amount" => "300",
        //             "provider" => "DOGE",
        //             "commission" => "0",
        //             "account" => "DOGE => DBVy8pF1f8yxaCVEHqHeR7kkcHecLQ8nRS",
        //             "order_id" => 69670170,
        //             "provider_type" => "crypto",
        //             "crypto_address" => "DBVy8pF1f8yxaCVEHqHeR7kkcHecLQ8nRS",
        //             "card_number" => "",
        //             "wallet_address" => "",
        //             "email" => "",
        //             "phone" => "",
        //             "extra" => array(
        //                 "txid" => "f2b66259ae1580f371d38dd27e31a23fff8c04122b65ee3ab5a3f612d579c792",
        //                 "confirmations" => null,
        //                 "excode" => "",
        //                 "invoice" => ""
        //             ),
        //             "error" => ""
        //          ),
        //
        $id = $this->safe_string_2($transaction, 'order_id', 'task_id');
        $timestamp = $this->safe_timestamp_2($transaction, 'dt', 'created');
        $updated = $this->safe_timestamp($transaction, 'updated');
        $amount = $this->safe_number($transaction, 'amount');
        if ($amount !== null) {
            $amount = abs($amount);
        }
        $status = $this->parse_transaction_status($this->safe_string_lower($transaction, 'status'));
        $txid = $this->safe_string($transaction, 'txid');
        if ($txid === null) {
            $extra = $this->safe_value($transaction, 'extra', array());
            $extraTxid = $this->safe_string($extra, 'txid');
            if ($extraTxid !== '') {
                $txid = $extraTxid;
            }
        }
        $type = $this->safe_string($transaction, 'type');
        $currencyId = $this->safe_string_2($transaction, 'curr', 'currency');
        $code = $this->safe_currency_code($currencyId, $currency);
        $address = null;
        $tag = null;
        $comment = null;
        $account = $this->safe_string($transaction, 'account');
        if ($type === 'deposit') {
            $comment = $account;
        } elseif ($type === 'withdrawal') {
            $address = $account;
            if ($address !== null) {
                $parts = explode(':', $address);
                $numParts = is_array($parts) ? count($parts) : 0;
                if ($numParts === 2) {
                    $address = $this->safe_string($parts, 1);
                    $address = str_replace(' ', '', $address);
                }
            }
        }
        $fee = null;
        // fixed funding fees only (for now)
        if (!$this->fees['funding']['percentage']) {
            $key = ($type === 'withdrawal') ? 'withdraw' : 'deposit';
            $feeCost = $this->safe_number($transaction, 'commission');
            if ($feeCost === null) {
                $feeCost = $this->safe_number($this->options['fundingFees'][$key], $code);
            }
            // users don't pay for cashbacks, no fees for that
            $provider = $this->safe_string($transaction, 'provider');
            if ($provider === 'cashback') {
                $feeCost = 0;
            }
            if ($feeCost !== null) {
                // withdrawal $amount includes the $fee
                if ($type === 'withdrawal') {
                    $amount = $amount - $feeCost;
                }
                $fee = array(
                    'cost' => $feeCost,
                    'currency' => $code,
                    'rate' => null,
                );
            }
        }
        $network = $this->safe_string($transaction, 'provider');
        return array(
            'info' => $transaction,
            'id' => $id,
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'currency' => $code,
            'amount' => $amount,
            'network' => $network,
            'address' => $address,
            'addressTo' => $address,
            'addressFrom' => null,
            'tag' => $tag,
            'tagTo' => $tag,
            'tagFrom' => null,
            'status' => $status,
            'type' => $type,
            'updated' => $updated,
            'comment' => $comment,
            'txid' => $txid,
            'fee' => $fee,
        );
    }

    public function fetch_transactions($code = null, $since = null, $limit = null, $params = array ()) {
        $this->load_markets();
        $request = array();
        if ($since !== null) {
            $request['date'] = intval($since / 1000);
        }
        $currency = null;
        if ($code !== null) {
            $currency = $this->currency($code);
        }
        $response = $this->privatePostWalletHistory (array_merge($request, $params));
        //
        //     {
        //       "result" => true,
        //       "error" => "",
        //       "begin" => "1493942400",
        //       "end" => "1494028800",
        //       "history" => [
        //          array(
        //            "dt" => 1461841192,
        //            "type" => "deposit",
        //            "curr" => "RUB",
        //            "status" => "processing",
        //            "provider" => "Qiwi (LA) [12345]",
        //            "amount" => "1",
        //            "account" => "",
        //            "txid" => "ec46f784ad976fd7f7539089d1a129fe46...",
        //          ),
        //          array(
        //            "dt" => 1463414785,
        //            "type" => "withdrawal",
        //            "curr" => "USD",
        //            "status" => "paid",
        //            "provider" => "EXCODE",
        //            "amount" => "-1",
        //            "account" => "EX-CODE_19371_USDda...",
        //            "txid" => "",
        //          ),
        //       ],
        //     }
        //
        return $this->parse_transactions($response['history'], $currency, $since, $limit);
    }

    public function fetch_withdrawals($code = null, $since = null, $limit = null, $params = array ()) {
        $this->load_markets();
        $currency = null;
        $request = array(
            'type' => 'withdraw',
        );
        if ($limit !== null) {
            $request['limit'] = $limit; // default => 100, maximum => 100
        }
        if ($code !== null) {
            $currency = $this->currency($code);
            $request['currency'] = $currency['id'];
        }
        $response = $this->privatePostWalletOperations (array_merge($request, $params));
        //
        //     {
        //         "items" => array(
        //         array(
        //             "operation_id" => 47412538520634344,
        //             "created" => 1573760013,
        //             "updated" => 1573760013,
        //             "type" => "withdraw",
        //             "currency" => "DOGE",
        //             "status" => "Paid",
        //             "amount" => "300",
        //             "provider" => "DOGE",
        //             "commission" => "0",
        //             "account" => "DOGE => DBVy8pF1f8yxaCVEHqHeR7kkcHecLQ8nRS",
        //             "order_id" => 69670170,
        //             "extra" => array(
        //                 "txid" => "f2b66259ae1580f371d38dd27e31a23fff8c04122b65ee3ab5a3f612d579c792",
        //                 "excode" => "",
        //                 "invoice" => ""
        //             ),
        //             "error" => ""
        //         ),
        //     ),
        //         "count" => 23
        //     }
        //
        $items = $this->safe_value($response, 'items', array());
        return $this->parse_transactions($items, $currency, $since, $limit);
    }

    public function fetch_withdrawal($id, $code = null, $params = array ()) {
        $this->load_markets();
        $currency = null;
        $request = array(
            'order_id' => $id,
            'type' => 'withdraw',
        );
        if ($code !== null) {
            $currency = $this->currency($code);
            $request['currency'] = $currency['id'];
        }
        $response = $this->privatePostWalletOperations (array_merge($request, $params));
        //
        //     {
        //         "items" => array(
        //         array(
        //             "operation_id" => 47412538520634344,
        //             "created" => 1573760013,
        //             "updated" => 1573760013,
        //             "type" => "deposit",
        //             "currency" => "DOGE",
        //             "status" => "Paid",
        //             "amount" => "300",
        //             "provider" => "DOGE",
        //             "commission" => "0",
        //             "account" => "DOGE => DBVy8pF1f8yxaCVEHqHeR7kkcHecLQ8nRS",
        //             "order_id" => 69670170,
        //             "extra" => array(
        //                 "txid" => "f2b66259ae1580f371d38dd27e31a23fff8c04122b65ee3ab5a3f612d579c792",
        //                 "excode" => "",
        //                 "invoice" => ""
        //             ),
        //             "error" => ""
        //         ),
        //     ),
        //         "count" => 23
        //     }
        //
        $items = $this->safe_value($response, 'items', array());
        $first = $this->safe_value($items, 0, array());
        return $this->parse_transaction($first, $currency);
    }

    public function fetch_deposit($id = null, $code = null, $params = array ()) {
        $this->load_markets();
        $currency = null;
        $request = array(
            'order_id' => $id,
            'type' => 'deposit',
        );
        if ($code !== null) {
            $currency = $this->currency($code);
            $request['currency'] = $currency['id'];
        }
        $response = $this->privatePostWalletOperations (array_merge($request, $params));
        //
        //     {
        //         "items" => array(
        //         array(
        //             "operation_id" => 47412538520634344,
        //             "created" => 1573760013,
        //             "updated" => 1573760013,
        //             "type" => "deposit",
        //             "currency" => "DOGE",
        //             "status" => "Paid",
        //             "amount" => "300",
        //             "provider" => "DOGE",
        //             "commission" => "0",
        //             "account" => "DOGE => DBVy8pF1f8yxaCVEHqHeR7kkcHecLQ8nRS",
        //             "order_id" => 69670170,
        //             "extra" => array(
        //                 "txid" => "f2b66259ae1580f371d38dd27e31a23fff8c04122b65ee3ab5a3f612d579c792",
        //                 "excode" => "",
        //                 "invoice" => ""
        //             ),
        //             "error" => ""
        //         ),
        //     ),
        //         "count" => 23
        //     }
        //
        $items = $this->safe_value($response, 'items', array());
        $first = $this->safe_value($items, 0, array());
        return $this->parse_transaction($first, $currency);
    }

    public function fetch_deposits($code = null, $since = null, $limit = null, $params = array ()) {
        $this->load_markets();
        $currency = null;
        $request = array(
            'type' => 'deposit',
        );
        if ($limit !== null) {
            $request['limit'] = $limit; // default => 100, maximum => 100
        }
        if ($code !== null) {
            $currency = $this->currency($code);
            $request['currency'] = $currency['id'];
        }
        $response = $this->privatePostWalletOperations (array_merge($request, $params));
        //
        //     {
        //         "items" => array(
        //         array(
        //             "operation_id" => 47412538520634344,
        //             "created" => 1573760013,
        //             "updated" => 1573760013,
        //             "type" => "deposit",
        //             "currency" => "DOGE",
        //             "status" => "Paid",
        //             "amount" => "300",
        //             "provider" => "DOGE",
        //             "commission" => "0",
        //             "account" => "DOGE => DBVy8pF1f8yxaCVEHqHeR7kkcHecLQ8nRS",
        //             "order_id" => 69670170,
        //             "extra" => array(
        //                 "txid" => "f2b66259ae1580f371d38dd27e31a23fff8c04122b65ee3ab5a3f612d579c792",
        //                 "excode" => "",
        //                 "invoice" => ""
        //             ),
        //             "error" => ""
        //         ),
        //     ),
        //         "count" => 23
        //     }
        //
        $items = $this->safe_value($response, 'items', array());
        return $this->parse_transactions($items, $currency, $since, $limit);
    }

    public function sign($path, $api = 'public', $method = 'GET', $params = array (), $headers = null, $body = null) {
        $url = $this->urls['api'][$api] . '/';
        if ($api !== 'web') {
            $url .= $this->version . '/';
        }
        $url .= $path;
        if (($api === 'public') || ($api === 'web')) {
            if ($params) {
                $url .= '?' . $this->urlencode($params);
            }
        } elseif ($api === 'private') {
            $this->check_required_credentials();
            $nonce = $this->nonce();
            $body = $this->urlencode(array_merge(array( 'nonce' => $nonce ), $params));
            $headers = array(
                'Content-Type' => 'application/x-www-form-urlencoded',
                'Key' => $this->apiKey,
                'Sign' => $this->hmac($this->encode($body), $this->encode($this->secret), 'sha512'),
            );
        }
        return array( 'url' => $url, 'method' => $method, 'body' => $body, 'headers' => $headers );
    }

    public function nonce() {
        return $this->milliseconds();
    }

    public function handle_errors($httpCode, $reason, $url, $method, $headers, $body, $response, $requestHeaders, $requestBody) {
        if ($response === null) {
            return; // fallback to default error handler
        }
        if ((is_array($response) && array_key_exists('result', $response)) || (is_array($response) && array_key_exists('errmsg', $response))) {
            //
            //     array("result":false,"error":"Error 50052 => Insufficient funds")
            //     array("s":"error","errmsg":"strconv.ParseInt => parsing \"\" => invalid syntax")
            //
            $success = $this->safe_value($response, 'result', false);
            if (gettype($success) === 'string') {
                if (($success === 'true') || ($success === '1')) {
                    $success = true;
                } else {
                    $success = false;
                }
            }
            if (!$success) {
                $code = null;
                $message = $this->safe_string_2($response, 'error', 'errmsg');
                $errorParts = explode(':', $message);
                $numParts = is_array($errorParts) ? count($errorParts) : 0;
                if ($numParts > 1) {
                    $errorSubParts = explode(' ', $errorParts[0]);
                    $numSubParts = is_array($errorSubParts) ? count($errorSubParts) : 0;
                    $code = ($numSubParts > 1) ? $errorSubParts[1] : $errorSubParts[0];
                }
                $feedback = $this->id . ' ' . $body;
                $this->throw_exactly_matched_exception($this->exceptions['exact'], $code, $feedback);
                $this->throw_broadly_matched_exception($this->exceptions['broad'], $message, $feedback);
                throw new ExchangeError($feedback);
            }
        }
    }
}
