<?php

namespace ccxt;

// PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
// https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

use Exception; // a common import
use \ccxt\ExchangeError;
use \ccxt\ArgumentsRequired;
use \ccxt\BadRequest;
use \ccxt\BadResponse;
use \ccxt\InvalidOrder;

class ftx extends Exchange {

    public function describe() {
        return $this->deep_extend(parent::describe (), array(
            'id' => 'ftx',
            'name' => 'FTX',
            'countries' => array( 'BS' ), // Bahamas
            //  hard limit of 7 requests per 200ms => 35 requests per 1000ms => 1000ms / 35 = 28.5714 ms between requests
            // 10 withdrawal requests per 30 seconds = (1000ms / rateLimit) / (1/3) = 90.1
            // cancels do not count towards rateLimit
            // only 'order-making' requests count towards ratelimit
            'rateLimit' => 28.57,
            'certified' => true,
            'pro' => true,
            'hostname' => 'ftx.com', // or ftx.us
            'urls' => array(
                'logo' => 'https://user-images.githubusercontent.com/1294454/67149189-df896480-f2b0-11e9-8816-41593e17f9ec.jpg',
                'www' => 'https://ftx.com',
                'api' => array(
                    'public' => 'https://{hostname}',
                    'private' => 'https://{hostname}',
                ),
                'doc' => 'https://github.com/ftexchange/ftx',
                'fees' => 'https://ftexchange.zendesk.com/hc/en-us/articles/360024479432-Fees',
                'referral' => array(
                    'url' => 'https://ftx.com/#a=ccxt',
                    'discount' => 0.05,
                ),
            ),
            'has' => array(
                'CORS' => null,
                'spot' => true,
                'margin' => true,
                'swap' => true,
                'future' => true,
                'option' => false,
                'cancelAllOrders' => true,
                'cancelOrder' => true,
                'createOrder' => true,
                'createReduceOnlyOrder' => true,
                'createStopLimitOrder' => true,
                'createStopMarketOrder' => true,
                'createStopOrder' => true,
                'editOrder' => true,
                'fetchBalance' => true,
                'fetchBorrowInterest' => true,
                'fetchBorrowRate' => true,
                'fetchBorrowRateHistories' => true,
                'fetchBorrowRateHistory' => true,
                'fetchBorrowRates' => true,
                'fetchClosedOrders' => null,
                'fetchCurrencies' => true,
                'fetchDepositAddress' => true,
                'fetchDeposits' => true,
                'fetchFundingHistory' => true,
                'fetchFundingRate' => true,
                'fetchFundingRateHistory' => true,
                'fetchFundingRates' => false,
                'fetchIndexOHLCV' => true,
                'fetchLeverageTiers' => false,
                'fetchMarketLeverageTiers' => false,
                'fetchMarkets' => true,
                'fetchMarkOHLCV' => false,
                'fetchMyTrades' => true,
                'fetchOHLCV' => true,
                'fetchOpenInterestHistory' => false,
                'fetchOpenOrders' => true,
                'fetchOrder' => true,
                'fetchOrderBook' => true,
                'fetchOrders' => true,
                'fetchOrderTrades' => true,
                'fetchPosition' => false,
                'fetchPositions' => true,
                'fetchPositionsRisk' => false,
                'fetchPremiumIndexOHLCV' => false,
                'fetchTicker' => true,
                'fetchTickers' => true,
                'fetchTime' => false,
                'fetchTrades' => true,
                'fetchTradingFee' => false,
                'fetchTradingFees' => true,
                'fetchTransactionFees' => null,
                'fetchTransfer' => null,
                'fetchTransfers' => null,
                'fetchWithdrawals' => true,
                'reduceMargin' => false,
                'setLeverage' => true,
                'setMarginMode' => false, // FTX only supports cross margin
                'setPositionMode' => false,
                'transfer' => true,
                'withdraw' => true,
            ),
            'timeframes' => array(
                '15s' => '15',
                '1m' => '60',
                '5m' => '300',
                '15m' => '900',
                '1h' => '3600',
                '4h' => '14400',
                '1d' => '86400',
                '3d' => '259200',
                '1w' => '604800',
                '2w' => '1209600',
                // the exchange does not align candles to the start of the month
                // it can only fetch candles in fixed intervals of multiples of whole days
                // that works for all timeframes, except the monthly timeframe
                // because months have varying numbers of days
                '1M' => '2592000',
            ),
            'api' => array(
                'public' => array(
                    'get' => array(
                        'coins' => 1,
                        // markets
                        'markets' => 1,
                        'markets/{market_name}' => 1,
                        'markets/{market_name}/orderbook' => 1, // ?depth={depth}
                        'markets/{market_name}/trades' => 1, // ?limit={limit}&start_time={start_time}&end_time={end_time}
                        'markets/{market_name}/candles' => 1, // ?resolution={resolution}&limit={limit}&start_time={start_time}&end_time={end_time}
                        // futures
                        'futures' => 1,
                        'futures/{future_name}' => 1,
                        'futures/{future_name}/stats' => 1,
                        'funding_rates' => 1,
                        'indexes/{index_name}/weights' => 1,
                        'expired_futures' => 1,
                        'indexes/{market_name}/candles' => 1, // ?resolution={resolution}&limit={limit}&start_time={start_time}&end_time={end_time}
                        // wallet
                        'wallet/coins' => 1,
                        // leverage tokens
                        'lt/tokens' => 1,
                        'lt/{token_name}' => 1,
                        // etfs
                        'etfs/rebalance_info' => 1,
                        // options
                        'options/requests' => 1,
                        'options/trades' => 1,
                        'options/historical_volumes/BTC' => 1,
                        'stats/24h_options_volume' => 1,
                        'options/open_interest/BTC' => 1,
                        'options/historical_open_interest/BTC' => 1,
                        // spot margin
                        'spot_margin/history' => 1,
                        'spot_margin/borrow_summary' => 1,
                        // nfts
                        'nft/nfts' => 1,
                        'nft/{nft_id}' => 1,
                        'nft/{nft_id}/trades' => 1,
                        'nft/all_trades' => 1,
                        'nft/{nft_id}/account_info' => 1,
                        'nft/collections' => 1,
                        // ftx pay
                        'ftxpay/apps/{user_specific_id}/details' => 1,
                    ),
                    'post' => array(
                        'ftxpay/apps/{user_specific_id}/orders' => 1,
                    ),
                ),
                'private' => array(
                    'get' => array(
                        // subaccounts
                        'subaccounts' => 1,
                        'subaccounts/{nickname}/balances' => 1,
                        // account
                        'account' => 1,
                        'positions' => 1,
                        // wallet
                        'wallet/balances' => 1,
                        'wallet/all_balances' => 1,
                        'wallet/deposit_address/{coin}' => 1, // ?method={method}
                        'wallet/deposits' => 1,
                        'wallet/withdrawals' => 1,
                        'wallet/airdrops' => 1,
                        'wallet/withdrawal_fee' => 1,
                        'wallet/saved_addresses' => 1,
                        // orders
                        'orders' => 1, // ?market={market}
                        'orders/history' => 1, // ?market={market}
                        'orders/{order_id}' => 1,
                        'orders/by_client_id/{client_order_id}' => 1,
                        // conditional orders
                        'conditional_orders' => 1, // ?market={market}
                        'conditional_orders/{conditional_order_id}/triggers' => 1,
                        'conditional_orders/history' => 1, // ?market={market}
                        'fills' => 1, // ?market={market}
                        'funding_payments' => 1,
                        // leverage tokens
                        'lt/balances' => 1,
                        'lt/creations' => 1,
                        'lt/redemptions' => 1,
                        // options
                        'options/my_requests' => 1,
                        'options/requests/{request_id}/quotes' => 1,
                        'options/my_quotes' => 1,
                        'options/account_info' => 1,
                        'options/positions' => 1,
                        'options/fills' => 1,
                        // staking
                        'staking/stakes' => 1,
                        'staking/unstake_requests' => 1,
                        'staking/balances' => 1,
                        'staking/staking_rewards' => 1,
                        // otc
                        'otc/quotes/{quoteId}' => 1,
                        // spot margin
                        'spot_margin/borrow_rates' => 1,
                        'spot_margin/lending_rates' => 1,
                        'spot_margin/market_info' => 1, // ?market={market}
                        'spot_margin/borrow_history' => 1,
                        'spot_margin/lending_history' => 1,
                        'spot_margin/offers' => 1,
                        'spot_margin/lending_info' => 1,
                        // nfts
                        'nft/balances' => 1,
                        'nft/bids' => 1,
                        'nft/deposits' => 1,
                        'nft/withdrawals' => 1,
                        'nft/fills' => 1,
                        'nft/gallery/{gallery_id}' => 1,
                        'nft/gallery_settings' => 1,
                        // latency statistics
                        'stats/latency_stats' => 1,
                        // pnl
                        'pnl/historical_changes' => 1,
                    ),
                    'post' => array(
                        // subaccounts
                        'subaccounts' => 1,
                        'subaccounts/update_name' => 1,
                        'subaccounts/transfer' => 1,
                        // account
                        'account/leverage' => 1,
                        // wallet
                        'wallet/deposit_address/list' => 1,
                        'wallet/withdrawals' => 90,
                        'wallet/saved_addresses' => 1,
                        // orders
                        'orders' => 1,
                        'conditional_orders' => 1,
                        'orders/{order_id}/modify' => 1,
                        'orders/by_client_id/{client_order_id}/modify' => 1,
                        'conditional_orders/{order_id}/modify' => 1,
                        // leverage tokens
                        'lt/{token_name}/create' => 1,
                        'lt/{token_name}/redeem' => 1,
                        // options
                        'options/requests' => 1,
                        'options/requests/{request_id}/quotes' => 1,
                        'options/quotes/{quote_id}/accept' => 1,
                        // staking
                        'staking/unstake_requests' => 1,
                        'srm_stakes/stakes' => 1,
                        // otc
                        'otc/quotes/{quote_id}/accept' => 1,
                        'otc/quotes' => 1,
                        // spot margin
                        'spot_margin/offers' => 1,
                        // nfts
                        'nft/offer' => 1,
                        'nft/buy' => 1,
                        'nft/auction' => 1,
                        'nft/edit_auction' => 1,
                        'nft/cancel_auction' => 1,
                        'nft/bids' => 1,
                        'nft/redeem' => 1,
                        'nft/gallery_settings' => 1,
                        // ftx pay
                        'ftxpay/apps/{user_specific_id}/orders' => 1,
                    ),
                    'delete' => array(
                        // subaccounts
                        'subaccounts' => 1,
                        // wallet
                        'wallet/saved_addresses/{saved_address_id}' => 1,
                        // orders
                        'orders/{order_id}' => 1,
                        'orders/by_client_id/{client_order_id}' => 1,
                        'orders' => 1,
                        'conditional_orders/{order_id}' => 1,
                        // options
                        'options/requests/{request_id}' => 1,
                        'options/quotes/{quote_id}' => 1,
                        // staking
                        'staking/unstake_requests/{request_id}' => 1,
                    ),
                ),
            ),
            'fees' => array(
                'trading' => array(
                    'tierBased' => true,
                    'percentage' => true,
                    'maker' => $this->parse_number('0.0002'),
                    'taker' => $this->parse_number('0.0007'),
                    'tiers' => array(
                        'taker' => array(
                            array( $this->parse_number('0'), $this->parse_number('0.0007') ),
                            array( $this->parse_number('2000000'), $this->parse_number('0.0006') ),
                            array( $this->parse_number('5000000'), $this->parse_number('0.00055') ),
                            array( $this->parse_number('10000000'), $this->parse_number('0.0005') ),
                            array( $this->parse_number('25000000'), $this->parse_number('0.0045') ),
                            array( $this->parse_number('50000000'), $this->parse_number('0.0004') ),
                        ),
                        'maker' => array(
                            array( $this->parse_number('0'), $this->parse_number('0.0002') ),
                            array( $this->parse_number('2000000'), $this->parse_number('0.00015') ),
                            array( $this->parse_number('5000000'), $this->parse_number('0.0001') ),
                            array( $this->parse_number('10000000'), $this->parse_number('0.00005') ),
                            array( $this->parse_number('25000000'), $this->parse_number('0') ),
                            array( $this->parse_number('50000000'), $this->parse_number('0') ),
                        ),
                    ),
                ),
                'funding' => array(
                    'withdraw' => array(),
                ),
            ),
            'exceptions' => array(
                'exact' => array(
                    'Slow down' => '\\ccxt\\RateLimitExceeded', // array("error":"Slow down","success":false)
                    'Size too small for provide' => '\\ccxt\\InvalidOrder', // array("error":"Size too small for provide","success":false)
                    'Not enough balances' => '\\ccxt\\InsufficientFunds', // array("error":"Not enough balances","success":false)
                    'InvalidPrice' => '\\ccxt\\InvalidOrder', // array("error":"Invalid price","success":false)
                    'Size too small' => '\\ccxt\\InvalidOrder', // array("error":"Size too small","success":false)
                    'Size too large' => '\\ccxt\\InvalidOrder', // array("error":"Size too large","success":false)
                    'Invalid price' => '\\ccxt\\InvalidOrder', // array("success":false,"error":"Invalid price")
                    'Missing parameter price' => '\\ccxt\\InvalidOrder', // array("error":"Missing parameter price","success":false)
                    'Order not found' => '\\ccxt\\OrderNotFound', // array("error":"Order not found","success":false)
                    'Order already closed' => '\\ccxt\\InvalidOrder', // array("error":"Order already closed","success":false)
                    'Trigger price too high' => '\\ccxt\\InvalidOrder', // array("error":"Trigger price too high","success":false)
                    'Trigger price too low' => '\\ccxt\\InvalidOrder', // array("error":"Trigger price too low","success":false)
                    'Order already queued for cancellation' => '\\ccxt\\CancelPending', // array("error":"Order already queued for cancellation","success":false)
                    'Duplicate client order ID' => '\\ccxt\\DuplicateOrderId', // array("error":"Duplicate client order ID","success":false)
                    'Spot orders cannot be reduce-only' => '\\ccxt\\InvalidOrder', // array("error":"Spot orders cannot be reduce-only","success":false)
                    'Invalid reduce-only order' => '\\ccxt\\InvalidOrder', // array("error":"Invalid reduce-only order","success":false)
                    'Account does not have enough balances' => '\\ccxt\\InsufficientFunds', // array("success":false,"error":"Account does not have enough balances")
                    'Not authorized for subaccount-specific access' => '\\ccxt\\PermissionDenied', // array("success":false,"error":"Not authorized for subaccount-specific access")
                    'Not approved to trade this product' => '\\ccxt\\PermissionDenied', // array("success":false,"error":"Not approved to trade this product")
                    'Internal Error' => '\\ccxt\\ExchangeNotAvailable', // array("success":false,"error":"Internal Error")
                ),
                'broad' => array(
                    // array("error":"Not logged in","success":false)
                    // array("error":"Not logged in => Invalid API key","success":false)
                    'Not logged in' => '\\ccxt\\AuthenticationError',
                    'Account does not have enough margin for order' => '\\ccxt\\InsufficientFunds',
                    'Invalid parameter' => '\\ccxt\\BadRequest', // array("error":"Invalid parameter start_time","success":false)
                    'The requested URL was not found on the server' => '\\ccxt\\BadRequest',
                    'No such coin' => '\\ccxt\\BadRequest',
                    'No such subaccount' => '\\ccxt\\AuthenticationError',
                    'No such future' => '\\ccxt\\BadSymbol',
                    'No such market' => '\\ccxt\\BadSymbol',
                    'Do not send more than' => '\\ccxt\\RateLimitExceeded',
                    'Cannot send more than' => '\\ccxt\\RateLimitExceeded', // array("success":false,"error":"Cannot send more than 1500 requests per minute")
                    'An unexpected error occurred' => '\\ccxt\\ExchangeNotAvailable', // array("error":"An unexpected error occurred, please try again later (58BC21C795).","success":false)
                    'Please retry request' => '\\ccxt\\ExchangeNotAvailable', // array("error":"Please retry request","success":false)
                    'Please try again' => '\\ccxt\\ExchangeNotAvailable', // array("error":"Please try again","success":false)
                    'Try again' => '\\ccxt\\ExchangeNotAvailable', // array("error":"Try again","success":false)
                    'Only have permissions for subaccount' => '\\ccxt\\PermissionDenied', // array("success":false,"error":"Only have permissions for subaccount *sub_name*")
                ),
            ),
            'precisionMode' => TICK_SIZE,
            'options' => array(
                // support for canceling conditional orders
                // https://github.com/ccxt/ccxt/issues/6669
                'fetchMarkets' => array(
                    // the expiry datetime may be null for expiring futures, https://github.com/ccxt/ccxt/pull/12692
                    'throwOnUndefinedExpiry' => false,
                ),
                'cancelOrder' => array(
                    'method' => 'privateDeleteOrdersOrderId', // privateDeleteConditionalOrdersOrderId
                ),
                'fetchOpenOrders' => array(
                    'method' => 'privateGetOrders', // privateGetConditionalOrders
                ),
                'fetchOrders' => array(
                    'method' => 'privateGetOrdersHistory', // privateGetConditionalOrdersHistory
                ),
                'sign' => array(
                    'ftx.com' => 'FTX',
                    'ftx.us' => 'FTXUS',
                ),
                'networks' => array(
                    'SOL' => 'sol',
                    'SPL' => 'sol',
                    'TRX' => 'trx',
                    'TRC20' => 'trx',
                    'ETH' => 'erc20',
                    'ERC20' => 'erc20',
                    'OMNI' => 'omni',
                    'BEP2' => 'bep2',
                    'BNB' => 'bep2',
                    'BEP20' => 'bsc',
                    'BSC' => 'bsc',
                ),
            ),
            'commonCurrencies' => array(
                'AMC' => 'AMC Entertainment Holdings',
                'STARS' => 'StarLaunch',
            ),
        ));
    }

    public function fetch_currencies($params = array ()) {
        /**
         * fetches all available $currencies on an exchange
         * @param {dict} $params extra parameters specific to the ftx api endpoint
         * @return {dict} an associative dictionary of $currencies
         */
        $response = $this->publicGetCoins ($params);
        $currencies = $this->safe_value($response, 'result', array());
        //
        //     {
        //         "success":true,
        //         "result" => array(
        //             array("id":"BTC","name":"Bitcoin"),
        //             array("id":"ETH","name":"Ethereum"),
        //             array("id":"ETHMOON","name":"10X Long Ethereum Token","underlying":"ETH"),
        //             array("id":"EOSBULL","name":"3X Long EOS Token","underlying":"EOS"),
        //         ),
        //     }
        //
        $result = array();
        for ($i = 0; $i < count($currencies); $i++) {
            $currency = $currencies[$i];
            $id = $this->safe_string($currency, 'id');
            $code = $this->safe_currency_code($id);
            $name = $this->safe_string($currency, 'name');
            $result[$code] = array(
                'id' => $id,
                'code' => $code,
                'info' => $currency,
                'type' => null,
                'name' => $name,
                'active' => null,
                'deposit' => null,
                'withdraw' => null,
                'fee' => null,
                'precision' => null,
                'limits' => array(
                    'withdraw' => array( 'min' => null, 'max' => null ),
                    'amount' => array( 'min' => null, 'max' => null ),
                ),
            );
        }
        return $result;
    }

    public function fetch_markets($params = array ()) {
        /**
         * retrieves data on all $markets for ftx
         * @param {dict} $params extra parameters specific to the exchange api endpoint
         * @return {[dict]} an array of objects representing $market data
         */
        $response = $this->publicGetMarkets ($params);
        //
        //     {
        //         'success' => true,
        //         "result" => array(
        //             array(
        //                 "ask":170.37,
        //                 "baseCurrency":null,
        //                 "bid":170.31,
        //                 "change1h":-0.019001554672655036,
        //                 "change24h":-0.024841165359738997,
        //                 "changeBod":-0.03816406029469881,
        //                 "enabled":true,
        //                 "last":170.37,
        //                 "name":"ETH-PERP",
        //                 "price":170.37,
        //                 "priceIncrement":0.01,
        //                 "quoteCurrency":null,
        //                 "quoteVolume24h":7742164.59889,
        //                 "sizeIncrement":0.001,
        //                 "type":"future",
        //                 "underlying":"ETH",
        //                 "volumeUsd24h":7742164.59889
        //             ),
        //             array(
        //                 "ask":170.44,
        //                 "baseCurrency":"ETH",
        //                 "bid":170.41,
        //                 "change1h":-0.018485459257126403,
        //                 "change24h":-0.023825887743413515,
        //                 "changeBod":-0.037605872388481086,
        //                 "enabled":true,
        //                 "last":172.72,
        //                 "name":"ETH/USD",
        //                 "price":170.44,
        //                 "priceIncrement":0.01,
        //                 "quoteCurrency":"USD",
        //                 "quoteVolume24h":382802.0252,
        //                 "sizeIncrement":0.001,
        //                 "type":"spot",
        //                 "underlying":null,
        //                 "volumeUsd24h":382802.0252
        //             ),
        //         ),
        //     }
        //
        //     {
        //         name => "BTC-PERP",
        //         enabled =>  true,
        //         postOnly =>  false,
        //         priceIncrement => "1.0",
        //         $sizeIncrement => "0.0001",
        //         $minProvideSize => "0.001",
        //         last => "60397.0",
        //         bid => "60387.0",
        //         ask => "60388.0",
        //         price => "60388.0",
        //         $type => "future",
        //         baseCurrency =>  null,
        //         quoteCurrency =>  null,
        //         underlying => "BTC",
        //         restricted =>  false,
        //         highLeverageFeeExempt =>  true,
        //         change1h => "-0.0036463231533270636",
        //         change24h => "-0.01844838515677064",
        //         changeBod => "-0.010130151132675475",
        //         quoteVolume24h => "2892083192.6099",
        //         volumeUsd24h => "2892083192.6099"
        //     }
        //
        $allFuturesResponse = null;
        if ($this->has['future'] && ($this->hostname !== 'ftx.us')) {
            $allFuturesResponse = $this->publicGetFutures ();
        }
        //
        //    {
        //        success => true,
        //        $result => array(
        //            array(
        //                name => "1INCH-PERP",
        //                underlying => "1INCH",
        //                description => "1INCH Token Perpetual Futures",
        //                $type => "perpetual",
        //                $expiry => null,
        //                $perpetual => true,
        //                expired => false,
        //                enabled => true,
        //                postOnly => false,
        //                priceIncrement => "0.0001",
        //                $sizeIncrement => "1.0",
        //                last => "2.5556",
        //                bid => "2.5555",
        //                ask => "2.5563",
        //                index => "2.5612449804010833",
        //                mark => "2.5587",
        //                imfFactor => "0.0005",
        //                lowerBound => "2.4315",
        //                upperBound => "2.6893",
        //                underlyingDescription => "1INCH Token",
        //                expiryDescription => "Perpetual",
        //                moveStart => null,
        //                marginPrice => "2.5587",
        //                positionLimitWeight => "20.0",
        //                group => "perpetual",
        //                change1h => "0.00799716356760164",
        //                change24h => "0.004909276569004792",
        //                changeBod => "0.008394419484511705",
        //                volumeUsd24h => "17834492.0818",
        //                volume => "7224898.0",
        //                openInterest => "5597917.0",
        //                openInterestUsd => "14323390.2279",
        //            ),
        //            ...
        //        ),
        //    }
        //
        $result = array();
        $markets = $this->safe_value($response, 'result', array());
        $allFutures = $this->safe_value($allFuturesResponse, 'result', array());
        $allFuturesDict = $this->index_by($allFutures, 'name');
        for ($i = 0; $i < count($markets); $i++) {
            $market = $markets[$i];
            $id = $this->safe_string($market, 'name');
            $future = $this->safe_value($allFuturesDict, $id);
            $marketType = $this->safe_string($market, 'type');
            $contract = ($marketType === 'future');
            $baseId = $this->safe_string_2($market, 'baseCurrency', 'underlying');
            $quoteId = $this->safe_string($market, 'quoteCurrency', 'USD');
            $settleId = $contract ? 'USD' : null;
            $base = $this->safe_currency_code($baseId);
            $quote = $this->safe_currency_code($quoteId);
            $settle = $this->safe_currency_code($settleId);
            $spot = !$contract;
            $margin = !$contract;
            $perpetual = $this->safe_value($future, 'perpetual', false);
            $swap = $perpetual;
            $option = false;
            $isFuture = $contract && !$swap;
            $expiry = null;
            $expiryDatetime = $this->safe_string($future, 'expiry');
            $type = 'spot';
            $symbol = $base . '/' . $quote;
            if ($swap) {
                $type = 'swap';
                $symbol = $base . '/' . $quote . ':' . $settle;
            } elseif ($isFuture) {
                $type = 'future';
                $expiry = $this->parse8601($expiryDatetime);
                if ($expiry === null) {
                    // it is likely a $future that is expiring in this moment
                    $options = $this->safe_value($this->options, 'fetchMarkets', array());
                    $throwOnUndefinedExpiry = $this->safe_value($options, 'throwOnUndefinedExpiry', false);
                    if ($throwOnUndefinedExpiry) {
                        throw new BadResponse($this->id . " $symbol '" . $id . "' is a $future $contract with an invalid $expiry datetime.");
                    } else {
                        continue;
                    }
                }
                $parsedId = explode('-', $id);
                $length = is_array($parsedId) ? count($parsedId) : 0;
                if ($length > 2) {
                    // handling for MOVE contracts
                    // BTC-MOVE-2022Q1
                    // BTC-MOVE-0106
                    // BTC-MOVE-WK-0121
                    array_pop($parsedId);
                    // remove $expiry
                    // array( 'BTC', 'MOVE' )
                    // array( 'BTC', 'MOVE' )
                    // array( 'BTC', 'MOVE', 'WK' )
                    $base = implode('-', $parsedId);
                }
                $symbol = $base . '/' . $quote . ':' . $settle . '-' . $this->yymmdd($expiry, '');
            }
            // check if a $market is a $spot or $future $market
            $sizeIncrement = $this->safe_string($market, 'sizeIncrement');
            $minProvideSize = $this->safe_string($market, 'minProvideSize');
            $minAmountString = $sizeIncrement;
            if ($minProvideSize !== null) {
                $minAmountString = Precise::string_gt($minProvideSize, $sizeIncrement) ? $sizeIncrement : $minProvideSize;
            }
            $result[] = array(
                'id' => $id,
                'symbol' => $symbol,
                'base' => $base,
                'quote' => $quote,
                'settle' => $settle,
                'baseId' => $baseId,
                'quoteId' => $quoteId,
                'settleId' => $settleId,
                'type' => $type,
                'spot' => $spot,
                'margin' => $margin,
                'swap' => $swap,
                'future' => $isFuture,
                'option' => $option,
                'active' => $this->safe_value($market, 'enabled'),
                'contract' => $contract,
                'linear' => $contract ? true : null,
                'inverse' => $contract ? false : null,
                'contractSize' => $this->parse_number('1'),
                'expiry' => $expiry,
                'expiryDatetime' => $this->iso8601($expiry),
                'strike' => null,
                'optionType' => null,
                'precision' => array(
                    'amount' => $this->parse_number($sizeIncrement),
                    'price' => $this->safe_number($market, 'priceIncrement'),
                ),
                'limits' => array(
                    'leverage' => array(
                        'min' => $this->parse_number('1'),
                        'max' => $this->parse_number('20'),
                    ),
                    'amount' => array(
                        'min' => $this->parse_number($minAmountString),
                        'max' => null,
                    ),
                    'price' => array(
                        'min' => null,
                        'max' => null,
                    ),
                    'cost' => array(
                        'min' => null,
                        'max' => null,
                    ),
                ),
                'info' => $market,
            );
        }
        return $result;
    }

    public function parse_ticker($ticker, $market = null) {
        //
        //     {
        //         "ask":171.29,
        //         "baseCurrency":null, // base currency for spot markets
        //         "bid":171.24,
        //         "change1h":-0.0012244897959183673,
        //         "change24h":-0.031603346901854366,
        //         "changeBod":-0.03297013492914808,
        //         "enabled":true,
        //         "last":171.44,
        //         "name":"ETH-PERP",
        //         "price":171.29,
        //         "priceIncrement":0.01,
        //         "quoteCurrency":null, // quote currency for spot markets
        //         "quoteVolume24h":8570651.12113,
        //         "sizeIncrement":0.001,
        //         "type":"future",
        //         "underlying":"ETH", // null for spot markets
        //         "volumeUsd24h":8570651.12113,
        //     }
        //
        $marketId = $this->safe_string($ticker, 'name');
        if (is_array($this->markets_by_id) && array_key_exists($marketId, $this->markets_by_id)) {
            $market = $this->markets_by_id[$marketId];
        }
        $symbol = $this->safe_symbol($marketId, $market);
        $last = $this->safe_string($ticker, 'last');
        $timestamp = $this->safe_timestamp($ticker, 'time', $this->milliseconds());
        $percentage = $this->safe_string($ticker, 'change24h');
        if ($percentage !== null) {
            $percentage = Precise::string_mul($percentage, '100');
        }
        return $this->safe_ticker(array(
            'symbol' => $symbol,
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'high' => $this->safe_string($ticker, 'high'),
            'low' => $this->safe_string($ticker, 'low'),
            'bid' => $this->safe_string($ticker, 'bid'),
            'bidVolume' => $this->safe_string($ticker, 'bidSize'),
            'ask' => $this->safe_string($ticker, 'ask'),
            'askVolume' => $this->safe_string($ticker, 'askSize'),
            'vwap' => null,
            'open' => null,
            'close' => $last,
            'last' => $last,
            'previousClose' => null,
            'change' => null,
            'percentage' => $percentage,
            'average' => null,
            'baseVolume' => null,
            'quoteVolume' => $this->safe_string($ticker, 'quoteVolume24h'),
            'info' => $ticker,
        ), $market);
    }

    public function fetch_ticker($symbol, $params = array ()) {
        /**
         * fetches a price ticker, a statistical calculation with the information calculated over the past 24 hours for a specific $market
         * @param {str} $symbol unified $symbol of the $market to fetch the ticker for
         * @param {dict} $params extra parameters specific to the ftx api endpoint
         * @return {dict} a {@link https://docs.ccxt.com/en/latest/manual.html#ticker-structure ticker structure}
         */
        $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'market_name' => $market['id'],
        );
        $response = $this->publicGetMarketsMarketName (array_merge($request, $params));
        //
        //     {
        //         "success":true,
        //         "result":{
        //             "ask":171.29,
        //             "baseCurrency":null, // base currency for spot markets
        //             "bid":171.24,
        //             "change1h":-0.0012244897959183673,
        //             "change24h":-0.031603346901854366,
        //             "changeBod":-0.03297013492914808,
        //             "enabled":true,
        //             "last":171.44,
        //             "name":"ETH-PERP",
        //             "price":171.29,
        //             "priceIncrement":0.01,
        //             "quoteCurrency":null, // quote currency for spot markets
        //             "quoteVolume24h":8570651.12113,
        //             "sizeIncrement":0.001,
        //             "type":"future",
        //             "underlying":"ETH", // null for spot markets
        //             "volumeUsd24h":8570651.12113,
        //         }
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        return $this->parse_ticker($result, $market);
    }

    public function fetch_tickers($symbols = null, $params = array ()) {
        /**
         * fetches price $tickers for multiple markets, statistical calculations with the information calculated over the past 24 hours each market
         * @param {[str]|null} $symbols unified $symbols of the markets to fetch the ticker for, all market $tickers are returned if not assigned
         * @param {dict} $params extra parameters specific to the ftx api endpoint
         * @return {dict} an array of {@link https://docs.ccxt.com/en/latest/manual.html#ticker-structure ticker structures}
         */
        $this->load_markets();
        $symbols = $this->market_symbols($symbols);
        $response = $this->publicGetMarkets ($params);
        //
        //     {
        //         'success' => true,
        //         "result" => array(
        //             array(
        //                 "ask":170.44,
        //                 "baseCurrency":"ETH",
        //                 "bid":170.41,
        //                 "change1h":-0.018485459257126403,
        //                 "change24h":-0.023825887743413515,
        //                 "changeBod":-0.037605872388481086,
        //                 "enabled":true,
        //                 "last":172.72,
        //                 "name":"ETH/USD",
        //                 "price":170.44,
        //                 "priceIncrement":0.01,
        //                 "quoteCurrency":"USD",
        //                 "quoteVolume24h":382802.0252,
        //                 "sizeIncrement":0.001,
        //                 "type":"spot",
        //                 "underlying":null,
        //                 "volumeUsd24h":382802.0252
        //             ),
        //         ),
        //     }
        //
        $tickers = $this->safe_value($response, 'result', array());
        return $this->parse_tickers($tickers, $symbols);
    }

    public function fetch_order_book($symbol, $limit = null, $params = array ()) {
        /**
         * fetches information on open orders with bid (buy) and ask (sell) prices, volumes and other data
         * @param {str} $symbol unified $symbol of the $market to fetch the order book for
         * @param {int|null} $limit the maximum amount of order book entries to return
         * @param {dict} $params extra parameters specific to the ftx api endpoint
         * @return {dict} A dictionary of {@link https://docs.ccxt.com/en/latest/manual.html#order-book-structure order book structures} indexed by $market symbols
         */
        $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'market_name' => $market['id'],
        );
        if ($limit !== null) {
            $request['depth'] = $limit; // max 100, default 20
        }
        $response = $this->publicGetMarketsMarketNameOrderbook (array_merge($request, $params));
        //
        //     {
        //         "success":true,
        //         "result":{
        //             "asks":[
        //                 [171.95,279.865],
        //                 [171.98,102.42],
        //                 [171.99,124.11],
        //             ],
        //             "bids":[
        //                 [171.93,69.749],
        //                 [171.9,288.325],
        //                 [171.88,87.47],
        //             ],
        //         }
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        return $this->parse_order_book($result, $symbol);
    }

    public function parse_ohlcv($ohlcv, $market = null) {
        //
        //     {
        //         "close":177.23,
        //         "high":177.45,
        //         "low":177.2,
        //         "open":177.43,
        //         "startTime":"2019-10-17T13:27:00+00:00",
        //         "time":1571318820000.0,
        //         "volume":0.0
        //     }
        //
        return array(
            $this->safe_integer($ohlcv, 'time'),
            $this->safe_number($ohlcv, 'open'),
            $this->safe_number($ohlcv, 'high'),
            $this->safe_number($ohlcv, 'low'),
            $this->safe_number($ohlcv, 'close'),
            $this->safe_number($ohlcv, 'volume'),
        );
    }

    public function get_market_id($symbol, $key, $params = array ()) {
        $parts = $this->get_market_params($symbol, $key, $params);
        return $this->safe_string($parts, 1, $symbol);
    }

    public function get_market_params($symbol, $key, $params = array ()) {
        $market = null;
        $marketId = null;
        if (is_array($this->markets) && array_key_exists($symbol, $this->markets)) {
            $market = $this->market($symbol);
            $marketId = $market['id'];
        } else {
            $marketId = $this->safe_string($params, $key, $symbol);
        }
        return array( $market, $marketId );
    }

    public function fetch_ohlcv($symbol, $timeframe = '1m', $since = null, $limit = null, $params = array ()) {
        /**
         * fetches historical candlestick data containing the open, high, low, and close $price, and the volume of a $market
         * @param {str} $symbol unified $symbol of the $market to fetch OHLCV data for
         * @param {str} $timeframe the length of time each candle represents
         * @param {int|null} $since timestamp in ms of the earliest candle to fetch
         * @param {int|null} $limit the maximum amount of candles to fetch
         * @param {dict} $params extra parameters specific to the ftx api endpoint
         * @return {[[int]]} A list of candles ordered as timestamp, open, high, low, close, volume
         */
        $this->load_markets();
        list($market, $marketId) = $this->get_market_params($symbol, 'market_name', $params);
        // max 1501 candles, including the current candle when $since is not specified
        $maxLimit = 5000;
        $defaultLimit = 1500;
        $limit = ($limit === null) ? $defaultLimit : min ($limit, $maxLimit);
        $request = array(
            'resolution' => $this->timeframes[$timeframe],
            'market_name' => $marketId,
            // 'start_time' => intval($since / 1000),
            // 'end_time' => $this->seconds(),
            'limit' => $limit,
        );
        $price = $this->safe_string($params, 'price');
        $params = $this->omit($params, 'price');
        if ($since !== null) {
            $startTime = intval($since / 1000);
            $request['start_time'] = $startTime;
            $duration = $this->parse_timeframe($timeframe);
            $endTime = $this->sum($startTime, $limit * $duration);
            $request['end_time'] = min ($endTime, $this->seconds());
            if ($duration > 86400) {
                $wholeDaysInTimeframe = intval($duration / 86400);
                $request['limit'] = min ($limit * $wholeDaysInTimeframe, $maxLimit);
            }
        }
        $method = 'publicGetMarketsMarketNameCandles';
        if ($price === 'index') {
            if (is_array($this->markets) && array_key_exists($symbol, $this->markets)) {
                $request['market_name'] = $market['baseId'];
            }
            $method = 'publicGetIndexesMarketNameCandles';
        }
        $response = $this->$method (array_merge($request, $params));
        //
        //     {
        //         "success" => true,
        //         "result":array(
        //             array(
        //                 "close":177.23,
        //                 "high":177.45,
        //                 "low":177.2,
        //                 "open":177.43,
        //                 "startTime":"2019-10-17T13:27:00+00:00",
        //                 "time":1571318820000.0,
        //                 "volume":0.0
        //             ),
        //             array(
        //                 "close":177.26,
        //                 "high":177.33,
        //                 "low":177.23,
        //                 "open":177.23,
        //                 "startTime":"2019-10-17T13:28:00+00:00",
        //                 "time":1571318880000.0,
        //                 "volume":0.0
        //             ),
        //         ),
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        return $this->parse_ohlcvs($result, $market, $timeframe, $since, $limit);
    }

    public function parse_trade($trade, $market = null) {
        //
        // fetchTrades (public)
        //
        //     {
        //         "id":1715826,
        //         "liquidation":false,
        //         "price":171.62,
        //         "side":"buy",
        //         "size":2.095,
        //         "time":"2019-10-18T12:59:54.288166+00:00"
        //     }
        //
        // fetchMyTrades (private)
        //
        //     {
        //         "fee" => 20.1374935,
        //         "feeRate" => 0.0005,
        //         "feeCurrency" => "USD",
        //         "future" => "EOS-0329",
        //         "id" => 11215,
        //         "liquidity" => "taker",
        //         "market" => "EOS-0329",
        //         "baseCurrency" => null,
        //         "quoteCurrency" => null,
        //         "orderId" => 8436981,
        //         "price" => 4.201,
        //         "side" => "buy",
        //         "size" => 9587,
        //         "time" => "2019-03-27T19:15:10.204619+00:00",
        //         "type" => "order"
        //     }
        //
        //     {
        //         "baseCurrency" => "BTC",
        //         "fee" => 0,
        //         "feeCurrency" => "USD",
        //         "feeRate" => 0,
        //         "future" => null,
        //         "id" => 664079556,
        //         "liquidity" => "taker",
        //         "market" => null,
        //         "orderId" => null,
        //         "price" => 34830.61359,
        //         "quoteCurrency" => "USD",
        //         "side" => "sell",
        //         "size" => 0.0005996,
        //         "time" => "2021-01-15T16:05:29.246135+00:00",
        //         "tradeId" => null,
        //         "type" => "otc"
        //     }
        //
        //     with -ve $fee
        //     {
        //         "id" => 1171258927,
        //         "fee" => -0.0000713875,
        //         "side" => "sell",
        //         "size" => 1,
        //         "time" => "2021-03-11T13:34:35.523627+00:00",
        //         "type" => "order",
        //         "price" => 14.2775,
        //         "future" => null,
        //         "market" => "SOL/USD",
        //         "feeRate" => -0.000005,
        //         "orderId" => 33182929044,
        //         "tradeId" => 582936801,
        //         "liquidity" => "maker",
        //         "feeCurrency" => "USD",
        //         "baseCurrency" => "SOL",
        //         "quoteCurrency" => "USD"
        //     }
        //
        //     // from OTC order
        //     {
        //         "id" => 1172129651,
        //         "fee" => 0,
        //         "side" => "sell",
        //         "size" => 1.47568846,
        //         "time" => "2021-03-11T15:04:46.893383+00:00",
        //         "type" => "otc",
        //         "price" => 14.60932598,
        //         "future" => null,
        //         "market" => null,
        //         "feeRate" => 0,
        //         "orderId" => null,
        //         "tradeId" => null,
        //         "liquidity" => "taker",
        //         "feeCurrency" => "USD",
        //         "baseCurrency" => "BCHA",
        //         "quoteCurrency" => "USD"
        //     }
        //
        $id = $this->safe_string($trade, 'id');
        $takerOrMaker = $this->safe_string($trade, 'liquidity');
        // a workaround for the OTC trades, they don't have a $symbol
        $baseId = $this->safe_string($trade, 'baseCurrency');
        $quoteId = $this->safe_string($trade, 'quoteCurrency');
        $defaultMarketId = null;
        if (($baseId !== null) && ($quoteId !== null)) {
            $defaultMarketId = $baseId . '/' . $quoteId;
        }
        $marketId = $this->safe_string($trade, 'market', $defaultMarketId);
        $market = $this->safe_market($marketId, $market, '/');
        $symbol = $market['symbol'];
        $timestamp = $this->parse8601($this->safe_string($trade, 'time'));
        $priceString = $this->safe_string($trade, 'price');
        $amountString = $this->safe_string($trade, 'size');
        $side = $this->safe_string($trade, 'side');
        $fee = null;
        $feeCostString = $this->safe_string($trade, 'fee');
        if ($feeCostString !== null) {
            $feeCurrencyId = $this->safe_string($trade, 'feeCurrency');
            $feeCurrencyCode = $this->safe_currency_code($feeCurrencyId);
            $fee = array(
                'cost' => $feeCostString,
                'currency' => $feeCurrencyCode,
                'rate' => $this->safe_string($trade, 'feeRate'),
            );
        }
        $orderId = $this->safe_string($trade, 'orderId');
        return $this->safe_trade(array(
            'info' => $trade,
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'symbol' => $symbol,
            'id' => $id,
            'order' => $orderId,
            'type' => null,
            'takerOrMaker' => $takerOrMaker,
            'side' => $side,
            'price' => $priceString,
            'amount' => $amountString,
            'cost' => null,
            'fee' => $fee,
        ), $market);
    }

    public function fetch_trades($symbol, $since = null, $limit = null, $params = array ()) {
        /**
         * get the list of most recent trades for a particular $symbol
         * @param {str} $symbol unified $symbol of the $market to fetch trades for
         * @param {int|null} $since timestamp in ms of the earliest trade to fetch
         * @param {int|null} $limit the maximum amount of trades to fetch
         * @param {dict} $params extra parameters specific to the ftx api endpoint
         * @return {[dict]} a list of ~@link https://docs.ccxt.com/en/latest/manual.html?#public-trades trade structures~
         */
        $this->load_markets();
        list($market, $marketId) = $this->get_market_params($symbol, 'market_name', $params);
        $request = array(
            'market_name' => $marketId,
        );
        if ($since !== null) {
            // the exchange aligns results to end_time returning 5000 trades max
            // the user must set the end_time (in seconds) close enough to start_time
            // for a proper pagination, fetch the most recent trades first
            // then set the end_time parameter to the timestamp of the last trade
            // start_time and end_time must be in seconds, divided by a thousand
            $request['start_time'] = intval($since / 1000);
            // start_time doesn't work without end_time
            $request['end_time'] = $this->seconds();
        }
        if ($limit !== null) {
            $request['limit'] = $limit;
        }
        $response = $this->publicGetMarketsMarketNameTrades (array_merge($request, $params));
        //
        //     {
        //         "success":true,
        //         "result":array(
        //             array(
        //                 "id":1715826,
        //                 "liquidation":false,
        //                 "price":171.62,
        //                 "side":"buy",
        //                 "size":2.095,
        //                 "time":"2019-10-18T12:59:54.288166+00:00"
        //             ),
        //             array(
        //                 "id":1715763,
        //                 "liquidation":false,
        //                 "price":171.89,
        //                 "side":"sell",
        //                 "size":1.477,
        //                 "time":"2019-10-18T12:58:38.443734+00:00"
        //             ),
        //         ),
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        return $this->parse_trades($result, $market, $since, $limit);
    }

    public function fetch_trading_fees($params = array ()) {
        $this->load_markets();
        $response = $this->privateGetAccount ($params);
        //
        //     {
        //         "success" => true,
        //         "result" => array(
        //             "backstopProvider" => true,
        //             "collateral" => 3568181.02691129,
        //             "freeCollateral" => 1786071.456884368,
        //             "initialMarginRequirement" => 0.12222384240257728,
        //             "liquidating" => false,
        //             "maintenanceMarginRequirement" => 0.07177992558058484,
        //             "makerFee" => 0.0002,
        //             "marginFraction" => 0.5588433331419503,
        //             "openMarginFraction" => 0.2447194090423075,
        //             "takerFee" => 0.0005,
        //             "totalAccountValue" => 3568180.98341129,
        //             "totalPositionSize" => 6384939.6992,
        //             "username" => "user@domain.com",
        //             "positions" => array(
        //                 array(
        //                     "cost" => -31.7906,
        //                     "entryPrice" => 138.22,
        //                     "future" => "ETH-PERP",
        //                     "initialMarginRequirement" => 0.1,
        //                     "longOrderSize" => 1744.55,
        //                     "maintenanceMarginRequirement" => 0.04,
        //                     "netSize" => -0.23,
        //                     "openSize" => 1744.32,
        //                     "realizedPnl" => 3.39441714,
        //                     "shortOrderSize" => 1732.09,
        //                     "side" => "sell",
        //                     "size" => 0.23,
        //                     "unrealizedPnl" => 0,
        //                 ),
        //             ),
        //         ),
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        $maker = $this->safe_number($result, 'makerFee');
        $taker = $this->safe_number($result, 'takerFee');
        $tradingFees = array();
        for ($i = 0; $i < count($this->symbols); $i++) {
            $symbol = $this->symbols[$i];
            $tradingFees[$symbol] = array(
                'info' => $response,
                'symbol' => $symbol,
                'maker' => $maker,
                'taker' => $taker,
                'percentage' => true,
                'tierBased' => true,
            );
        }
        return $tradingFees;
    }

    public function fetch_funding_rate_history($symbol = null, $since = null, $limit = null, $params = array ()) {
        /**
         * fetches historical funding rate prices
         * @param {str|null} $symbol unified $symbol of the $market to fetch the funding rate history for
         * @param {int|null} $since $timestamp in ms of the earliest funding rate to fetch
         * @param {int|null} $limit the maximum amount of ~@link https://docs.ccxt.com/en/latest/manual.html?#funding-rate-history-structure funding rate structures~ to fetch
         * @param {dict} $params extra parameters specific to the ftx api endpoint
         * @param {int|null} $params->till extra parameters specific to the ftx api endpoint
         * @return {[dict]} a list of ~@link https://docs.ccxt.com/en/latest/manual.html?#funding-rate-history-structure funding rate structures~
         */
        $this->load_markets();
        $request = array();
        if ($symbol !== null) {
            $market = $this->market($symbol);
            $symbol = $market['symbol'];
            $request['future'] = $market['id'];
        }
        if ($since !== null) {
            $request['start_time'] = intval($since / 1000);
        }
        $till = $this->safe_integer($params, 'till'); // unified in milliseconds
        $endTime = $this->safe_string($params, 'end_time'); // exchange-specific in seconds
        $params = $this->omit($params, array( 'end_time', 'till' ));
        if ($till !== null) {
            $request['end_time'] = intval($till / 1000);
        } elseif ($endTime !== null) {
            $request['end_time'] = $endTime;
        }
        $response = $this->publicGetFundingRates (array_merge($request, $params));
        //
        //     {
        //        "success" => true,
        //        "result" => array(
        //          {
        //            "future" => "BTC-PERP",
        //            "rate" => 0.0025,
        //            "time" => "2019-06-02T08:00:00+00:00"
        //          }
        //        )
        //      }
        //
        $result = $this->safe_value($response, 'result', array());
        $rates = array();
        for ($i = 0; $i < count($result); $i++) {
            $entry = $result[$i];
            $marketId = $this->safe_string($entry, 'future');
            $timestamp = $this->parse8601($this->safe_string($entry, 'time'));
            $rates[] = array(
                'info' => $entry,
                'symbol' => $this->safe_symbol($marketId),
                'fundingRate' => $this->safe_number($entry, 'rate'),
                'timestamp' => $timestamp,
                'datetime' => $this->iso8601($timestamp),
            );
        }
        $sorted = $this->sort_by($rates, 'timestamp');
        return $this->filter_by_symbol_since_limit($sorted, $symbol, $since, $limit);
    }

    public function parse_balance($response) {
        $result = array(
            'info' => $response,
        );
        $balances = $this->safe_value($response, 'result', array());
        for ($i = 0; $i < count($balances); $i++) {
            $balance = $balances[$i];
            $code = $this->safe_currency_code($this->safe_string($balance, 'coin'));
            $account = $this->account();
            $account['free'] = $this->safe_string_2($balance, 'availableWithoutBorrow', 'free');
            $account['total'] = $this->safe_string($balance, 'total');
            $result[$code] = $account;
        }
        return $this->safe_balance($result);
    }

    public function fetch_balance($params = array ()) {
        /**
         * query for balance and get the amount of funds available for trading or funds locked in orders
         * @param {dict} $params extra parameters specific to the ftx api endpoint
         * @return {dict} a ~@link https://docs.ccxt.com/en/latest/manual.html?#balance-structure balance structure~
         */
        $this->load_markets();
        $response = $this->privateGetWalletBalances ($params);
        //
        //     {
        //         "success" => true,
        //         "result" => array(
        //             array(
        //                 "coin" => "USDTBEAR",
        //                 "free" => 2320.2,
        //                 "total" => 2340.2
        //             ),
        //         ),
        //     }
        //
        return $this->parse_balance($response);
    }

    public function parse_order_status($status) {
        $statuses = array(
            'new' => 'open',
            'open' => 'open',
            'closed' => 'closed', // filled or canceled
            'triggered' => 'closed',
        );
        return $this->safe_string($statuses, $status, $status);
    }

    public function parse_order($order, $market = null) {
        //
        // limit orders - fetchOrder, fetchOrders, fetchOpenOrders, createOrder, editOrder
        //
        //     {
        //         "createdAt" => "2019-03-05T09:56:55.728933+00:00",
        //         "filledSize" => 0,
        //         "future" => "XRP-PERP",
        //         "id" => 9596912,
        //         "market" => "XRP-PERP",
        //         "price" => 0.306525,
        //         "remainingSize" => 31431,
        //         "side" => "sell",
        //         "size" => 31431,
        //         "status" => "open",
        //         "type" => "limit",
        //         "reduceOnly" => false,
        //         "ioc" => false,
        //         "postOnly" => false,
        //         "clientId" => null,
        //     }
        //
        // $market orders - fetchOrder, fetchOrders, fetchOpenOrders, createOrder
        //
        //     {
        //         "avgFillPrice" => 2666.0,
        //         "clientId" => None,
        //         "createdAt" => "2020-02-12T00 => 53 => 49.009726+00 => 00",
        //         "filledSize" => 0.0007,
        //         "future" => None,
        //         "id" => 3109208514,
        //         "ioc" => True,
        //         "market" => "BNBBULL/USD",
        //         "postOnly" => False,
        //         "price" => None,
        //         "reduceOnly" => False,
        //         "remainingSize" => 0.0,
        //         "side" => "buy",
        //         "size" => 0.0007,
        //         "status" => "closed",
        //         "type" => "market"
        //     }
        //
        // createOrder (conditional, "stop", "trailingStop", or "takeProfit")
        //
        //     {
        //         "createdAt" => "2019-03-05T09:56:55.728933+00:00",
        //         "future" => "XRP-PERP",
        //         "id" => 9596912,
        //         "market" => "XRP-PERP",
        //         "triggerPrice" => 0.306525,
        //         "orderId" => null,
        //         "side" => "sell",
        //         "size" => 31431,
        //         "status" => "open",
        //         "type" => "stop",
        //         "orderPrice" => null,
        //         "error" => null,
        //         "triggeredAt" => null,
        //         "reduceOnly" => false
        //     }
        //
        // editOrder (conditional, stop, trailing stop, take profit)
        //
        //     {
        //         "createdAt" => "2019-03-05T09:56:55.728933+00:00",
        //         "future" => "XRP-PERP",
        //         "id" => 9596912,
        //         "market" => "XRP-PERP",
        //         "triggerPrice" => 0.306225,
        //         "orderId" => null,
        //         "side" => "sell",
        //         "size" => 31431,
        //         "status" => "open",
        //         "type" => "stop",
        //         "orderPrice" => null,
        //         "error" => null,
        //         "triggeredAt" => null,
        //         "reduceOnly" => false,
        //         "orderType" => "market",
        //         "filledSize" => 0,
        //         "avgFillPrice" => null,
        //         "retryUntilFilled" => false
        //     }
        //
        // canceled $order with a closed $status
        //
        //     {
        //         "avgFillPrice":null,
        //         "clientId":null,
        //         "createdAt":"2020-09-01T13:45:57.119695+00:00",
        //         "filledSize":0.0,
        //         "future":null,
        //         "id":8553541288,
        //         "ioc":false,
        //         "liquidation":false,
        //         "market":"XRP/USDT",
        //         "postOnly":false,
        //         "price":0.5,
        //         "reduceOnly":false,
        //         "remainingSize":0.0,
        //         "side":"sell",
        //         "size":46.0,
        //         "status":"closed",
        //         "type":"limit"
        //     }
        //
        $id = $this->safe_string($order, 'id');
        $timestamp = $this->parse8601($this->safe_string($order, 'createdAt'));
        $status = $this->parse_order_status($this->safe_string($order, 'status'));
        $amount = $this->safe_string($order, 'size');
        $filled = $this->safe_string($order, 'filledSize');
        $remaining = $this->safe_string($order, 'remainingSize');
        if (Precise::string_equals($remaining, '0')) {
            $remaining = Precise::string_sub($amount, $filled);
            if (Precise::string_gt($remaining, '0')) {
                $status = 'canceled';
            }
        }
        $marketId = $this->safe_string($order, 'market');
        $market = $this->safe_market($marketId, $market);
        $symbol = $market['symbol'];
        if ($symbol === null) {
            // support for delisted $market ids
            // https://github.com/ccxt/ccxt/issues/7113
            $symbol = $marketId;
        }
        $side = $this->safe_string($order, 'side');
        $type = $this->safe_string($order, 'type');
        $average = $this->safe_string($order, 'avgFillPrice');
        $price = $this->safe_string_2($order, 'price', 'triggerPrice', $average);
        $lastTradeTimestamp = $this->parse8601($this->safe_string($order, 'triggeredAt'));
        $clientOrderId = $this->safe_string($order, 'clientId');
        $stopPrice = $this->safe_number($order, 'triggerPrice');
        $postOnly = $this->safe_value($order, 'postOnly');
        $ioc = $this->safe_value($order, 'ioc');
        $timeInForce = null;
        if ($ioc) {
            $timeInForce = 'IOC';
        }
        if ($postOnly) {
            $timeInForce = 'PO';
        }
        return $this->safe_order(array(
            'info' => $order,
            'id' => $id,
            'clientOrderId' => $clientOrderId,
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'lastTradeTimestamp' => $lastTradeTimestamp,
            'symbol' => $symbol,
            'type' => $type,
            'timeInForce' => $timeInForce,
            'postOnly' => $postOnly,
            'reduceOnly' => $this->safe_value($order, 'reduceOnly'),
            'side' => $side,
            'price' => $price,
            'stopPrice' => $stopPrice,
            'amount' => $amount,
            'cost' => null,
            'average' => $average,
            'filled' => $filled,
            'remaining' => $remaining,
            'status' => $status,
            'fee' => null,
            'trades' => null,
        ), $market);
    }

    public function create_order($symbol, $type, $side, $amount, $price = null, $params = array ()) {
        $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'market' => $market['id'],
            'side' => $side, // "buy" or "sell"
            // 'price' => 0.306525, // send null for $market orders
            'type' => $type, // "limit", "market", "stop", "trailingStop", or "takeProfit"
            'size' => floatval($this->amount_to_precision($symbol, $amount)),
            // 'reduceOnly' => false, // optional, default is false
            // 'ioc' => false, // optional, default is false, limit or $market orders only
            // 'postOnly' => false, // optional, default is false, limit or $market orders only
            // 'clientId' => 'abcdef0123456789', // string, optional, client order id, limit or $market orders only
            // 'triggerPrice' => 0.306525, // required for stop and takeProfit orders
            // 'trailValue' => -0.306525, // required for trailingStop orders, negative for "sell"; positive for "buy"
            // 'orderPrice' => 0.306525, // optional, for stop and takeProfit orders only ($market by default). If not specified, a $market order will be submitted
        );
        $reduceOnly = $this->safe_value($params, 'reduceOnly');
        if ($reduceOnly === true) {
            $request['reduceOnly'] = $reduceOnly;
        }
        $clientOrderId = $this->safe_string_2($params, 'clientId', 'clientOrderId');
        if ($clientOrderId !== null) {
            $request['clientId'] = $clientOrderId;
            $params = $this->omit($params, array( 'clientId', 'clientOrderId' ));
        }
        $method = null;
        $stopPrice = $this->safe_value_2($params, 'stopPrice', 'triggerPrice');
        $params = $this->omit($params, array( 'stopPrice', 'triggerPrice' ));
        if ((($type === 'limit') || ($type === 'market')) && ($stopPrice === null)) {
            $method = 'privatePostOrders';
            if ($type === 'limit') {
                $request['price'] = floatval($this->price_to_precision($symbol, $price));
            } elseif ($type === 'market') {
                $request['price'] = null;
            }
            $timeInForce = $this->safe_string($params, 'timeInForce');
            $postOnly = $this->safe_value($params, 'postOnly', false);
            $params = $this->omit($params, array( 'timeInForce', 'postOnly' ));
            if ($timeInForce !== null) {
                if (!(($timeInForce === 'IOC') || ($timeInForce === 'PO'))) {
                    throw new InvalidOrder($this->id . ' createOrder () does not accept $timeInForce => ' . $timeInForce . ' orders, only IOC and PO orders are allowed');
                }
            }
            $maker = (($timeInForce === 'PO') || $postOnly);
            if (($type === 'market') && $maker) {
                throw new InvalidOrder($this->id . ' createOrder () does not accept $postOnly => true or $timeInForce => PO for $market orders');
            }
            $ioc = ($timeInForce === 'IOC');
            if ($maker) {
                $request['postOnly'] = true;
            }
            if ($ioc) {
                $request['ioc'] = true;
            }
        } elseif (($type === 'stop') || ($type === 'takeProfit') || ($stopPrice !== null)) {
            $method = 'privatePostConditionalOrders';
            if ($stopPrice === null) {
                throw new ArgumentsRequired($this->id . ' createOrder () requires a $stopPrice parameter or a triggerPrice parameter for ' . $type . ' orders');
            } else {
                $request['triggerPrice'] = floatval($this->price_to_precision($symbol, $stopPrice));
            }
            if (($type === 'limit') && ($price === null)) {
                throw new ArgumentsRequired($this->id . ' createOrder () requires a $price argument for stop limit orders');
            }
            if ($price !== null) {
                $request['orderPrice'] = floatval($this->price_to_precision($symbol, $price)); // optional, order $type is limit if this is specified, otherwise $market
            }
            if (($type === 'limit') || ($type === 'market')) {
                // default to stop orders for main argument
                $request['type'] = 'stop';
            }
        } elseif ($type === 'trailingStop') {
            $trailValue = $this->safe_number($params, 'trailValue', $price);
            if ($trailValue === null) {
                throw new ArgumentsRequired($this->id . ' createOrder () requires a $trailValue parameter or a $price argument (negative or positive) for a ' . $type . ' order');
            }
            $method = 'privatePostConditionalOrders';
            $request['trailValue'] = floatval($this->price_to_precision($symbol, $trailValue)); // negative for "sell", positive for "buy"
        } else {
            throw new InvalidOrder($this->id . ' createOrder () does not support order $type ' . $type . ', only limit, $market, stop, trailingStop, or takeProfit orders are supported');
        }
        $response = $this->$method (array_merge($request, $params));
        //
        // orders
        //
        //     {
        //         "success" => true,
        //         "result" => array(
        //             {
        //                 "createdAt" => "2019-03-05T09:56:55.728933+00:00",
        //                 "filledSize" => 0,
        //                 "future" => "XRP-PERP",
        //                 "id" => 9596912,
        //                 "market" => "XRP-PERP",
        //                 "price" => 0.306525,
        //                 "remainingSize" => 31431,
        //                 "side" => "sell",
        //                 "size" => 31431,
        //                 "status" => "open",
        //                 "type" => "limit",
        //                 "reduceOnly" => false,
        //                 "ioc" => false,
        //                 "postOnly" => false,
        //                 "clientId" => null,
        //             }
        //         )
        //     }
        //
        // conditional orders
        //
        //     {
        //         "success" => true,
        //         "result" => array(
        //             {
        //                 "createdAt" => "2019-03-05T09:56:55.728933+00:00",
        //                 "future" => "XRP-PERP",
        //                 "id" => 9596912,
        //                 "market" => "XRP-PERP",
        //                 "triggerPrice" => 0.306525,
        //                 "orderId" => null,
        //                 "side" => "sell",
        //                 "size" => 31431,
        //                 "status" => "open",
        //                 "type" => "stop",
        //                 "orderPrice" => null,
        //                 "error" => null,
        //                 "triggeredAt" => null,
        //                 "reduceOnly" => false
        //             }
        //         )
        //     }
        //
        //
        $result = $this->safe_value($response, 'result', array());
        return $this->parse_order($result, $market);
    }

    public function edit_order($id, $symbol, $type, $side, $amount, $price = null, $params = array ()) {
        $this->load_markets();
        $market = $this->market($symbol);
        $request = array();
        $method = null;
        $clientOrderId = $this->safe_string_2($params, 'client_order_id', 'clientOrderId');
        $triggerPrice = $this->safe_number($params, 'triggerPrice');
        $orderPrice = $this->safe_number($params, 'orderPrice');
        $trailValue = $this->safe_number($params, 'trailValue');
        $params = $this->omit($params, array( 'client_order_id', 'clientOrderId', 'triggerPrice', 'orderPrice', 'trailValue' ));
        $triggerPriceIsDefined = ($triggerPrice !== null);
        $orderPriceIsDefined = ($orderPrice !== null);
        $trailValueIsDefined = ($trailValue !== null);
        if ($triggerPriceIsDefined || $orderPriceIsDefined || $trailValueIsDefined) {
            $method = 'privatePostConditionalOrdersOrderIdModify';
            $request['order_id'] = $id;
            if ($triggerPriceIsDefined) {
                $request['triggerPrice'] = floatval($this->price_to_precision($symbol, $triggerPrice));
            }
            if ($orderPriceIsDefined) {
                // only for stop limit or take profit limit orders
                $request['orderPrice'] = floatval($this->price_to_precision($symbol, $orderPrice));
            }
            if ($trailValueIsDefined) {
                // negative for sell orders, positive for buy orders
                $request['trailValue'] = floatval($this->price_to_precision($symbol, $trailValue));
            }
        } else {
            if ($clientOrderId === null) {
                $method = 'privatePostOrdersOrderIdModify';
                $request['order_id'] = $id;
            } else {
                $method = 'privatePostOrdersByClientIdClientOrderIdModify';
                $request['client_order_id'] = $clientOrderId;
                // $request['clientId'] = $clientOrderId;
            }
            if ($price !== null) {
                $request['price'] = floatval($this->price_to_precision($symbol, $price));
            }
        }
        if ($amount !== null) {
            $request['size'] = floatval($this->amount_to_precision($symbol, $amount));
        }
        $response = $this->$method (array_merge($request, $params));
        //
        // regular order
        //
        //     {
        //         "success" => true,
        //         "result" => {
        //             "createdAt" => "2019-03-05T11:56:55.728933+00:00",
        //             "filledSize" => 0,
        //             "future" => "XRP-PERP",
        //             "id" => 9596932,
        //             "market" => "XRP-PERP",
        //             "price" => 0.326525,
        //             "remainingSize" => 31431,
        //             "side" => "sell",
        //             "size" => 31431,
        //             "status" => "open",
        //             "type" => "limit",
        //             "reduceOnly" => false,
        //             "ioc" => false,
        //             "postOnly" => false,
        //             "clientId" => null,
        //         }
        //     }
        //
        // conditional trigger order
        //
        //     {
        //         "success" => true,
        //         "result" => {
        //             "createdAt" => "2019-03-05T09:56:55.728933+00:00",
        //             "future" => "XRP-PERP",
        //             "id" => 9596912,
        //             "market" => "XRP-PERP",
        //             "triggerPrice" => 0.306225,
        //             "orderId" => null,
        //             "side" => "sell",
        //             "size" => 31431,
        //             "status" => "open",
        //             "type" => "stop",
        //             "orderPrice" => null,
        //             "error" => null,
        //             "triggeredAt" => null,
        //             "reduceOnly" => false,
        //             "orderType" => "market",
        //             "filledSize" => 0,
        //             "avgFillPrice" => null,
        //             "retryUntilFilled" => false
        //         }
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        return $this->parse_order($result, $market);
    }

    public function cancel_order($id, $symbol = null, $params = array ()) {
        $this->load_markets();
        $request = array();
        // support for canceling conditional orders
        // https://github.com/ccxt/ccxt/issues/6669
        $options = $this->safe_value($this->options, 'cancelOrder', array());
        $defaultMethod = $this->safe_string($options, 'method', 'privateDeleteOrdersOrderId');
        $method = $this->safe_string($params, 'method', $defaultMethod);
        $type = $this->safe_value($params, 'type');
        $clientOrderId = $this->safe_value_2($params, 'client_order_id', 'clientOrderId');
        if ($clientOrderId === null) {
            $request['order_id'] = intval($id);
            if (($type === 'stop') || ($type === 'trailingStop') || ($type === 'takeProfit')) {
                $method = 'privateDeleteConditionalOrdersOrderId';
            }
        } else {
            $request['client_order_id'] = $clientOrderId;
            $method = 'privateDeleteOrdersByClientIdClientOrderId';
        }
        $query = $this->omit($params, array( 'method', 'type', 'client_order_id', 'clientOrderId' ));
        $response = $this->$method (array_merge($request, $query));
        //
        //     {
        //         "success" => true,
        //         "result" => "Order queued for cancelation"
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        return $result;
    }

    public function cancel_all_orders($symbol = null, $params = array ()) {
        $this->load_markets();
        $request = array(
            // 'market' => market['id'], // optional
            // 'conditionalOrdersOnly' => false, // cancel conditional orders only
            // 'limitOrdersOnly' => false, // cancel existing limit orders (non-conditional orders) only
        );
        $marketId = $this->get_market_id($symbol, 'market', $params);
        if ($marketId !== null) {
            $request['market'] = $marketId;
        }
        $response = $this->privateDeleteOrders (array_merge($request, $params));
        $result = $this->safe_value($response, 'result', array());
        //
        //     {
        //         "success" => true,
        //         "result" => "Orders queued for cancelation"
        //     }
        //
        return $result;
    }

    public function fetch_order($id, $symbol = null, $params = array ()) {
        $this->load_markets();
        $request = array();
        $clientOrderId = $this->safe_value_2($params, 'client_order_id', 'clientOrderId');
        $method = 'privateGetOrdersOrderId';
        if ($clientOrderId === null) {
            $request['order_id'] = $id;
        } else {
            $request['client_order_id'] = $clientOrderId;
            $params = $this->omit($params, array( 'client_order_id', 'clientOrderId' ));
            $method = 'privateGetOrdersByClientIdClientOrderId';
        }
        $response = $this->$method (array_merge($request, $params));
        //
        //     {
        //         "success" => true,
        //         "result" => {
        //             "createdAt" => "2019-03-05T09:56:55.728933+00:00",
        //             "filledSize" => 10,
        //             "future" => "XRP-PERP",
        //             "id" => 9596912,
        //             "market" => "XRP-PERP",
        //             "price" => 0.306525,
        //             "avgFillPrice" => 0.306526,
        //             "remainingSize" => 31421,
        //             "side" => "sell",
        //             "size" => 31431,
        //             "status" => "open",
        //             "type" => "limit",
        //             "reduceOnly" => false,
        //             "ioc" => false,
        //             "postOnly" => false,
        //             "clientId" => null
        //         }
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        return $this->parse_order($result);
    }

    public function fetch_open_orders($symbol = null, $since = null, $limit = null, $params = array ()) {
        $this->load_markets();
        $request = array();
        list($market, $marketId) = $this->get_market_params($symbol, 'market', $params);
        if ($marketId !== null) {
            $request['market'] = $marketId;
        }
        // support for canceling conditional orders
        // https://github.com/ccxt/ccxt/issues/6669
        $options = $this->safe_value($this->options, 'fetchOpenOrders', array());
        $defaultMethod = $this->safe_string($options, 'method', 'privateGetOrders');
        $method = $this->safe_string($params, 'method', $defaultMethod);
        $type = $this->safe_value($params, 'type');
        $stop = $this->safe_value($params, 'stop');
        if ($stop || ($type === 'stop') || ($type === 'trailingStop') || ($type === 'takeProfit')) {
            $method = 'privateGetConditionalOrders';
        }
        $query = $this->omit($params, array( 'method', 'type', 'stop' ));
        $response = $this->$method (array_merge($request, $query));
        //
        //     {
        //         "success" => true,
        //         "result" => array(
        //             {
        //                 "createdAt" => "2019-03-05T09:56:55.728933+00:00",
        //                 "filledSize" => 10,
        //                 "future" => "XRP-PERP",
        //                 "id" => 9596912,
        //                 "market" => "XRP-PERP",
        //                 "price" => 0.306525,
        //                 "avgFillPrice" => 0.306526,
        //                 "remainingSize" => 31421,
        //                 "side" => "sell",
        //                 "size" => 31431,
        //                 "status" => "open",
        //                 "type" => "limit",
        //                 "reduceOnly" => false,
        //                 "ioc" => false,
        //                 "postOnly" => false,
        //                 "clientId" => null
        //             }
        //         )
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        return $this->parse_orders($result, $market, $since, $limit);
    }

    public function fetch_orders($symbol = null, $since = null, $limit = null, $params = array ()) {
        $this->load_markets();
        $request = array();
        list($market, $marketId) = $this->get_market_params($symbol, 'market', $params);
        if ($marketId !== null) {
            $request['market'] = $marketId;
        }
        if ($limit !== null) {
            $request['limit'] = $limit; // default 100, max 100
        }
        if ($since !== null) {
            $request['start_time'] = intval($since / 1000);
        }
        // support for canceling conditional orders
        // https://github.com/ccxt/ccxt/issues/6669
        $options = $this->safe_value($this->options, 'fetchOrders', array());
        $defaultMethod = $this->safe_string($options, 'method', 'privateGetOrdersHistory');
        $method = $this->safe_string($params, 'method', $defaultMethod);
        $type = $this->safe_value($params, 'type');
        if (($type === 'stop') || ($type === 'trailingStop') || ($type === 'takeProfit')) {
            $method = 'privateGetConditionalOrdersHistory';
        }
        $query = $this->omit($params, array( 'method', 'type' ));
        $response = $this->$method (array_merge($request, $query));
        //
        //     {
        //         "success" => true,
        //         "result" => array(
        //             {
        //                 "createdAt" => "2019-03-05T09:56:55.728933+00:00",
        //                 "filledSize" => 10,
        //                 "future" => "XRP-PERP",
        //                 "id" => 9596912,
        //                 "market" => "XRP-PERP",
        //                 "price" => 0.306525,
        //                 "avgFillPrice" => 0.306526,
        //                 "remainingSize" => 31421,
        //                 "side" => "sell",
        //                 "size" => 31431,
        //                 "status" => "open",
        //                 "type" => "limit",
        //                 "reduceOnly" => false,
        //                 "ioc" => false,
        //                 "postOnly" => false,
        //                 "clientId" => null
        //             }
        //         )
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        return $this->parse_orders($result, $market, $since, $limit);
    }

    public function fetch_order_trades($id, $symbol = null, $since = null, $limit = null, $params = array ()) {
        $request = array(
            'orderId' => $id,
        );
        return $this->fetch_my_trades($symbol, $since, $limit, array_merge($request, $params));
    }

    public function fetch_my_trades($symbol = null, $since = null, $limit = null, $params = array ()) {
        /**
         * fetch $trades specific to you account
         * @param {str} $symbol unified $market $symbol
         * @param {int} $since timestamp in ms of the earliest trade
         * @param {int} $limit not sent to exchange but filtered internally by CCXT
         * @param {dict} $params exchange specific parameters
         * @param {int} $params->till timestamp in ms of the latest trade
         * @return A list of {@link https://docs.ccxt.com/en/latest/manual.html#trade-structure trade structures}
         */
        $this->load_markets();
        list($market, $marketId) = $this->get_market_params($symbol, 'market', $params);
        $request = array();
        if ($marketId !== null) {
            $request['market'] = $marketId;
        }
        if ($market !== null) {
            $symbol = $market['symbol'];
        }
        $till = $this->safe_integer($params, 'till');
        if ($since !== null) {
            $request['start_time'] = intval($since / 1000);
            $request['end_time'] = $this->seconds();
        }
        if ($till !== null) {
            $request['end_time'] = intval($till / 1000);
            $params = $this->omit($params, 'till');
        }
        $response = $this->privateGetFills (array_merge($request, $params));
        //
        //     {
        //         "success" => true,
        //         "result" => array(
        //             {
        //                 "fee" => 20.1374935,
        //                 "feeRate" => 0.0005,
        //                 "future" => "EOS-0329",
        //                 "id" => 11215,
        //                 "liquidity" => "taker",
        //                 "market" => "EOS-0329",
        //                 "baseCurrency" => null,
        //                 "quoteCurrency" => null,
        //                 "orderId" => 8436981,
        //                 "price" => 4.201,
        //                 "side" => "buy",
        //                 "size" => 9587,
        //                 "time" => "2019-03-27T19:15:10.204619+00:00",
        //                 "type" => "order"
        //             }
        //         )
        //     }
        //
        $trades = $this->safe_value($response, 'result', array());
        return $this->parse_trades($trades, $market, $since, $limit);
    }

    public function transfer($code, $amount, $fromAccount, $toAccount, $params = array ()) {
        $this->load_markets();
        $currency = $this->currency($code);
        $request = array(
            'coin' => $currency['id'],
            'source' => $fromAccount,
            'destination' => $toAccount,
            'size' => $amount,
        );
        $response = $this->privatePostSubaccountsTransfer (array_merge($request, $params));
        //
        //     {
        //         success => true,
        //         $result => {
        //             id => '31222278',
        //             coin => 'USDT',
        //             size => '1.0',
        //             time => '2022-04-01T11:18:27.194188+00:00',
        //             notes => 'Transfer from main account to testSubaccount',
        //             status => 'complete'
        //         }
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        return $this->parse_transfer($result, $currency);
    }

    public function parse_transfer($transfer, $currency = null) {
        //
        //     {
        //         id => '31222278',
        //         coin => 'USDT',
        //         size => '1.0',
        //         time => '2022-04-01T11:18:27.194188+00:00',
        //         $notes => 'Transfer from main account to testSubaccount',
        //         $status => 'complete'
        //     }
        //
        $currencyId = $this->safe_string($transfer, 'coin');
        $notes = $this->safe_string($transfer, 'notes', '');
        $status = $this->safe_string($transfer, 'status');
        $fromTo = str_replace('Transfer from ', '', $notes);
        $parts = explode(' to ', $fromTo);
        $fromAccount = $this->safe_string($parts, 0);
        $fromAccount = str_replace(' account', '', $fromAccount);
        $toAccount = $this->safe_string($parts, 1);
        $toAccount = str_replace(' account', '', $toAccount);
        return array(
            'info' => $transfer,
            'id' => $this->safe_string($transfer, 'id'),
            'timestamp' => null,
            'datetime' => $this->safe_string($transfer, 'time'),
            'currency' => $this->safe_currency_code($currencyId, $currency),
            'amount' => $this->safe_number($transfer, 'size'),
            'fromAccount' => $fromAccount,
            'toAccount' => $toAccount,
            'status' => $this->parse_transfer_status($status),
        );
    }

    public function parse_transfer_status($status) {
        $statuses = array(
            'complete' => 'ok',
        );
        return $this->safe_string($statuses, $status, $status);
    }

    public function withdraw($code, $amount, $address, $tag = null, $params = array ()) {
        list($tag, $params) = $this->handle_withdraw_tag_and_params($tag, $params);
        $this->load_markets();
        $this->check_address($address);
        $currency = $this->currency($code);
        $request = array(
            'coin' => $currency['id'],
            'size' => $amount,
            'address' => $address,
            // 'password' => 'string', // optional withdrawal password if it is required for your account
            // 'code' => '192837', // optional 2fa $code if it is required for your account
        );
        if ($this->password !== null) {
            $request['password'] = $this->password;
        }
        if ($tag !== null) {
            $request['tag'] = $tag;
        }
        $networks = $this->safe_value($this->options, 'networks', array());
        $network = $this->safe_string_upper($params, 'network'); // this line allows the user to specify either ERC20 or ETH
        $network = $this->safe_string_lower($networks, $network, $network); // handle ERC20>ETH alias
        if ($network !== null) {
            $request['method'] = $network;
            $params = $this->omit($params, 'network');
        }
        $response = $this->privatePostWalletWithdrawals (array_merge($request, $params));
        //
        //     {
        //         "success" => true,
        //         "result" => {
        //             "coin" => "USDTBEAR",
        //             "address" => "0x83a127952d266A6eA306c40Ac62A4a70668FE3BE",
        //             "tag" => "null",
        //             "fee" => 0,
        //             "id" => 1,
        //             "size" => "20.2",
        //             "status" => "requested",
        //             "time" => "2019-03-05T09:56:55.728933+00:00",
        //             "txid" => "null"
        //         }
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        return $this->parse_transaction($result, $currency);
    }

    public function fetch_positions($symbols = null, $params = array ()) {
        $this->load_markets();
        $request = array(
            'showAvgPrice' => true,
        );
        $response = $this->privateGetPositions (array_merge($request, $params));
        //
        //     {
        //         "success" => true,
        //         "result" => array(
        //             {
        //                 "cost" => -31.7906,
        //                 "entryPrice" => 138.22,
        //                 "estimatedLiquidationPrice" => 152.1,
        //                 "future" => "ETH-PERP",
        //                 "initialMarginRequirement" => 0.1,
        //                 "longOrderSize" => 1744.55,
        //                 "maintenanceMarginRequirement" => 0.04,
        //                 "netSize" => -0.23,
        //                 "openSize" => 1744.32,
        //                 "realizedPnl" => 3.39441714,
        //                 "shortOrderSize" => 1732.09,
        //                 "recentAverageOpenPrice" => 278.98,
        //                 "recentPnl" => 2.44,
        //                 "recentBreakEvenPrice" => 278.98,
        //                 "side" => "sell",
        //                 "size" => 0.23,
        //                 "unrealizedPnl" => 0,
        //                 "collateralUsed" => 3.17906
        //             }
        //         )
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        return $this->parse_positions($result, $symbols);
    }

    public function parse_position($position, $market = null) {
        //
        //   {
        //     "future" => "XMR-PERP",
        //     "size" => "0.0",
        //     "side" => "buy",
        //     "netSize" => "0.0",
        //     "longOrderSize" => "0.0",
        //     "shortOrderSize" => "0.0",
        //     "cost" => "0.0",
        //     "entryPrice" => null,
        //     "unrealizedPnl" => "0.0",
        //     "realizedPnl" => "0.0",
        //     "initialMarginRequirement" => "0.02",
        //     "maintenanceMarginRequirement" => "0.006",
        //     "openSize" => "0.0",
        //     "collateralUsed" => "0.0",
        //     "estimatedLiquidationPrice" => null
        //   }
        //
        $contractsString = $this->safe_string($position, 'size');
        $rawSide = $this->safe_string($position, 'side');
        $side = ($rawSide === 'buy') ? 'long' : 'short';
        $marketId = $this->safe_string($position, 'future');
        $symbol = $this->safe_symbol($marketId, $market);
        $liquidationPriceString = $this->safe_string($position, 'estimatedLiquidationPrice');
        $initialMarginPercentage = $this->safe_string($position, 'initialMarginRequirement');
        $leverage = intval(Precise::string_div('1', $initialMarginPercentage, 0));
        // on ftx the entryPrice is actually the mark price
        $markPriceString = $this->safe_string($position, 'entryPrice');
        $notionalString = Precise::string_mul($contractsString, $markPriceString);
        $initialMargin = Precise::string_mul($notionalString, $initialMarginPercentage);
        $maintenanceMarginPercentageString = $this->safe_string($position, 'maintenanceMarginRequirement');
        $maintenanceMarginString = Precise::string_mul($notionalString, $maintenanceMarginPercentageString);
        $unrealizedPnlString = $this->safe_string($position, 'recentPnl');
        $percentage = $this->parse_number(Precise::string_mul(Precise::string_div($unrealizedPnlString, $initialMargin, 4), '100'));
        $entryPriceString = $this->safe_string($position, 'recentAverageOpenPrice');
        $difference = null;
        $collateral = null;
        $marginRatio = null;
        if (($entryPriceString !== null) && (Precise::string_gt($liquidationPriceString, '0'))) {
            // $collateral = maintenanceMargin ± ((markPrice - liquidationPrice) * size)
            if ($side === 'long') {
                $difference = Precise::string_sub($markPriceString, $liquidationPriceString);
            } else {
                $difference = Precise::string_sub($liquidationPriceString, $markPriceString);
            }
            $loss = Precise::string_mul($difference, $contractsString);
            $collateral = Precise::string_add($loss, $maintenanceMarginString);
            $marginRatio = $this->parse_number(Precise::string_div($maintenanceMarginString, $collateral, 4));
        }
        // ftx has a weird definition of realizedPnl
        // it keeps the historical record of the realizedPnl per contract forever
        // so we cannot use this data
        return array(
            'info' => $position,
            'symbol' => $symbol,
            'timestamp' => null,
            'datetime' => null,
            'initialMargin' => $this->parse_number($initialMargin),
            'initialMarginPercentage' => $this->parse_number($initialMarginPercentage),
            'maintenanceMargin' => $this->parse_number($maintenanceMarginString),
            'maintenanceMarginPercentage' => $this->parse_number($maintenanceMarginPercentageString),
            'entryPrice' => $this->parse_number($entryPriceString),
            'notional' => $this->parse_number($notionalString),
            'leverage' => $leverage,
            'unrealizedPnl' => $this->parse_number($unrealizedPnlString),
            'contracts' => $this->parse_number($contractsString),
            'contractSize' => $this->safe_value($market, 'contractSize'),
            'marginRatio' => $marginRatio,
            'liquidationPrice' => $this->parse_number($liquidationPriceString),
            'markPrice' => $this->parse_number($markPriceString),
            'collateral' => $this->parse_number($collateral),
            'marginMode' => 'cross',
            'marginType' => 'cross', // deprecated
            'side' => $side,
            'percentage' => $percentage,
        );
    }

    public function fetch_deposit_address($code, $params = array ()) {
        $this->load_markets();
        $currency = $this->currency($code);
        $request = array(
            'coin' => $currency['id'],
        );
        $networks = $this->safe_value($this->options, 'networks', array());
        $network = $this->safe_string_upper($params, 'network'); // this line allows the user to specify either ERC20 or ETH
        $network = $this->safe_string_lower($networks, $network, $network); // handle ERC20>ETH alias
        if ($network !== null) {
            $request['method'] = $network;
            $params = $this->omit($params, 'network');
        }
        $response = $this->privateGetWalletDepositAddressCoin (array_merge($request, $params));
        //
        //     {
        //         "success" => true,
        //         "result" => {
        //             "address" => "0x83a127952d266A6eA306c40Ac62A4a70668FE3BE",
        //             "tag" => null,
        //             "method" => "erc20",
        //             "coin" => null
        //         }
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        $networkId = $this->safe_string($result, 'method');
        $address = $this->safe_string($result, 'address');
        $this->check_address($address);
        return array(
            'currency' => $code,
            'address' => $address,
            'tag' => $this->safe_string($result, 'tag'),
            'network' => $this->safe_network($networkId),
            'info' => $response,
        );
    }

    public function safe_network($networkId) {
        $networksById = array(
            'trx' => 'TRC20',
            'erc20' => 'ERC20',
            'sol' => 'SOL',
            'bsc' => 'BEP20',
            'bep2' => 'BEP2',
        );
        return $this->safe_string($networksById, $networkId, $networkId);
    }

    public function parse_transaction_status($status) {
        $statuses = array(
            // what are other $statuses here?
            'confirmed' => 'ok', // deposits
            'complete' => 'ok', // withdrawals
            'cancelled' => 'canceled', // deposits
        );
        return $this->safe_string($statuses, $status, $status);
    }

    public function parse_transaction($transaction, $currency = null) {
        //
        // fetchDeposits
        //
        //     airdrop
        //
        //     {
        //         "id" => 9147072,
        //         "coin" => "SRM_LOCKED",
        //         "size" => 3.12,
        //         "time" => "2021-04-27T23:59:03.565983+00:00",
        //         "notes" => "SRM Airdrop for FTT holdings",
        //         "status" => "complete"
        //     }
        //
        //     regular deposits
        //
        //     {
        //         "coin" => "TUSD",
        //         "confirmations" => 64,
        //         "confirmedTime" => "2019-03-05T09:56:55.728933+00:00",
        //         "fee" => 0,
        //         "id" => 1,
        //         "sentTime" => "2019-03-05T09:56:55.735929+00:00",
        //         "size" => "99.0",
        //         "status" => "confirmed",
        //         "time" => "2019-03-05T09:56:55.728933+00:00",
        //         "txid" => "0x8078356ae4b06a036d64747546c274af19581f1c78c510b60505798a7ffcaf1"
        //     }
        //
        // fetchWithdrawals
        //
        //     {
        //         "coin" => "TUSD",
        //         "address" => "0x83a127952d266A6eA306c40Ac62A4a70668FE3BE",
        //         "tag" => "null",
        //         "fee" => 0,
        //         "id" => 1,
        //         "size" => "99.0",
        //         "status" => "complete",
        //         "time" => "2019-03-05T09:56:55.728933+00:00",
        //         "txid" => "0x8078356ae4b06a036d64747546c274af19581f1c78c510b60505798a7ffcaf1"
        //     }
        //
        //     {
        //         "coin" => 'BTC',
        //         "id" => 1969806,
        //         "notes" => 'Transfer to Dd6gi7m2Eg4zzBbPAxuwfEaHs6tYvyUX5hbPpsTcNPXo',
        //         "size" => 0.003,
        //         "status" => 'complete',
        //         "time" => '2021-02-03T20:28:54.918146+00:00'
        //     }
        //
        $code = $this->safe_currency_code($this->safe_string($transaction, 'coin'));
        $id = $this->safe_string($transaction, 'id');
        $amount = $this->safe_number($transaction, 'size');
        $status = $this->parse_transaction_status($this->safe_string($transaction, 'status'));
        $timestamp = $this->parse8601($this->safe_string($transaction, 'time'));
        $txid = $this->safe_string($transaction, 'txid');
        $tag = null;
        $address = $this->safe_value($transaction, 'address');
        if (gettype($address) !== 'string') {
            $tag = $this->safe_string($address, 'tag');
            $address = $this->safe_string($address, 'address');
        }
        if ($address === null) {
            // parse $address from internal transfer
            $notes = $this->safe_string($transaction, 'notes');
            if (($notes !== null) && (mb_strpos($notes, 'Transfer to') !== false)) {
                $address = mb_substr($notes, 12);
            }
        }
        $fee = $this->safe_number($transaction, 'fee');
        return array(
            'info' => $transaction,
            'id' => $id,
            'txid' => $txid,
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'network' => null,
            'addressFrom' => null,
            'address' => $address,
            'addressTo' => $address,
            'tagFrom' => null,
            'tag' => $tag,
            'tagTo' => $tag,
            'type' => null,
            'amount' => $amount,
            'currency' => $code,
            'status' => $status,
            'updated' => null,
            'fee' => array(
                'currency' => $code,
                'cost' => $fee,
                'rate' => null,
            ),
        );
    }

    public function fetch_deposits($code = null, $since = null, $limit = null, $params = array ()) {
        $this->load_markets();
        $response = $this->privateGetWalletDeposits ($params);
        //
        //     {
        //         "success" => true,
        //         "result" => {
        //             "coin" => "TUSD",
        //             "confirmations" => 64,
        //             "confirmedTime" => "2019-03-05T09:56:55.728933+00:00",
        //             "fee" => 0,
        //             "id" => 1,
        //             "sentTime" => "2019-03-05T09:56:55.735929+00:00",
        //             "size" => "99.0",
        //             "status" => "confirmed",
        //             "time" => "2019-03-05T09:56:55.728933+00:00",
        //             "txid" => "0x8078356ae4b06a036d64747546c274af19581f1c78c510b60505798a7ffcaf1"
        //         }
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        $currency = null;
        if ($code !== null) {
            $currency = $this->currency($code);
        }
        return $this->parse_transactions($result, $currency, $since, $limit, array( 'type' => 'deposit' ));
    }

    public function fetch_withdrawals($code = null, $since = null, $limit = null, $params = array ()) {
        $this->load_markets();
        $response = $this->privateGetWalletWithdrawals ($params);
        //
        //     {
        //         "success" => true,
        //         "result" => {
        //             "coin" => "TUSD",
        //             "address" => "0x83a127952d266A6eA306c40Ac62A4a70668FE3BE",
        //             "tag" => "null",
        //             "fee" => 0,
        //             "id" => 1,
        //             "size" => "99.0",
        //             "status" => "complete",
        //             "time" => "2019-03-05T09:56:55.728933+00:00",
        //             "txid" => "0x8078356ae4b06a036d64747546c274af19581f1c78c510b60505798a7ffcaf1"
        //         }
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        $currency = null;
        if ($code !== null) {
            $currency = $this->currency($code);
        }
        return $this->parse_transactions($result, $currency, $since, $limit, array( 'type' => 'withdrawal' ));
    }

    public function sign($path, $api = 'public', $method = 'GET', $params = array (), $headers = null, $body = null) {
        $request = '/api/' . $this->implode_params($path, $params);
        $signOptions = $this->safe_value($this->options, 'sign', array());
        $headerPrefix = $this->safe_string($signOptions, $this->hostname, 'FTX');
        $subaccountField = $headerPrefix . '-SUBACCOUNT';
        $chosenSubaccount = $this->safe_string_2($params, $subaccountField, 'subaccount');
        if ($chosenSubaccount !== null) {
            $params = $this->omit($params, array( $subaccountField, 'subaccount' ));
        }
        $query = $this->omit($params, $this->extract_params($path));
        $baseUrl = $this->implode_hostname($this->urls['api'][$api]);
        $url = $baseUrl . $request;
        if ($method !== 'POST') {
            if ($query) {
                $suffix = '?' . $this->urlencode($query);
                $url .= $suffix;
                $request .= $suffix;
            }
        }
        if ($api === 'private') {
            $this->check_required_credentials();
            $timestamp = (string) $this->milliseconds();
            $auth = $timestamp . $method . $request;
            $headers = array();
            if (($method === 'POST') || ($method === 'DELETE')) {
                $body = $this->json($query);
                $auth .= $body;
                $headers['Content-Type'] = 'application/json';
            }
            $signature = $this->hmac($this->encode($auth), $this->encode($this->secret), 'sha256');
            $headers[$headerPrefix . '-KEY'] = $this->apiKey;
            $headers[$headerPrefix . '-TS'] = $timestamp;
            $headers[$headerPrefix . '-SIGN'] = $signature;
            if ($chosenSubaccount !== null) {
                $headers[$subaccountField] = $chosenSubaccount;
            }
        }
        return array( 'url' => $url, 'method' => $method, 'body' => $body, 'headers' => $headers );
    }

    public function handle_errors($code, $reason, $url, $method, $headers, $body, $response, $requestHeaders, $requestBody) {
        if ($response === null) {
            return; // fallback to the default $error handler
        }
        //
        //     array("error":"Invalid parameter start_time","success":false)
        //     array("error":"Not enough balances","success":false)
        //
        $success = $this->safe_value($response, 'success');
        if (!$success) {
            $feedback = $this->id . ' ' . $body;
            $error = $this->safe_string($response, 'error');
            $this->throw_exactly_matched_exception($this->exceptions['exact'], $error, $feedback);
            $this->throw_broadly_matched_exception($this->exceptions['broad'], $error, $feedback);
            throw new ExchangeError($feedback); // unknown message
        }
    }

    public function set_leverage($leverage, $symbol = null, $params = array ()) {
        // WARNING => THIS WILL INCREASE LIQUIDATION PRICE FOR OPEN ISOLATED LONG POSITIONS
        // AND DECREASE LIQUIDATION PRICE FOR OPEN ISOLATED SHORT POSITIONS
        if (($leverage < 1) || ($leverage > 20)) {
            throw new BadRequest($this->id . ' setLeverage () $leverage should be between 1 and 20');
        }
        $request = array(
            'leverage' => $leverage,
        );
        return $this->privatePostAccountLeverage (array_merge($request, $params));
    }

    public function parse_income($income, $market = null) {
        //
        //   {
        //       "future" => "ETH-PERP",
        //        "id" => 33830,
        //        "payment" => 0.0441342,
        //        "time" => "2019-05-15T18:00:00+00:00",
        //        "rate" => 0.0001
        //   }
        //
        $marketId = $this->safe_string($income, 'future');
        $symbol = $this->safe_symbol($marketId, $market);
        $amount = $this->safe_number($income, 'payment');
        $code = $this->safe_currency_code('USD');
        $id = $this->safe_string($income, 'id');
        $time = $this->safe_string($income, 'time');
        $timestamp = $this->parse8601($time);
        $rate = $this->safe_number($income, 'rate');
        return array(
            'info' => $income,
            'symbol' => $symbol,
            'code' => $code,
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'id' => $id,
            'amount' => $amount,
            'rate' => $rate,
        );
    }

    public function parse_incomes($incomes, $market = null, $since = null, $limit = null) {
        $result = array();
        for ($i = 0; $i < count($incomes); $i++) {
            $entry = $incomes[$i];
            $parsed = $this->parse_income($entry, $market);
            $result[] = $parsed;
        }
        $sorted = $this->sort_by($result, 'timestamp');
        return $this->filter_by_since_limit($sorted, $since, $limit, 'timestamp');
    }

    public function fetch_funding_history($symbol = null, $since = null, $limit = null, $params = array ()) {
        $this->load_markets();
        $request = array();
        $market = null;
        if ($symbol !== null) {
            $market = $this->market($symbol);
            $request['future'] = $market['id'];
        }
        if ($since !== null) {
            $request['startTime'] = $since;
        }
        $response = $this->privateGetFundingPayments (array_merge($request, $params));
        $result = $this->safe_value($response, 'result', array());
        return $this->parse_incomes($result, $market, $since, $limit);
    }

    public function parse_funding_rate($fundingRate, $market = null) {
        //
        // perp
        //     {
        //       "volume" => "71294.7636",
        //       "nextFundingRate" => "0.000033",
        //       "nextFundingTime" => "2021-10-14T20:00:00+00:00",
        //       "openInterest" => "47142.994"
        //     }
        //
        // delivery
        //     {
        //       "volume" => "4998.727",
        //       "predictedExpirationPrice" => "3798.820141757",
        //       "openInterest" => "48307.96"
        //     }
        //
        $fundingRateDatetimeRaw = $this->safe_string($fundingRate, 'nextFundingTime');
        $fundingRateTimestamp = $this->parse8601($fundingRateDatetimeRaw);
        $estimatedSettlePrice = $this->safe_number($fundingRate, 'predictedExpirationPrice');
        return array(
            'info' => $fundingRate,
            'symbol' => $market['symbol'],
            'markPrice' => null,
            'indexPrice' => null,
            'interestRate' => $this->parse_number('0'),
            'estimatedSettlePrice' => $estimatedSettlePrice,
            'timestamp' => null,
            'datetime' => null,
            'fundingRate' => $this->safe_number($fundingRate, 'nextFundingRate'),
            'fundingTimestamp' => $fundingRateTimestamp,
            'fundingDatetime' => $this->iso8601($fundingRateTimestamp),
            'nextFundingRate' => null,
            'nextFundingTimestamp' => null,
            'nextFundingDatetime' => null,
            'previousFundingRate' => null,
            'previousFundingTimestamp' => null,
            'previousFundingDatetime' => null,
        );
    }

    public function fetch_funding_rate($symbol, $params = array ()) {
        $this->load_markets();
        $market = $this->market($symbol);
        $request = array(
            'future_name' => $market['id'],
        );
        $response = $this->publicGetFuturesFutureNameStats (array_merge($request, $params));
        //
        //     {
        //       "success" => true,
        //       "result" => {
        //         "volume" => "71294.7636",
        //         "nextFundingRate" => "0.000033",
        //         "nextFundingTime" => "2021-10-14T20:00:00+00:00",
        //         "openInterest" => "47142.994"
        //       }
        //     }
        //
        $result = $this->safe_value($response, 'result', array());
        return $this->parse_funding_rate($result, $market);
    }

    public function fetch_borrow_rates($params = array ()) {
        $this->load_markets();
        $response = $this->privateGetSpotMarginBorrowRates ($params);
        //
        //     {
        //         "success":true,
        //         "result":array(
        //             array("coin":"1INCH","previous":4.8763e-6,"estimate":4.8048e-6),
        //             array("coin":"AAPL","previous":0.0000326469,"estimate":0.0000326469),
        //             array("coin":"AAVE","previous":1.43e-6,"estimate":1.43e-6),
        //         )
        //     }
        //
        $result = $this->safe_value($response, 'result');
        return $this->parse_borrow_rates($result, 'coin');
    }

    public function fetch_borrow_rate_histories($codes = null, $since = null, $limit = null, $params = array ()) {
        /**
         * Gets the history of the borrow rate for mutiple currencies
         * @param {str} code Unified $currency code
         * @param {int} $since Timestamp in ms of the earliest time to fetch the borrow rate
         * @param {int} $limit Max number of {@link https://docs.ccxt.com/en/latest/manual.html#borrow-rate-structure borrow rate structures} to return per $currency, max=48 for multiple currencies, max=5000 for a single $currency
         * @param {dict} $params Exchange specific parameters
         * @param {dict} $params->till Timestamp in ms of the latest time to fetch the borrow rate
         * @return A dictionary of {@link https://docs.ccxt.com/en/latest/manual.html#borrow-rate-structure borrow rate structures} with unified $currency $codes as keys
         */
        $this->load_markets();
        $request = array();
        $numCodes = 0;
        $endTime = $this->safe_number_2($params, 'till', 'end_time');
        if ($codes !== null) {
            $numCodes = is_array($codes) ? count($codes) : 0;
        }
        if ($numCodes === 1) {
            $millisecondsPer5000Hours = 18000000000;
            if (($limit !== null) && ($limit > 5000)) {
                throw new BadRequest($this->id . ' fetchBorrowRateHistories () $limit cannot exceed 5000 for a single currency');
            }
            if (($endTime !== null) && ($since !== null) && (($endTime - $since) > $millisecondsPer5000Hours)) {
                throw new BadRequest($this->id . ' fetchBorrowRateHistories () requires the time range between the $since time and the end time to be less than 5000 hours for a single currency');
            }
            $currency = $this->currency($codes[0]);
            $request['coin'] = $currency['id'];
        } else {
            $millisecondsPer2Days = 172800000;
            if (($limit !== null) && ($limit > 48)) {
                throw new BadRequest($this->id . ' fetchBorrowRateHistories () $limit cannot exceed 48 for multiple currencies');
            }
            if (($endTime !== null) && ($since !== null) && (($endTime - $since) > $millisecondsPer2Days)) {
                throw new BadRequest($this->id . ' fetchBorrowRateHistories () requires the time range between the $since time and the end time to be less than 48 hours for multiple currencies');
            }
        }
        $millisecondsPerHour = 3600000;
        if ($since !== null) {
            $request['start_time'] = intval($since / 1000);
            if ($endTime === null) {
                $now = $this->milliseconds();
                $sinceLimit = ($limit === null) ? 2 : $limit;
                $endTime = $this->sum($since, $millisecondsPerHour * ($sinceLimit - 1));
                $endTime = min ($endTime, $now);
            }
        } else {
            if ($limit !== null) {
                if ($endTime === null) {
                    $endTime = $this->milliseconds();
                }
                $startTime = $this->sum(($endTime - $millisecondsPerHour * $limit), 1000);
                $request['start_time'] = intval($startTime / 1000);
            }
        }
        if ($endTime !== null) {
            $request['end_time'] = intval($endTime / 1000);
        }
        $response = $this->publicGetSpotMarginHistory (array_merge($request, $params));
        //
        //    {
        //        "success" => true,
        //        "result" => array(
        //            array(
        //                "coin" => "PYPL",
        //                "time" => "2022-01-24T13:00:00+00:00",
        //                "size" => 0.00500172,
        //                "rate" => 1e-6
        //            ),
        //            ...
        //        )
        //    }
        //
        $result = $this->safe_value($response, 'result');
        return $this->parse_borrow_rate_histories($result, $codes, $since, $limit);
    }

    public function fetch_borrow_rate_history($code, $since = null, $limit = null, $params = array ()) {
        /**
         * retrieves a history of a currencies borrow interest rate at specific time slots
         * @param {str} $code unified currency $code
         * @param {int|null} $since timestamp for the earliest borrow rate
         * @param {int|null} $limit the maximum number of {@link https://docs.ccxt.com/en/latest/manual.html#borrow-rate-structure borrow rate structures} to retrieve
         * @param {dict} $params extra parameters specific to the exchange api endpoint
         * @param {dict} $params->till Timestamp in ms of the latest time to fetch the borrow rate
         * @return {[dict]} an array of {@link https://docs.ccxt.com/en/latest/manual.html#borrow-rate-structure borrow rate structures}
         */
        $histories = $this->fetch_borrow_rate_histories(array( $code ), $since, $limit, $params);
        $borrowRateHistory = $this->safe_value($histories, $code);
        if ($borrowRateHistory === null) {
            throw new BadRequest($this->id . ' fetchBorrowRateHistory () returned no data for ' . $code);
        }
        return $borrowRateHistory;
    }

    public function parse_borrow_rate_histories($response, $codes, $since, $limit) {
        // How to calculate borrow rate
        // https://help.ftx.com/hc/en-us/articles/360053007671-Spot-Margin-Trading-Explainer
        $takerFee = (string) $this->fees['trading']['taker'];
        $spotMarginBorrowRate = Precise::string_mul('500', $takerFee);
        $borrowRateHistories = array();
        for ($i = 0; $i < count($response); $i++) {
            $item = $response[$i];
            $code = $this->safe_currency_code($this->safe_string($item, 'coin'));
            if ($codes === null || $this->in_array($code, $codes)) {
                if (!(is_array($borrowRateHistories) && array_key_exists($code, $borrowRateHistories))) {
                    $borrowRateHistories[$code] = array();
                }
                $lendingRate = $this->safe_string($item, 'rate');
                $borrowRate = Precise::string_mul($lendingRate, Precise::string_add('1', $spotMarginBorrowRate));
                $borrowRateStructure = array_merge($this->parse_borrow_rate($item), array( 'rate' => $borrowRate ));
                $borrowRateHistories[$code][] = $borrowRateStructure;
            }
        }
        $keys = is_array($borrowRateHistories) ? array_keys($borrowRateHistories) : array();
        for ($i = 0; $i < count($keys); $i++) {
            $code = $keys[$i];
            $borrowRateHistories[$code] = $this->filter_by_currency_since_limit($borrowRateHistories[$code], $code, $since, $limit);
        }
        return $borrowRateHistories;
    }

    public function parse_borrow_rates($response, $codeKey) {
        $result = array();
        for ($i = 0; $i < count($response); $i++) {
            $item = $response[$i];
            $currency = $this->safe_string($item, $codeKey);
            $code = $this->safe_currency_code($currency);
            $borrowRate = $this->parse_borrow_rate($item);
            $result[$code] = $borrowRate;
        }
        return $result;
    }

    public function parse_borrow_rate($info, $currency = null) {
        //
        //    {
        //        "coin" => "1INCH",
        //        "previous" => 0.0000462375,
        //        "estimate" => 0.0000462375
        //    }
        //
        $coin = $this->safe_string($info, 'coin');
        $datetime = $this->safe_string($info, 'time');
        $timestamp = $this->parse8601($datetime);
        return array(
            'currency' => $this->safe_currency_code($coin),
            'rate' => $this->safe_number($info, 'previous'),
            'period' => 3600000,
            'timestamp' => $timestamp,
            'datetime' => $this->iso8601($timestamp),
            'info' => $info,
        );
    }

    public function fetch_borrow_interest($code = null, $symbol = null, $since = null, $limit = null, $params = array ()) {
        $this->load_markets();
        $request = array();
        if ($since !== null) {
            $request['start_time'] = intval($since / 1000);
        }
        $response = $this->privateGetSpotMarginBorrowHistory (array_merge($request, $params));
        //
        //     {
        //         "success":true,
        //         "result":array(
        //             array("coin":"USDT","time":"2021-12-26T01:00:00+00:00","size":4593.74214725,"rate":3.3003e-6,"cost":0.0151607272085692,"feeUsd":0.0151683341034461),
        //             array("coin":"USDT","time":"2021-12-26T00:00:00+00:00","size":4593.97110361,"rate":3.3003e-6,"cost":0.0151614828332441,"feeUsd":0.015169697173028324),
        //             array("coin":"USDT","time":"2021-12-25T23:00:00+00:00","size":4594.20005922,"rate":3.3003e-6,"cost":0.0151622384554438,"feeUsd":0.015170200298479137),
        //         )
        //     }
        //
        $result = $this->safe_value($response, 'result');
        $interest = $this->parse_borrow_interests($result);
        return $this->filter_by_currency_since_limit($interest, $code, $since, $limit);
    }

    public function parse_borrow_interest($info, $market = null) {
        $coin = $this->safe_string($info, 'coin');
        $datetime = $this->safe_string($info, 'time');
        return array(
            'account' => 'cross',
            'symbol' => null,
            'marginMode' => 'cross',
            'marginType' => 'cross', // deprecated
            'currency' => $this->safe_currency_code($coin),
            'interest' => $this->safe_number($info, 'cost'),
            'interestRate' => $this->safe_number($info, 'rate'),
            'amountBorrowed' => $this->safe_number($info, 'size'),
            'timestamp' => $this->parse8601($datetime),
            'datetime' => $datetime,
            'info' => $info,
        );
    }
}
